@@ -14,6 +14,8 @@
 using namespace std;
 using namespace PoDoFo;
 
+namespace
+{
 template <typename TStream>
 size_t getPosition(TStream& stream)
 {
@@ -65,24 +67,26 @@ size_t getLength(TStream& stream)
 template <typename TStream>
 void seek(TStream& stream, ssize_t pos, SeekDirection direction)
 {
-    ios_base::seekdir seekdir;
     switch (direction)
     {
         case SeekDirection::Begin:
-            seekdir = ios_base::beg;
+            (void)utls::stream_helper<TStream>::seek(stream, (std::streampos)pos, ios_base::beg);
             break;
         case SeekDirection::Current:
-            seekdir = ios_base::cur;
+            (void)utls::stream_helper<TStream>::seek(stream, (std::streampos)pos, ios_base::cur);
             break;
         case SeekDirection::End:
-            seekdir = ios_base::end;
+            (void)utls::stream_helper<TStream>::seek(stream, (std::streampos)pos, ios_base::end);
             break;
         default:
             PODOFO_RAISE_ERROR(PdfErrorCode::InvalidEnumValue);
     }
-    (void)utls::stream_helper<TStream>::seek(stream, (std::streampos)pos, seekdir);
 }
 
+}
+
+static FILE* createFile(const string_view& filename, FileMode mode, DeviceAccess access);
+
 StreamDevice::StreamDevice(DeviceAccess access)
     : InputStreamDevice(false), OutputStreamDevice(false)
 {
@@ -96,7 +100,7 @@ size_t StreamDevice::SeekPosition(size_t curpos, size_t devlen, ssize_t offset,
         case SeekDirection::Begin:
         {
             if (offset < 0)
-                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDeviceOperation, "Invalid negative seek");
+                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Invalid negative seek");
             else if ((size_t)offset > devlen)
                 PODOFO_RAISE_ERROR_INFO(PdfErrorCode::ValueOutOfRange, "Invalid seek out of bounds");
 
@@ -125,7 +129,7 @@ size_t StreamDevice::SeekPosition(size_t curpos, size_t devlen, ssize_t offset,
         case SeekDirection::End:
         {
             if (offset > 0)
-                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDeviceOperation, "Invalid positive seek");
+                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Invalid positive seek");
             else if ((size_t)-offset > devlen)
                 PODOFO_RAISE_ERROR_INFO(PdfErrorCode::ValueOutOfRange, "Invalid seek out of bounds");
 
@@ -142,25 +146,25 @@ StandardStreamDevice::StandardStreamDevice(ostream& stream)
     : StandardStreamDevice(DeviceAccess::Write, &stream, nullptr, &stream, false)
 {
     if (stream.fail())
-        PODOFO_RAISE_ERROR(PdfErrorCode::InvalidDeviceOperation);
+        PODOFO_RAISE_ERROR(PdfErrorCode::IOError);
 }
 
 StandardStreamDevice::StandardStreamDevice(istream& stream)
     : StandardStreamDevice(DeviceAccess::Read, &stream, &stream, nullptr, false)
 {
     if (stream.fail())
-        PODOFO_RAISE_ERROR(PdfErrorCode::InvalidDeviceOperation);
+        PODOFO_RAISE_ERROR(PdfErrorCode::IOError);
 }
 
 StandardStreamDevice::StandardStreamDevice(iostream& stream)
     : StandardStreamDevice(DeviceAccess::ReadWrite, &stream, &stream, &stream, false)
 {
     if (stream.fail())
-        PODOFO_RAISE_ERROR(PdfErrorCode::InvalidDeviceOperation);
+        PODOFO_RAISE_ERROR(PdfErrorCode::IOError);
 
     if (stream.tellp() != stream.tellg())
-        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDeviceOperation,
-            "Unsupported mistmatch between read and read position in stream");
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError,
+            "Unsupported mismatch between read and read position in stream");
 }
 
 StandardStreamDevice::StandardStreamDevice(DeviceAccess access, ios& stream, bool streamOwned)
@@ -190,13 +194,13 @@ size_t StandardStreamDevice::GetLength() const
     switch (GetAccess())
     {
         case DeviceAccess::Read:
+        case DeviceAccess::ReadWrite: // We just take the input stream as the reference
         {
             PODOFO_INVARIANT(m_istream != nullptr);
             ret = getLength(*m_istream);
             break;
         }
         case DeviceAccess::Write:
-        case DeviceAccess::ReadWrite:
         {
             PODOFO_INVARIANT(m_ostream != nullptr);
             ret = getLength(*m_ostream);
@@ -207,7 +211,7 @@ size_t StandardStreamDevice::GetLength() const
     }
 
     if (m_Stream->fail())
-        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDeviceOperation, "Failed to retrieve length for this stream");
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Failed to retrieve length for this stream");
 
     return ret;
 }
@@ -218,13 +222,13 @@ size_t StandardStreamDevice::GetPosition() const
     switch (GetAccess())
     {
         case DeviceAccess::Read:
+        case DeviceAccess::ReadWrite: // We just take the input stream as the reference
         {
             PODOFO_INVARIANT(m_istream != nullptr);
             ret = getPosition(*m_istream);
             break;
         }
         case DeviceAccess::Write:
-        case DeviceAccess::ReadWrite:
         {
             PODOFO_INVARIANT(m_ostream != nullptr);
             ret = getPosition(*m_ostream);
@@ -235,7 +239,7 @@ size_t StandardStreamDevice::GetPosition() const
     }
 
     if (m_Stream->fail())
-        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDeviceOperation, "Failed to get current position in the stream");
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Failed to get current position in the stream");
 
     return ret;
 }
@@ -252,10 +256,44 @@ bool StandardStreamDevice::Eof() const
 
 void StandardStreamDevice::writeBuffer(const char* buffer, size_t size)
 {
-    PODOFO_INVARIANT(m_ostream != nullptr);
-    m_ostream->write(buffer, size);
-    if (m_ostream->fail())
-        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDeviceOperation, "Failed to write the given buffer");
+    switch (GetAccess())
+    {
+        case DeviceAccess::Write:
+        {
+            PODOFO_INVARIANT(m_ostream != nullptr);
+            m_ostream->write(buffer, size);
+            if (m_ostream->fail())
+                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Failed to write the given buffer");
+            break;
+        }
+        case DeviceAccess::ReadWrite:
+        {
+            PODOFO_INVARIANT(m_ostream != nullptr && m_istream != nullptr);
+            // Since some iostreams such as std::stringstream have different position
+            // indicators for input and output sequences. Synchronize the ostream
+            // position indicator with the istream (the reference one) before writing
+            // NOTE: Some c++ libraries don't reset eofbit prior seeking
+            auto pos = getPosition(*m_istream);
+            m_Stream->clear(m_Stream->rdstate() & ~ios_base::eofbit);
+            ::seek(*m_ostream, (ssize_t)pos, SeekDirection::Begin);
+            m_ostream->write(buffer, size);
+            if (m_ostream->fail())
+                goto Fail;
+
+            // After writing, finally synchronize the istream position indicator
+            ::seek(*m_istream, (ssize_t)(pos + size), SeekDirection::Begin);
+            if (m_istream->fail())
+                goto Fail;
+
+            break;
+
+        Fail:
+            PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Failed to write the given buffer");
+            break;
+        }
+        default:
+            PODOFO_RAISE_ERROR(PdfErrorCode::InternalLogic);
+    }
 }
 
 void StandardStreamDevice::flush()
@@ -263,7 +301,7 @@ void StandardStreamDevice::flush()
     PODOFO_INVARIANT(m_ostream != nullptr);
     m_ostream->flush();
     if (m_ostream->fail())
-        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDeviceOperation, "Failed to flush the stream");
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Failed to flush the stream");
 }
 
 size_t StandardStreamDevice::readBuffer(char* buffer, size_t size, bool& eof)
@@ -302,7 +340,7 @@ bool StandardStreamDevice::peek(char& ch) const
 
     read = m_istream->peek();
     if (m_istream->fail())
-        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDeviceOperation, "Failed to peek current character");
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Failed to peek current character");
 
     if (read == char_traits<char>::eof())
         goto Eof;
@@ -319,20 +357,30 @@ void StandardStreamDevice::seek(ssize_t offset, SeekDirection direction)
 {
     // NOTE: Some c++ libraries don't reset eofbit prior seeking
     m_Stream->clear(m_Stream->rdstate() & ~ios_base::eofbit);
-    if ((GetAccess() & DeviceAccess::Read) != DeviceAccess{ })
+    switch (GetAccess())
     {
-        PODOFO_INVARIANT(m_istream != nullptr);
-        ::seek(*m_istream, offset, direction);
-    }
-
-    if ((GetAccess() & DeviceAccess::Write) != DeviceAccess{ })
-    {
-        PODOFO_INVARIANT(m_ostream != nullptr);
-        ::seek(*m_ostream, offset, direction);
+        case DeviceAccess::Read:
+        case DeviceAccess::ReadWrite:
+        {
+            PODOFO_INVARIANT(m_istream != nullptr);
+            ::seek(*m_istream, offset, direction);
+            break;
+        }
+        case DeviceAccess::Write:
+        {
+            PODOFO_INVARIANT(m_ostream != nullptr);
+            ::seek(*m_ostream, offset, direction);
+            break;
+        }
     }
 
     if (m_Stream->fail())
-        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDeviceOperation, "Failed to seek to given position in the stream");
+        goto Fail;
+
+    return;
+
+Fail:
+    PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Failed to seek to given position in the stream");
 }
 
 FileStreamDevice::FileStreamDevice(const string_view& filepath)
@@ -346,108 +394,163 @@ FileStreamDevice::FileStreamDevice(const string_view& filepath, FileMode mode)
 }
 
 FileStreamDevice::FileStreamDevice(const string_view& filepath, FileMode mode, DeviceAccess access)
-    : StandardStreamDevice(access, *getFileStream(filepath, mode, access), true), m_Filepath(filepath)
+    : StreamDevice(access), m_file(createFile(filepath, mode, access))
 {
 }
 
+FileStreamDevice::~FileStreamDevice()
+{
+    try
+    {
+        close();
+    }
+    catch (...)
+    {
+        // Do nothing, it should not throw
+    }
+}
+
+size_t FileStreamDevice::GetLength() const
+{
+    ssize_t offset;
+    ssize_t previousOffset;
+    previousOffset = utls::ftell(m_file);
+    if (previousOffset == -1)
+        goto Fail;
+
+    if (utls::fseek(m_file, 0, SEEK_END) != 0)
+        goto Fail;
+
+    offset = utls::ftell(m_file);
+    if (offset == -1)
+        goto Fail;
+
+    if (utls::fseek(m_file, previousOffset, SEEK_SET) != 0)
+        goto Fail;
+
+    return offset;
+
+Fail:
+    PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Failed to determine the current file length");
+}
+
+size_t FileStreamDevice::GetPosition() const
+{
+    ssize_t offset = utls::ftell(m_file);
+    if (offset == -1)
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Failed to read the current file position");
+
+    return offset;
+}
+
+bool FileStreamDevice::CanSeek() const
+{
+    return true;
+}
+
+bool FileStreamDevice::Eof() const
+{
+    return std::feof(m_file) != 0;
+}
+
+void FileStreamDevice::writeBuffer(const char* buffer, size_t size)
+{
+    if (std::fwrite(buffer, sizeof(char), size, m_file) != size)
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Failed to write the given buffer");
+}
+
+void FileStreamDevice::flush()
+{
+    int rc = std::fflush(m_file);
+    if (rc == EOF)
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Failed to flush the stream");
+}
+
+size_t FileStreamDevice::readBuffer(char* buffer, size_t size, bool& eof)
+{
+    size_t ret = std::fread(buffer, 1, (size_t)size, m_file);
+    if (std::ferror(m_file) != 0)
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Failed to read the amount of bytes requested");
+
+    if (std::feof(m_file) != 0)
+    {
+        eof = true;
+        return ret;
+    }
+
+    eof = false;
+    return ret;
+}
+
+bool FileStreamDevice::readChar(char& ch)
+{
+    int rc = std::fgetc(m_file);
+    if (std::ferror(m_file) != 0)
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Stream I/O error while reading");
+
+    if (std::feof(m_file) != 0)
+    {
+        ch = '\0';
+        return false;
+    }
+
+    ch = (char)(unsigned char)rc;
+    return true;
+}
+
+bool FileStreamDevice::peek(char& ch) const
+{
+    int rc = std::fgetc(m_file);
+    if (std::ferror(m_file) != 0)
+        goto Fail;
+
+    if (std::feof(m_file) != 0)
+    {
+        ch = '\0';
+        return false;
+    }
+
+    if (std::ungetc(rc, m_file) == EOF)
+        goto Fail;
+
+    ch = (char)(unsigned char)rc;
+    return true;
+
+Fail:
+    PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Stream I/O error while reading");
+}
+
+void FileStreamDevice::seek(ssize_t offset, SeekDirection direction)
+{
+    int origin;
+    switch (direction)
+    {
+        case SeekDirection::Begin:
+            origin = SEEK_SET;
+            break;
+        case SeekDirection::Current:
+            origin = SEEK_CUR;
+            break;
+        case SeekDirection::End:
+            origin = SEEK_END;
+            break;
+        default:
+            PODOFO_RAISE_ERROR(PdfErrorCode::InvalidEnumValue);
+    }
+
+    if (utls::fseek(m_file, offset, origin) != 0)
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Failed to seek to given position in the stream");
+}
+
 void FileStreamDevice::close()
 {
-    dynamic_cast<fstream&>(GetStream()).close();
-}
+    if (m_file == nullptr)
+        return;
 
-fstream* FileStreamDevice::getFileStream(const string_view& filename, FileMode mode, DeviceAccess access)
-{
-    switch (mode)
-    {
-        case FileMode::CreateNew:
-            if (access == DeviceAccess::Read)
-                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDeviceOperation, "Invalid combination FileMode::CreateNew and DeviceAccess::Read");
-            break;
-        case FileMode::Create:
-            if (access == DeviceAccess::Read)
-                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDeviceOperation, "Invalid combination FileMode::Create and DeviceAccess::Read");
-            break;
-        case FileMode::OpenOrCreate:
-            if (access == DeviceAccess::Read)
-                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDeviceOperation, "Invalid combination FileMode::OpenOrCreate and DeviceAccess::Read");
-            break;
-        case FileMode::Truncate:
-            if (access == DeviceAccess::Read)
-                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDeviceOperation, "Invalid combination FileMode::Truncate and DeviceAccess::Read");
-            break;
-        case FileMode::Append:
-            if ((access & DeviceAccess::Read) != DeviceAccess{ })
-                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDeviceOperation, "Invalid combination FileMode::Append and DeviceAccess::Read or DeviceAccess::ReadWrite");
-            break;
-        default:
-            break;
-    }
+    if (std::fclose(m_file) == EOF)
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Failed to close stream");
 
-    ios_base::openmode openmode = fstream::binary;
-    switch (access)
-    {
-        case DeviceAccess::Read:
-            openmode |= ios_base::in;
-            break;
-        case DeviceAccess::Write:
-            openmode |= ios_base::out;
-            break;
-        case DeviceAccess::ReadWrite:
-            openmode |= ios_base::in | ios_base::out;
-            break;
-        default:
-            PODOFO_RAISE_ERROR(PdfErrorCode::InvalidEnumValue);
-    }
-
-    switch (mode)
-    {
-        case FileMode::CreateNew:
-            if (fs::exists(filename))
-                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDeviceOperation, "The file must not exist");
-            break;
-        case FileMode::Create:
-            openmode |= ios_base::trunc;
-            break;
-        case FileMode::Open:
-            if ((access & DeviceAccess::Write) != DeviceAccess{ } && !fs::exists(filename))
-                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDeviceOperation, "The file must exist");
-            break;
-        case FileMode::OpenOrCreate:
-            break;
-        case FileMode::Truncate:
-            if (!fs::exists(filename))
-                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDeviceOperation, "The file must exist");
-            openmode |= ios_base::trunc;
-            break;
-        case FileMode::Append:
-            // NOTE: ios_base::app and ios_base::ate don't work as intended,
-            // as tellp returns 0. Also opening the file with ios_base::out
-            // will truncate the file anyway, even without ios_base::trunc.
-            // So, we have to create the file read/write and do our seeking
-            openmode |= ios_base::in;
-            break;
-        default:
-            PODOFO_RAISE_ERROR(PdfErrorCode::InvalidEnumValue);
-    }
-
-    auto stream = new fstream(utls::open_fstream(filename, openmode));
-    if (mode == FileMode::Append)
-    {
-        // NOTE: Do our manual seeking at the end
-        stream->seekg(0, ios_base::end);
-        stream->seekp(0, ios_base::end);
-
-        // TODO: We should prevent seeking before the initial
-        // position
-    }
-
-    if (stream->fail())
-    {
-        delete stream;
-        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDeviceOperation, "Error accessing file {}", filename);
-    }
-
-    return stream;
+    m_file = nullptr;
 }
 
 NullStreamDevice::NullStreamDevice()
@@ -615,3 +718,143 @@ void SpanStreamDevice::seek(ssize_t offset, SeekDirection direction)
 {
     m_Position = SeekPosition(m_Position, m_Length, offset, direction);
 }
+
+FILE* createFile(const string_view& filepath, FileMode mode, DeviceAccess access)
+{
+    string cmode;
+    switch (mode)
+    {
+        case FileMode::CreateNew:
+        {
+            if (access == DeviceAccess::Read)
+                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Invalid combination FileMode::CreateNew and DeviceAccess::Read");
+
+            if (fs::exists(fs::u8path(filepath)))
+                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "The file {} must not exist", filepath);
+
+            switch (access)
+            {
+                case DeviceAccess::Write:
+                    cmode.append("w");
+                    break;
+                case DeviceAccess::ReadWrite:
+                    cmode.append("w+");
+                    break;
+                default:
+                    PODOFO_RAISE_ERROR(PdfErrorCode::InvalidEnumValue);
+            }
+
+            break;
+        }
+        case FileMode::Create:
+        {
+            if (access == DeviceAccess::Read)
+                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Invalid combination FileMode::Create and DeviceAccess::Read");
+
+            switch (access)
+            {
+                case DeviceAccess::Write:
+                    cmode.append("w");
+                    break;
+                case DeviceAccess::ReadWrite:
+                    cmode.append("w+");
+                    break;
+                default:
+                    PODOFO_RAISE_ERROR(PdfErrorCode::InvalidEnumValue);
+            }
+
+            break;
+        }
+        case FileMode::Open:
+        {
+            if ((access & DeviceAccess::Write) != DeviceAccess{ } && !fs::exists(fs::u8path(filepath)))
+                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "The file {} must exist", filepath);
+
+            switch (access)
+            {
+                case DeviceAccess::Read:
+                    cmode.append("r");
+                    break;
+                case DeviceAccess::Write:
+                    cmode.append("w");
+                    break;
+                case DeviceAccess::ReadWrite:
+                    cmode.append("r+");
+                    break;
+                default:
+                    PODOFO_RAISE_ERROR(PdfErrorCode::InvalidEnumValue);
+            }
+
+            break;
+        }
+        case FileMode::OpenOrCreate:
+        {
+            if (access == DeviceAccess::Read)
+                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Invalid combination FileMode::OpenOrCreate and DeviceAccess::Read");
+
+            bool exists = fs::exists(fs::u8path(filepath));
+            switch (access)
+            {
+                case DeviceAccess::Write:
+                    cmode.append("w");
+                    break;
+                case DeviceAccess::ReadWrite:
+                    cmode.append(exists ? "r+" : "w+");
+                    break;
+                default:
+                    PODOFO_RAISE_ERROR(PdfErrorCode::InvalidEnumValue);
+            }
+
+            break;
+        }
+        case FileMode::Truncate:
+        {
+            if (access == DeviceAccess::Read)
+                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Invalid combination FileMode::Truncate and DeviceAccess::Read");
+
+            if (!fs::exists(fs::u8path(filepath)))
+                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "The file {} must exist", filepath);
+
+            switch (access)
+            {
+                case DeviceAccess::Write:
+                    cmode.append("w");
+                    break;
+                case DeviceAccess::ReadWrite:
+                    cmode.append("w+");
+                    break;
+                default:
+                    PODOFO_RAISE_ERROR(PdfErrorCode::InvalidEnumValue);
+            }
+            break;
+        }
+        case FileMode::Append:
+            if (access == DeviceAccess::Read)
+                PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Invalid combination FileMode::Append and DeviceAccess::Read");
+
+            switch (access)
+            {
+                case DeviceAccess::Write:
+                    cmode.append("a");
+                    break;
+                case DeviceAccess::ReadWrite:
+                    cmode.append("a+");
+                    break;
+                default:
+                    PODOFO_RAISE_ERROR(PdfErrorCode::InvalidEnumValue);
+            }
+
+            break;
+        default:
+            PODOFO_RAISE_ERROR(PdfErrorCode::InvalidEnumValue);
+    }
+#if _WIN32
+    cmode.append("b");
+#endif // _WIN32
+
+    auto stream = utls::fopen(filepath, cmode.data());
+    if (stream == nullptr)
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::IOError, "Error accessing file {}", filepath);
+
+    return stream;
+}
