@@ -8,25 +8,9 @@
 #include <podofo/private/XMPUtils.h>
 #include "PdfDocument.h"
 
-#include <algorithm>
-#include <deque>
-
-#include "PdfArray.h"
-#include "PdfDictionary.h"
-#include "PdfImmediateWriter.h"
-#include "PdfObjectStream.h"
-#include "PdfIndirectObjectList.h"
-#include "PdfAcroForm.h"
+#include "PdfExtGState.h"
 #include "PdfDestination.h"
 #include "PdfFileSpec.h"
-#include "PdfFontMetrics.h"
-#include "PdfInfo.h"
-#include "PdfNameTree.h"
-#include "PdfOutlines.h"
-#include "PdfPage.h"
-#include "PdfPageCollection.h"
-#include "PdfXObjectForm.h"
-#include "PdfImage.h"
 
 using namespace std;
 using namespace PoDoFo;
@@ -37,22 +21,7 @@ PdfDocument::PdfDocument(bool empty) :
     m_FontManager(*this)
 {
     if (!empty)
-    {
-        m_TrailerObj.reset(new PdfObject()); // The trailer is NO part of the vector of objects
-        m_TrailerObj->SetDocument(this);
-        auto& catalog = m_Objects.CreateDictionaryObject("Catalog");
-        m_Trailer.reset(new PdfTrailer(*m_TrailerObj));
-
-        m_Catalog.reset(new PdfCatalog(catalog));
-        m_TrailerObj->GetDictionary().AddKeyIndirect("Root", catalog);
-
-        auto& info = m_Objects.CreateDictionaryObject();
-        m_Info.reset(new PdfInfo(info,
-            PdfInfoInitial::WriteProducer | PdfInfoInitial::WriteCreationTime));
-        m_TrailerObj->GetDictionary().AddKeyIndirect("Info", info);
-
-        Init();
-    }
+        resetPrivate();
 }
 
 PdfDocument::PdfDocument(const PdfDocument& doc) :
@@ -66,20 +35,40 @@ PdfDocument::PdfDocument(const PdfDocument& doc) :
 
 PdfDocument::~PdfDocument()
 {
-    // Do nothing, all members will autoclear
+    // NOTE: Members will autoclear
 }
 
-void PdfDocument::Clear() 
+void PdfDocument::Reset()
+{
+    Clear();
+    resetPrivate();
+    reset();
+}
+
+void PdfDocument::reset()
+{
+    // Do nothing, to be overridden
+}
+
+void PdfDocument::Clear()
 {
     m_FontManager.Clear();
+    m_Metadata.Invalidate();
+    m_TrailerObj = nullptr;
+    m_Trailer = nullptr;
     m_Catalog = nullptr;
     m_Info = nullptr;
     m_Pages = nullptr;
     m_AcroForm = nullptr;
     m_Outlines = nullptr;
-    m_NameTree = nullptr;
+    m_NameTrees = nullptr;
     m_Objects.Clear();
-    m_Objects.SetCanReuseObjectNumbers(true);
+    clear();
+}
+
+void PdfDocument::clear()
+{
+    // Do nothing, to be overridden
 }
 
 void PdfDocument::Init()
@@ -88,7 +77,7 @@ void PdfDocument::Init()
     if (pagesRootObj == nullptr)
     {
         m_Pages.reset(new PdfPageCollection(*this));
-        m_Catalog->GetDictionary().AddKey("Pages", m_Pages->GetObject().GetIndirectReference());
+        m_Catalog->GetDictionary().AddKey("Pages"_n, m_Pages->GetObject().GetIndirectReference());
     }
     else
     {
@@ -98,11 +87,7 @@ void PdfDocument::Init()
     auto& catalogDict = m_Catalog->GetDictionary();
     auto namesObj = catalogDict.FindKey("Names");
     if (namesObj != nullptr)
-        m_NameTree.reset(new PdfNameTree(*namesObj));
-
-    auto outlinesObj = catalogDict.FindKey("Outlines");
-    if (outlinesObj != nullptr)
-        m_Outlines.reset(new PdfOutlines(*outlinesObj));
+        m_NameTrees.reset(new PdfNameTrees(*namesObj));
 
     auto acroformObj = catalogDict.FindKey("AcroForm");
     if (acroformObj != nullptr)
@@ -116,12 +101,12 @@ void PdfDocument::AppendDocumentPages(const PdfDocument& doc)
 
 void PdfDocument::append(const PdfDocument& doc, bool appendAll)
 {
-    unsigned difference = static_cast<unsigned>(m_Objects.GetSize() + m_Objects.GetFreeObjects().size());
-
-    // Because GetNextObject uses m_ObjectCount instead 
-    // of m_Objects.GetSize() + m_Objects.GetFreeObjects().size() + 1
-    // make sure the free objects are already present before appending to
-    // prevent overlapping obj-numbers
+    // CHECK-ME: The following is fishy. We switched from m_Objects.GetSize() to m_Objects.GetObjectCount()
+    // to not fall in overlaps in case of removed objects (see https://github.com/podofo/podofo/issues/253),
+    // but in this way free objects should be already taken into account. We'll eventually fix it by not
+    // relying on computing a static difference between inserted objects and objects being inserted but just
+    // inserting objects normally and remapping them with a support map
+    unsigned difference = static_cast<unsigned>(m_Objects.GetObjectCount() + m_Objects.GetFreeObjects().size());
 
     // create all free objects again, to have a clean free object list
     for (auto& ref : doc.GetObjects().GetFreeObjects())
@@ -137,7 +122,7 @@ void PdfDocument::append(const PdfDocument& doc, bool appendAll)
         m_Objects.PushObject(newObj);
         *newObj = *obj;
 
-        PoDoFo::LogMessage(PdfLogSeverity::Information, "Fixing references in {} {} R by {}",
+        PoDoFo::LogMessage(PdfLogSeverity::Debug, "Fixing references in {} {} R by {}",
             newObj->GetIndirectReference().ObjectNumber(), newObj->GetIndirectReference().GenerationNumber(), difference);
         fixObjectReferences(*newObj, difference);
     }
@@ -145,11 +130,11 @@ void PdfDocument::append(const PdfDocument& doc, bool appendAll)
     if (appendAll)
     {
         const PdfName inheritableAttributes[] = {
-            PdfName("Resources"),
-            PdfName("MediaBox"),
-            PdfName("CropBox"),
-            PdfName("Rotate"),
-            PdfName::KeyNull
+            "Resources"_n,
+            "MediaBox"_n,
+            "CropBox"_n,
+            "Rotate"_n,
+            PdfName::Null
         };
 
         // append all pages now to our page tree
@@ -179,18 +164,20 @@ void PdfDocument::append(const PdfDocument& doc, bool appendAll)
             m_Pages->InsertPageAt(m_Pages->GetCount(), *new PdfPage(obj));
         }
 
-        // append all outlines
-        PdfOutlineItem* root = this->GetOutlines();
-        PdfOutlines* appendRoot = const_cast<PdfDocument&>(doc).GetOutlines();
-        if (appendRoot && appendRoot->First())
+        // Append all outlines
+        const PdfOutlineItem* appendRoot = doc.GetOutlines();
+        if (appendRoot != nullptr && (appendRoot = appendRoot->First()) != nullptr)
         {
-            // only append outlines if appended document has outlines
-            while (root && root->Next())
+            // Get or create outlines
+            PdfOutlineItem* root = &this->GetOrCreateOutlines();
+
+            // Find actual item where to append
+            while (root->Next() != nullptr)
                 root = root->Next();
 
-            PdfReference ref(appendRoot->First()->GetObject().GetIndirectReference().ObjectNumber()
-                + difference, appendRoot->First()->GetObject().GetIndirectReference().GenerationNumber());
-            root->InsertChild(new PdfOutlines(m_Objects.MustGetObject(ref)));
+            PdfReference ref(appendRoot->GetObject().GetIndirectReference().ObjectNumber()
+                + difference, appendRoot->GetObject().GetIndirectReference().GenerationNumber());
+            root->InsertChild(unique_ptr<PdfOutlineItem>(new PdfOutlines(m_Objects.MustGetObject(ref))));
         }
     }
 
@@ -200,23 +187,21 @@ void PdfDocument::append(const PdfDocument& doc, bool appendAll)
 
 void PdfDocument::InsertDocumentPageAt(unsigned atIndex, const PdfDocument& doc, unsigned pageIndex)
 {
-    // copy of PdfDocument::Append, only restricts which page to add
-    unsigned difference = static_cast<unsigned>(m_Objects.GetSize() + m_Objects.GetFreeObjects().size());
-
-
-    // Because GetNextObject uses m_ObjectCount instead 
-    // of m_Objects.GetSize() + m_Objects.GetFreeObjects().size() + 1
-    // make sure the free objects are already present before appending to
-    // prevent overlapping obj-numbers
+    // CHECK-ME: The following is fishy. We switched from m_Objects.GetSize() to m_Objects.GetObjectCount()
+    // to not fall in overlaps in case of removed objects (see https://github.com/podofo/podofo/issues/253),
+    // but in this way free objects should be already taken into account. We'll eventually fix it by not
+    // relying on computing a static difference between inserted objects and objects being inserted but just
+    // inserting objects normally and remapping them with a support map
+    unsigned difference = static_cast<unsigned>(m_Objects.GetObjectCount() + m_Objects.GetFreeObjects().size());
 
     // create all free objects again, to have a clean free object list
-    for (auto& freeObj : GetObjects().GetFreeObjects())
+    for (auto& freeObj : doc.GetObjects().GetFreeObjects())
     {
         m_Objects.AddFreeObject(PdfReference(freeObj.ObjectNumber() + difference, freeObj.GenerationNumber()));
     }
 
     // append all objects first and fix their references
-    for (auto& obj : GetObjects())
+    for (auto& obj : doc.GetObjects())
     {
         PdfReference ref(static_cast<uint32_t>(obj->GetIndirectReference().ObjectNumber() + difference), obj->GetIndirectReference().GenerationNumber());
         auto newObj = new PdfObject(PdfDictionary());
@@ -225,62 +210,42 @@ void PdfDocument::InsertDocumentPageAt(unsigned atIndex, const PdfDocument& doc,
         m_Objects.PushObject(newObj);
         *newObj = *obj;
 
-        PoDoFo::LogMessage(PdfLogSeverity::Information, "Fixing references in {} {} R by {}",
+        PoDoFo::LogMessage(PdfLogSeverity::Debug, "Fixing references in {} {} R by {}",
             newObj->GetIndirectReference().ObjectNumber(), newObj->GetIndirectReference().GenerationNumber(), difference);
         fixObjectReferences(*newObj, difference);
     }
 
     const PdfName inheritableAttributes[] = {
-        PdfName("Resources"),
-        PdfName("MediaBox"),
-        PdfName("CropBox"),
-        PdfName("Rotate"),
-        PdfName::KeyNull
+        "Resources"_n,
+        "MediaBox"_n,
+        "CropBox"_n,
+        "Rotate"_n,
+        PdfName::Null
     };
 
-    // append all pages now to our page tree
-    for (unsigned i = 0; i < doc.GetPages().GetCount(); i++)
+    // append all page to our page tree
+    auto& page = doc.GetPages().GetPageAt(pageIndex);
+    auto& obj = m_Objects.MustGetObject(PdfReference(page.GetObject().GetIndirectReference().ObjectNumber()
+                                                     + difference, page.GetObject().GetIndirectReference().GenerationNumber()));
+    if (obj.IsDictionary() && obj.GetDictionary().HasKey("Parent"))
+        obj.GetDictionary().RemoveKey("Parent");
+
+    // Deal with inherited attributes
+    const PdfName* inherited = inheritableAttributes;
+    while (!inherited->IsNull())
     {
-        if (i != pageIndex)
-            continue;
-
-        auto& page = doc.GetPages().GetPageAt(i);
-        auto& obj = m_Objects.MustGetObject(PdfReference(page.GetObject().GetIndirectReference().ObjectNumber()
-            + difference, page.GetObject().GetIndirectReference().GenerationNumber()));
-        if (obj.IsDictionary() && obj.GetDictionary().HasKey("Parent"))
-            obj.GetDictionary().RemoveKey("Parent");
-
-        // Deal with inherited attributes
-        const PdfName* inherited = inheritableAttributes;
-        while (!inherited->IsNull())
+        auto attribute = page.GetDictionary().FindKeyParent(*inherited);
+        if (attribute != nullptr)
         {
-            auto attribute = page.GetDictionary().FindKeyParent(*inherited);
-            if (attribute != nullptr)
-            {
-                PdfObject attributeCopy(*attribute);
-                fixObjectReferences(attributeCopy, difference);
-                obj.GetDictionary().AddKey(*inherited, attributeCopy);
-            }
-
-            inherited++;
+            PdfObject attributeCopy(*attribute);
+            fixObjectReferences(attributeCopy, difference);
+            obj.GetDictionary().AddKey(*inherited, attributeCopy);
         }
 
-        m_Pages->InsertPageAt(atIndex, *new PdfPage(obj));
+        inherited++;
     }
 
-    // append all outlines
-    PdfOutlineItem* root = this->GetOutlines();
-    PdfOutlines* appendRoot = const_cast<PdfDocument&>(doc).GetOutlines();
-    if (appendRoot != nullptr && appendRoot->First())
-    {
-        // only append outlines if appended document has outlines
-        while (root != nullptr && root->Next())
-            root = root->Next();
-
-        PdfReference ref(appendRoot->First()->GetObject().GetIndirectReference().ObjectNumber()
-            + difference, appendRoot->First()->GetObject().GetIndirectReference().GenerationNumber());
-        root->InsertChild(new PdfOutlines(m_Objects.MustGetObject(ref)));
-    }
+    m_Pages->InsertPageAt(atIndex, *new PdfPage(obj));
 
     // TODO: merge name trees
     // ToDictionary -> then iteratate over all keys and add them to the new one
@@ -332,25 +297,65 @@ void PdfDocument::deletePages(unsigned atIndex, unsigned pageCount)
         this->GetPages().RemovePageAt(atIndex);
 }
 
+void PdfDocument::resetPrivate()
+{
+    m_TrailerObj.reset(new PdfObject()); // The trailer is NO part of the vector of objects
+    m_TrailerObj->SetDocument(this);
+    auto& catalog = m_Objects.CreateDictionaryObject("Catalog"_n);
+    m_Trailer.reset(new PdfTrailer(*m_TrailerObj));
+
+    m_Catalog.reset(new PdfCatalog(catalog));
+    m_TrailerObj->GetDictionary().AddKeyIndirect("Root"_n, catalog);
+
+    auto& info = m_Objects.CreateDictionaryObject();
+    m_Info.reset(new PdfInfo(info,
+        PdfInfoInitial::WriteProducer | PdfInfoInitial::WriteCreationTime));
+    m_TrailerObj->GetDictionary().AddKeyIndirect("Info"_n, info);
+
+    Init();
+}
+
+void PdfDocument::initOutlines()
+{
+    if (m_Outlines != nullptr)
+        return;
+
+    auto outlinesObj = m_Catalog->GetDictionary().FindKey("Outlines");
+    if (outlinesObj == nullptr)
+        m_Outlines *= nullptr;
+    else
+        m_Outlines = unique_ptr<PdfOutlines>(new PdfOutlines(*outlinesObj));
+}
+
 PdfInfo& PdfDocument::GetOrCreateInfo()
 {
     if (m_Info == nullptr)
     {
         auto info = &m_Objects.CreateDictionaryObject();
         m_Info.reset(new PdfInfo(*info));
-        m_TrailerObj->GetDictionary().AddKeyIndirect("Info", *info);
+        m_TrailerObj->GetDictionary().AddKeyIndirect("Info"_n, *info);
     }
 
     return *m_Info;
 }
 
+void PdfDocument::createAction(PdfActionType type, unique_ptr<PdfAction>& action)
+{
+    action = PdfAction::Create(*this, type);
+}
+
 Rect PdfDocument::FillXObjectFromPage(PdfXObjectForm& xobj, const PdfPage& page, bool useTrimBox)
 {
     unsigned difference = 0;
     auto& sourceDoc = page.GetDocument();
     if (this != &sourceDoc)
     {
-        difference = static_cast<unsigned>(m_Objects.GetSize() + m_Objects.GetFreeObjects().size());
+        // CHECK-ME: The following is fishy. We switched from m_Objects.GetSize() to m_Objects.GetObjectCount()
+        // to not fall in overlaps in case of removed objects (see https://github.com/podofo/podofo/issues/253),
+        // but in this way free objects should be already taken into account. We'll eventually fix it by not
+        // relying on computing a static difference between inserted objects and objects being inserted but just
+        // inserting objects normally and remapping them with a support map
+        difference = static_cast<unsigned>(m_Objects.GetObjectCount() + m_Objects.GetFreeObjects().size());
         append(sourceDoc, false);
     }
 
@@ -369,7 +374,7 @@ Rect PdfDocument::FillXObjectFromPage(PdfXObjectForm& xobj, const PdfPage& page,
 
     // link resources from external doc to x-object
     if (pageObj.IsDictionary() && pageObj.GetDictionary().HasKey("Resources"))
-        xobj.GetObject().GetDictionary().AddKey("Resources", *pageObj.GetDictionary().GetKey("Resources"));
+        xobj.GetDictionary().AddKey("Resources"_n, *pageObj.GetDictionary().GetKey("Resources"));
 
     // copy top-level content from external doc to x-object
     if (pageObj.IsDictionary() && pageObj.GetDictionary().HasKey("Contents"))
@@ -488,24 +493,25 @@ void PdfDocument::CollectGarbage()
 
 PdfOutlines& PdfDocument::GetOrCreateOutlines()
 {
-    if (m_Outlines != nullptr)
-        return *m_Outlines.get();
+    initOutlines();
+    if (*m_Outlines != nullptr)
+        return **m_Outlines;
 
-    m_Outlines.reset(new PdfOutlines(*this));
-    m_Catalog->GetDictionary().AddKey("Outlines", m_Outlines->GetObject().GetIndirectReference());
-    return *m_Outlines.get();
+    m_Outlines = unique_ptr<PdfOutlines>(new PdfOutlines(*this));
+    m_Catalog->GetDictionary().AddKey("Outlines"_n, (*m_Outlines)->GetObject().GetIndirectReference());
+    return **m_Outlines;
 }
 
-PdfNameTree& PdfDocument::GetOrCreateNameTree()
+PdfNameTrees& PdfDocument::GetOrCreateNames()
 {
-    if (m_NameTree != nullptr)
-        return *m_NameTree;
+    if (m_NameTrees != nullptr)
+        return *m_NameTrees;
 
-    PdfNameTree tmpTree(*this);
+    PdfNameTrees tmpTree(*this);
     auto obj = &tmpTree.GetObject();
-    m_Catalog->GetDictionary().AddKey("Names", obj->GetIndirectReference());
-    m_NameTree.reset(new PdfNameTree(*obj));
-    return *m_NameTree;
+    m_Catalog->GetDictionary().AddKey("Names"_n, obj->GetIndirectReference());
+    m_NameTrees.reset(new PdfNameTrees(*obj));
+    return *m_NameTrees;
 }
 
 PdfAcroForm& PdfDocument::GetOrCreateAcroForm(PdfAcroFormDefaulAppearance defaultAppearance)
@@ -514,34 +520,10 @@ PdfAcroForm& PdfDocument::GetOrCreateAcroForm(PdfAcroFormDefaulAppearance defaul
         return *m_AcroForm.get();
 
     m_AcroForm.reset(new PdfAcroForm(*this, defaultAppearance));
-    m_Catalog->GetDictionary().AddKey("AcroForm", m_AcroForm->GetObject().GetIndirectReference());
+    m_Catalog->GetDictionary().AddKey("AcroForm"_n, m_AcroForm->GetObject().GetIndirectReference());
     return *m_AcroForm.get();
 }
 
-void PdfDocument::AddNamedDestination(const PdfDestination& dest, const PdfString& name)
-{
-    auto& names = GetOrCreateNameTree();
-    names.AddValue("Dests", name, dest.GetObject().GetIndirectReference());
-}
-
-void PdfDocument::AttachFile(const PdfFileSpec& fileSpec)
-{
-    auto& names = GetOrCreateNameTree();
-    names.AddValue("EmbeddedFiles", fileSpec.GetFilename(false), fileSpec.GetObject().GetIndirectReference());
-}
-    
-PdfFileSpec* PdfDocument::GetAttachment(const PdfString& name)
-{
-    if (m_NameTree == nullptr)
-        return nullptr;
-
-    auto obj = m_NameTree->GetValue("EmbeddedFiles", name);
-    if (obj == nullptr)
-        return nullptr;
-
-    return new PdfFileSpec(*obj);
-}
-
 void PdfDocument::SetTrailer(unique_ptr<PdfObject> obj)
 {
     if (obj == nullptr)
@@ -553,7 +535,7 @@ void PdfDocument::SetTrailer(unique_ptr<PdfObject> obj)
 
     auto catalog = m_TrailerObj->GetDictionary().FindKey("Root");
     if (catalog == nullptr)
-        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::NoObject, "Catalog object not found!");
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::ObjectNotFound, "Catalog object not found!");
 
     m_Catalog.reset(new PdfCatalog(*catalog));
 
@@ -607,12 +589,227 @@ bool PdfDocument::IsHighPrintAllowed() const
     return GetEncrypt() == nullptr ? true : GetEncrypt()->IsHighPrintAllowed();
 }
 
-unique_ptr<PdfImage> PdfDocument::CreateImage(const string_view& prefix)
+void PdfDocument::PushPdfExtension(const PdfExtension& extension)
 {
-    return unique_ptr<PdfImage>(new PdfImage(*this, prefix));
+    PdfDictionary* extensionDict;
+    if (!this->GetCatalog().GetDictionary().TryFindKeyAs("Extensions", extensionDict))
+    {
+        auto& obj = GetObjects().CreateDictionaryObject();
+        this->GetCatalog().GetDictionary().AddKeyIndirect("Extensions"_n, obj);
+        extensionDict = &obj.GetDictionary();
+    }
+
+    auto& newExtension = GetObjects().CreateDictionaryObject();
+    auto& newExtensionDict = newExtension.GetDictionary();
+    auto version = extension.GetBaseVersion();
+    if (version == PdfVersion::Unknown)
+        version = GetPdfVersion();
+
+    newExtensionDict.AddKey("BaseVersion"_n, PoDoFo::GetPdfVersionName(version));
+    newExtensionDict.AddKey("ExtensionLevel"_n, extension.GetLevel());
+    if (extension.GetUrl() != nullptr)
+        newExtensionDict.AddKey("URL"_n, *extension.GetUrl());
+    if (extension.GetExtensionRevision() != nullptr)
+        newExtensionDict.AddKey("ExtensionRevision"_n, *extension.GetExtensionRevision());
+
+    extensionDict->AddKeyIndirect(extension.GetNamespace(), newExtension);
 }
 
-unique_ptr<PdfXObjectForm> PdfDocument::CreateXObjectForm(const Rect& rect, const string_view& prefix)
+bool PdfDocument::HasPdfExtension(const string_view& ns, int64_t level) const
 {
-    return unique_ptr<PdfXObjectForm>(new PdfXObjectForm(*this, rect, prefix));
+    const PdfDictionary* dict;
+    int64_t num;
+    if (this->GetCatalog().GetDictionary().TryFindKeyAs("Extensions", dict)
+        && dict->TryFindKeyAs(ns, dict)
+        && dict->TryFindKeyAs("ExtensionLevel", num)
+        && num == level)
+    {
+        return true;
+    }
+
+    return false;
+}
+
+vector<PdfExtension> PdfDocument::GetPdfExtensions() const
+{
+    vector<PdfExtension> ret;
+    const PdfDictionary* dict;
+    if (!this->GetCatalog().GetDictionary().TryFindKeyAs("Extensions", dict))
+        return ret;
+
+    // Loop through all declared extensions
+    const PdfName* name;
+    int64_t num;
+    PdfVersion version;
+    const PdfString* url;
+    const PdfString* extensionRevision;
+    for (auto& pair : dict->GetIndirectIterator())
+    {
+        if (!pair.second->TryGetDictionary(dict)
+            || !dict->TryFindKeyAs("ExtensionLevel", num)
+            || !dict->TryFindKeyAs("BaseVersion", name)
+            || (version = PoDoFo::GetPdfVersion(name->GetString())) == PdfVersion::Unknown)
+        {
+            continue;
+        }
+
+        (void)dict->TryFindKeyAs("URL", url);
+        (void)dict->TryFindKeyAs("ExtensionRevision", extensionRevision);
+
+        ret.push_back(PdfExtension(pair.first, num, version,
+            url == nullptr ? nullable<const PdfString&>{ } : *url,
+            extensionRevision == nullptr ? nullable<const PdfString&>{ } : *extensionRevision));
+    }
+
+    return ret;
+}
+
+void PdfDocument::RemovePdfExtension(const string_view& ns, int64_t level)
+{
+    PdfDictionary* dict;
+    int64_t num;
+    if (!this->GetCatalog().GetDictionary().TryFindKeyAs("Extensions", dict)
+        || !dict->TryFindKeyAs(ns, dict)
+        || !dict->TryFindKeyAs("ExtensionLevel", num)
+        || num != level)
+    {
+        return;
+    }
+
+    dict->RemoveKey(ns);
+}
+
+PdfAcroForm& PdfDocument::MustGetAcroForm()
+{
+    if (m_AcroForm == nullptr)
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidHandle, "AcroForm is not present");
+
+    return *m_AcroForm;
+}
+
+const PdfAcroForm& PdfDocument::MustGetAcroForm() const
+{
+    if (m_AcroForm == nullptr)
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidHandle, "AcroForm is not present");
+
+    return *m_AcroForm;
+}
+
+PdfNameTrees& PdfDocument::MustGetNames()
+{
+    if (m_NameTrees == nullptr)
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidHandle, "Names are not present");
+
+    return *m_NameTrees;
+}
+
+const PdfNameTrees& PdfDocument::MustGetNames() const
+{
+    if (m_NameTrees == nullptr)
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidHandle, "Names are not present");
+
+    return *m_NameTrees;
+}
+
+PdfOutlines* PdfDocument::GetOutlines()
+{
+    initOutlines();
+    return (*m_Outlines).get();
+}
+
+const PdfOutlines* PdfDocument::GetOutlines() const
+{
+    const_cast<PdfDocument&>(*this).initOutlines();
+    return (*m_Outlines).get();
+}
+
+PdfOutlines& PdfDocument::MustGetOutlines()
+{
+    initOutlines();
+    if (*m_Outlines == nullptr)
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidHandle, "Outlines are not present");
+
+    return **m_Outlines;
+}
+
+const PdfOutlines& PdfDocument::MustGetOutlines() const
+{
+    const_cast<PdfDocument&>(*this).initOutlines();
+    if (*m_Outlines == nullptr)
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidHandle, "Outlines are not present");
+
+    return **m_Outlines;
+}
+
+PdfDocumentFieldIterable PdfDocument::GetFieldsIterator()
+{
+    return PdfDocumentFieldIterable(*this);
+}
+
+PdfDocumentConstFieldIterable PdfDocument::GetFieldsIterator() const
+{
+    return PdfDocumentConstFieldIterable(const_cast<PdfDocument&>(*this));
+}
+
+unique_ptr<PdfImage> PdfDocument::CreateImage()
+{
+    return unique_ptr<PdfImage>(new PdfImage(*this));
+}
+
+unique_ptr<PdfXObjectForm> PdfDocument::CreateXObjectForm(const Rect& rect)
+{
+    return unique_ptr<PdfXObjectForm>(new PdfXObjectForm(*this, rect));
+}
+
+unique_ptr<PdfDestination> PdfDocument::CreateDestination()
+{
+    return unique_ptr<PdfDestination>(new PdfDestination(*this));
+}
+
+unique_ptr<PdfColorSpace> PdfDocument::CreateColorSpace(PdfColorSpaceFilterPtr filter)
+{
+    if (filter->IsTrivial())
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidDataType, "Only non trivial color spaces can be constructed through document");
+
+    return unique_ptr<PdfColorSpace>(new PdfColorSpace(*this, std::move(filter)));
+}
+
+unique_ptr<PdfFunction> PdfDocument::CreateFunction(PdfFunctionDefinitionPtr definition)
+{
+    return unique_ptr<PdfFunction>(new PdfFunction(*this, std::move(definition)));
+}
+
+unique_ptr<PdfUncolouredTilingPattern> PdfDocument::CreateTilingPattern(shared_ptr<PdfUncolouredTilingPatternDefinition> definition)
+{
+    return unique_ptr<PdfUncolouredTilingPattern>(new PdfUncolouredTilingPattern(*this, std::move(definition)));
+}
+
+unique_ptr<PdfColouredTilingPattern> PdfDocument::CreateTilingPattern(shared_ptr<PdfColouredTilingPatternDefinition> definition)
+{
+    return unique_ptr<PdfColouredTilingPattern>(new PdfColouredTilingPattern(*this, std::move(definition)));
+}
+
+unique_ptr<PdfShadingPattern> PdfDocument::CreateShadingPattern(PdfShadingPatternDefinitionPtr definition)
+{
+    return unique_ptr<PdfShadingPattern>(new PdfShadingPattern(*this, std::move(definition)));
+}
+
+unique_ptr<PdfShadingDictionary> PdfDocument::CreateShadingDictionary(PdfShadingDefinitionPtr definition)
+{
+    return unique_ptr<PdfShadingDictionary>(new PdfShadingDictionary(*this, std::move(definition)));
+}
+
+unique_ptr<PdfExtGState> PdfDocument::CreateExtGState(PdfExtGStateDefinitionPtr definition)
+{
+    return unique_ptr<PdfExtGState>(new PdfExtGState(*this, std::move(definition)));
+}
+
+unique_ptr<PdfAction> PdfDocument::CreateAction(PdfActionType type)
+{
+    return PdfAction::Create(*this, type);
+}
+
+unique_ptr<PdfFileSpec> PdfDocument::CreateFileSpec()
+{
+    return unique_ptr<PdfFileSpec>(new PdfFileSpec(*this));
 }
