@@ -19,48 +19,75 @@
 using namespace std;
 using namespace PoDoFo;
 
+namespace PoDoFo
+{
+    class PdfDynamicEncodingMap : public PdfEncodingMapBase
+    {
+    public:
+        PdfDynamicEncodingMap(shared_ptr<PdfCharCodeMap> map);
+
+        void AppendCodeSpaceRange(OutputStream& stream, charbuff& temp) const override;
+    };
+}
+
 static PdfCharCode fetchFallbackCharCode(string_view::iterator& it, const string_view::iterator& end, const PdfEncodingLimits& limits);
+static void pushCodeRangeSize(vector<unsigned char>& codeRangeSizes, unsigned char codeRangeSize);
 
 PdfEncoding::PdfEncoding()
-    : PdfEncoding(NullEncodingId, PdfEncodingMapFactory::GetNullEncodingMap(), nullptr)
+    : PdfEncoding(NullEncodingId, shared_ptr(PdfEncodingMapFactory::GetNullEncodingInstancePtr()), nullptr)
 {
 }
 
-PdfEncoding::PdfEncoding(const PdfEncodingMapConstPtr& encoding, const PdfToUnicodeMapConstPtr& toUnicode)
-    : PdfEncoding(GetNextId(), encoding, toUnicode)
+PdfEncoding::PdfEncoding(PdfEncodingMapConstPtr encoding, PdfToUnicodeMapConstPtr toUnicode)
+    : PdfEncoding(GetNextId(), std::move(encoding), std::move(toUnicode))
 {
-    if (toUnicode != nullptr && toUnicode->GetType() != PdfEncodingMapType::CMap)
+    if (m_ToUnicode != nullptr && m_ToUnicode->GetType() != PdfEncodingMapType::CMap)
         PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InternalLogic, "The encoding map must be CMap type");
 }
 
-PdfEncoding::PdfEncoding(size_t id, const PdfEncodingMapConstPtr& encoding, const PdfEncodingMapConstPtr& toUnicode)
-    : m_Id(id), m_Encoding(encoding), m_ToUnicode(toUnicode)
+PdfEncoding::PdfEncoding(unsigned id, PdfEncodingMapConstPtr&& encoding, PdfEncodingMapConstPtr&& toUnicode)
+    : m_Id(id), m_IsObjectLoaded(false), m_Font(nullptr), m_Encoding(std::move(encoding)), m_ToUnicode(std::move(toUnicode))
 {
-    if (encoding == nullptr)
+    if (m_Encoding == nullptr)
         PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidHandle, "Main encoding must be not null");
 }
 
-PdfEncoding::PdfEncoding(const PdfObject& fontObj, const PdfEncodingMapConstPtr& encoding, const PdfEncodingMapConstPtr& toUnicode)
-    : PdfEncoding(GetNextId(), encoding, toUnicode)
+PdfEncoding::PdfEncoding(unsigned id, bool isObjectLoaded, const PdfEncodingLimits& limits, PdfFont* font,
+        PdfEncodingMapConstPtr&& encoding, PdfEncodingMapConstPtr&& toUnicode,
+        PdfCIDToGIDMapConstPtr&& cidToGidMap) :
+    m_Id(id), m_IsObjectLoaded(isObjectLoaded), m_ParsedLimits(limits), m_Font(font),
+    m_Encoding(std::move(encoding)), m_ToUnicode(std::move(toUnicode)), m_CIDToGIDMap(std::move(cidToGidMap))
 {
-    auto firstCharObj = fontObj.GetDictionary().FindKey("FirstChar");
-    if (firstCharObj != nullptr)
-        m_ParsedLimits.FirstChar = PdfCharCode(static_cast<unsigned>(firstCharObj->GetNumber()));
-
-    auto lastCharObj = fontObj.GetDictionary().FindKey("LastChar");
-    if (lastCharObj != nullptr)
-        m_ParsedLimits.LastChar = PdfCharCode(static_cast<unsigned>(lastCharObj->GetNumber()));
-
-    if (m_ParsedLimits.LastChar.Code > m_ParsedLimits.FirstChar.Code)
-    {
-        // If found valid /FirstChar and /LastChar, valorize
-        //  also the code size limits
-        m_ParsedLimits.MinCodeSize = utls::GetCharCodeSize(m_ParsedLimits.FirstChar.Code);
-        m_ParsedLimits.MaxCodeSize = utls::GetCharCodeSize(m_ParsedLimits.LastChar.Code);
-    }
 }
 
-PdfEncoding::~PdfEncoding() { }
+PdfEncoding PdfEncoding::Create(const PdfEncoding& ref, PdfToUnicodeMapConstPtr&& toUnicode)
+{
+    return PdfEncoding(GetNextId(), ref.IsObjectLoaded(), ref.GetLimits(),
+        nullptr, ref.GetEncodingMapPtr(), std::move(toUnicode), nullptr);
+}
+
+PdfEncoding PdfEncoding::Create(const PdfEncodingLimits& parsedLimits, PdfEncodingMapConstPtr&& encoding,
+    PdfEncodingMapConstPtr&& toUnicode, PdfCIDToGIDMapConstPtr&& cidToGidMap)
+{
+    return PdfEncoding(GetNextId(), true, parsedLimits, nullptr,
+        std::move(encoding), std::move(toUnicode), std::move(cidToGidMap));
+}
+
+unique_ptr<PdfEncoding> PdfEncoding::CreateSchim(const PdfEncoding& encoding, PdfFont& font)
+{
+    unique_ptr<PdfEncoding> ret(new PdfEncoding(encoding));
+    ret->m_Font = &font;
+    return ret;
+}
+
+unique_ptr<PdfEncoding> PdfEncoding::CreateDynamicEncoding(const shared_ptr<PdfCharCodeMap>& cidMap,
+    const shared_ptr<PdfCharCodeMap>& toUnicodeMap, PdfFont& font)
+{
+    unique_ptr<PdfEncoding> ret(new PdfEncoding(GetNextId(), PdfEncodingMapConstPtr(new PdfDynamicEncodingMap(cidMap)),
+        PdfEncodingMapConstPtr(new PdfDynamicEncodingMap(toUnicodeMap))));
+    ret->m_Font = &font;
+    return ret;
+}
 
 string PdfEncoding::ConvertToUtf8(const PdfString& encodedStr) const
 {
@@ -90,11 +117,11 @@ bool PdfEncoding::TryConvertToEncoded(const string_view& str, charbuff& encoded)
     if (str.empty())
         return true;
 
-    auto& font = GetFont();
-    if (font.IsObjectLoaded() || !font.GetMetrics().HasUnicodeMapping())
+    PODOFO_ASSERT(m_Font != nullptr);
+    if (m_IsObjectLoaded || !m_Font->GetMetrics().HasUnicodeMapping())
     {
-        // The font is loaded from object. We will attempt to use
-        // just the loaded map to perform the conversion
+        // The font is loaded from object or substitute. We will attempt
+        // to use the loaded map to perform the conversion
         const PdfEncodingMap* toUnicode;
         if (!GetToUnicodeMapSafe(toUnicode))
             return false;
@@ -117,7 +144,7 @@ bool PdfEncoding::TryConvertToEncoded(const string_view& str, charbuff& encoded)
         // If the font is not loaded from object but created
         // from scratch, we will attempt first to infer GIDs
         // from Unicode code points using the font metrics
-        auto& metrics = font.GetMetrics();
+        auto& metrics = m_Font->GetMetrics();
         auto it = str.begin();
         auto end = str.end();
         vector<unsigned> gids;
@@ -146,7 +173,7 @@ bool PdfEncoding::TryConvertToEncoded(const string_view& str, charbuff& encoded)
         {
             unsigned char cpsSpanSize = backwardMap[i];
             unicodeview span(cps.data() + cpOffset, cpsSpanSize);
-            if (!tryGetCharCode(font, gids[i], span, codeUnit))
+            if (!tryGetCharCode(*m_Font, gids[i], span, codeUnit))
                 return false;
 
             codeUnit.AppendTo(encoded);
@@ -159,10 +186,17 @@ bool PdfEncoding::TryConvertToEncoded(const string_view& str, charbuff& encoded)
 
 bool PdfEncoding::tryGetCharCode(PdfFont& font, unsigned gid, const unicodeview& codePoints, PdfCharCode& codeUnit) const
 {
-    if (font.IsSubsettingEnabled())
+    if (font.IsSubsettingEnabled() && !font.IsProxy())
     {
-        codeUnit = font.AddSubsetGIDSafe(gid, codePoints).Unit;
-        return true;
+        PdfCID cid;
+        if (font.TryAddSubsetGID(gid, codePoints, cid))
+        {
+            codeUnit = cid.Unit;
+            return true;
+        }
+
+        codeUnit = { };
+        return false;
     }
     else
     {
@@ -173,10 +207,7 @@ bool PdfEncoding::tryGetCharCode(PdfFont& font, unsigned gid, const unicodeview&
         }
         else
         {
-            if (!GetToUnicodeMapSafe().TryGetCharCode(codePoints, codeUnit))
-                return false;
-
-            return true;
+            return GetToUnicodeMapSafe().TryGetCharCode(codePoints, codeUnit);
         }
     }
 }
@@ -192,23 +223,23 @@ bool PdfEncoding::tryConvertEncodedToUtf8(const string_view& encoded, string& st
     bool success = true;
     auto it = encoded.begin();
     auto end = encoded.end();
-    vector<char32_t> codePoints;
+    CodePointSpan codePoints;
     while (it != end)
     {
         if (!map.TryGetNextCodePoints(it, end, codePoints))
         {
             success = false;
-            codePoints.clear();
-            codePoints.push_back((char32_t)fetchFallbackCharCode(it, end, limits).Code);
+            codePoints = CodePointSpan((char32_t)fetchFallbackCharCode(it, end, limits).Code);
         }
 
-        for (size_t i = 0; i < codePoints.size(); i++)
+        auto view = codePoints.view();
+        for (size_t i = 0; i < view.size(); i++)
         {
-            char32_t codePoint = codePoints[i];
+            char32_t codePoint = view[i];
             if (codePoint != U'\0' && utf8::internal::is_code_point_valid(codePoint))
             {
                 // Validate codepoints to insert
-                utf8::unchecked::append((uint32_t)codePoints[i], std::back_inserter(str));
+                utf8::unchecked::append((uint32_t)view[i], std::back_inserter(str));
             }
         }
     }
@@ -233,8 +264,8 @@ bool PdfEncoding::TryGetCIDId(const PdfCharCode& codeUnit, unsigned& cid) const
     else
     {
         PODOFO_INVARIANT(m_Encoding->IsSimpleEncoding());
-        auto& font = GetFont();
-        if (font.IsObjectLoaded() || !font.GetMetrics().HasUnicodeMapping())
+        PODOFO_ASSERT(m_Font != nullptr);
+        if (m_IsObjectLoaded || !m_Font->GetMetrics().HasUnicodeMapping())
         {
             // Assume cid == charcode
             cid = codeUnit.Code;
@@ -246,7 +277,7 @@ bool PdfEncoding::TryGetCIDId(const PdfCharCode& codeUnit, unsigned& cid) const
             // a GID from the metrics
             char32_t cp = GetCodePoint(codeUnit);
             unsigned gid;
-            if (cp == U'\0' || !font.GetMetrics().TryGetGID(cp, gid))
+            if (cp == U'\0' || !m_Font->GetMetrics().TryGetGID(cp, gid))
             {
                 cid = 0;
                 return false;
@@ -308,49 +339,59 @@ const PdfCharCode& PdfEncoding::GetLastChar() const
     return limits.LastChar;
 }
 
-void PdfEncoding::ExportToFont(PdfFont& font, PdfEncodingExportFlags flags) const
+void PdfEncoding::ExportToFont(PdfFont& font) const
 {
-    auto& fontDict = font.GetObject().GetDictionary();
-    if (font.IsCIDKeyed())
+    exportToFont(font, nullptr);
+}
+
+void PdfEncoding::ExportToFont(PdfFont& font, const PdfCIDSystemInfo& cidInfo) const
+{
+    exportToFont(font, &cidInfo);
+}
+
+void PdfEncoding::exportToFont(PdfFont& font, const PdfCIDSystemInfo* cidInfo) const
+{
+    auto& fontDict = font.GetDictionary();
+    if (font.IsCIDFont())
     {
-        auto fontName = font.GetName();
+        PODOFO_ASSERT(cidInfo != nullptr);
 
         // The CIDSystemInfo, should be an indirect object
-        auto& cidSystemInfo = font.GetDocument().GetObjects().CreateDictionaryObject();
-        cidSystemInfo.GetDictionary().AddKey("Registry", PdfString(CMAP_REGISTRY_NAME));
-        cidSystemInfo.GetDictionary().AddKey("Ordering", PdfString(fontName));
-        cidSystemInfo.GetDictionary().AddKey("Supplement", static_cast<int64_t>(0));
+        auto& cidInfoObj = font.GetDocument().GetObjects().CreateDictionaryObject();
+        cidInfoObj.GetDictionary().AddKey("Registry"_n, cidInfo->Registry);
+        cidInfoObj.GetDictionary().AddKey("Ordering"_n, cidInfo->Ordering);
+        cidInfoObj.GetDictionary().AddKey("Supplement"_n, static_cast<int64_t>(cidInfo->Supplement));
 
         // NOTE: Setting the CIDSystemInfo params in the descendant font object is required
-        font.GetDescendantFontObject().GetDictionary().AddKeyIndirect("CIDSystemInfo", cidSystemInfo);
+        font.GetDescendantFontObject().GetDictionary().AddKeyIndirect("CIDSystemInfo"_n, cidInfoObj);
 
         // Some CMap encodings has a name representation, such as
         // Identity-H/Identity-V. NOTE: Use a fixed representation only
-        // if we are not subsetting. In that case we want a CID mapping
-        if (font.IsSubsettingEnabled() || !tryExportObjectTo(fontDict, true))
+        // if we are not subsetting. In that case we unconditionally want a CID mapping
+        if (font.HasCIDSubset() || !tryExportEncodingTo(fontDict, true))
         {
-            // If it doesn't have a name represenation, try to export a CID CMap
+            // If it doesn't have a name representation, try to export a CID CMap
             auto& cmapObj = fontDict.GetOwner()->GetDocument()->GetObjects().CreateDictionaryObject();
 
             // NOTE: Setting the CIDSystemInfo params in the CMap stream object is required
-            cmapObj.GetDictionary().AddKeyIndirect("CIDSystemInfo", cidSystemInfo);
+            cmapObj.GetDictionary().AddKeyIndirect("CIDSystemInfo"_n, cidInfoObj);
 
-            writeCIDMapping(cmapObj, GetFont(), fontName);
-            fontDict.AddKeyIndirect("Encoding", cmapObj);
+            writeCIDMapping(cmapObj, font, *cidInfo);
+            fontDict.AddKeyIndirect("Encoding"_n, cmapObj);
         }
     }
-    else
+    else // Simple font
     {
-        if (!tryExportObjectTo(fontDict, false))
+        if (!tryExportEncodingTo(fontDict, false))
             PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InternalLogic, "The encoding should supply an export object");
+
+        fontDict.AddKey("FirstChar"_n, PdfVariant(static_cast<int64_t>(GetFirstChar().Code)));
+        fontDict.AddKey("LastChar"_n, PdfVariant(static_cast<int64_t>(GetLastChar().Code)));
     }
 
-    if ((flags & PdfEncodingExportFlags::SkipToUnicode) == PdfEncodingExportFlags::None)
-    {
-        auto& cmapObj = fontDict.GetOwner()->GetDocument()->GetObjects().CreateDictionaryObject();
-        writeToUnicodeCMap(cmapObj);
-        fontDict.AddKeyIndirect("ToUnicode", cmapObj);
-    }
+    auto& cmapObj = fontDict.GetOwner()->GetDocument()->GetObjects().CreateDictionaryObject();
+    writeToUnicodeCMap(cmapObj, font);
+    fontDict.AddKeyIndirect("ToUnicode"_n, cmapObj);
 }
 
 PdfStringScanContext PdfEncoding::StartStringScan(const PdfString& encodedStr)
@@ -358,7 +399,7 @@ PdfStringScanContext PdfEncoding::StartStringScan(const PdfString& encodedStr)
     return PdfStringScanContext(encodedStr.GetRawData(), *this);
 }
 
-bool PdfEncoding::tryExportObjectTo(PdfDictionary& dictionary, bool wantCIDMapping) const
+bool PdfEncoding::tryExportEncodingTo(PdfDictionary& dictionary, bool wantCIDMapping) const
 {
     if (wantCIDMapping && !HasCIDMapping())
     {
@@ -375,9 +416,9 @@ bool PdfEncoding::tryExportObjectTo(PdfDictionary& dictionary, bool wantCIDMappi
         return false;
 
     if (obj == nullptr)
-        dictionary.AddKey("Encoding", name);
+        dictionary.AddKey("Encoding"_n, name);
     else
-        dictionary.AddKeyIndirect("Encoding", *obj);
+        dictionary.AddKeyIndirect("Encoding"_n, *obj);
 
     return true;
 }
@@ -421,38 +462,33 @@ bool PdfEncoding::HasParsedLimits() const
 
 bool PdfEncoding::IsDynamicEncoding() const
 {
-    return false;
-}
-
-PdfFont& PdfEncoding::GetFont() const
-{
-    PODOFO_RAISE_ERROR_INFO(PdfErrorCode::NotImplemented, "The encoding has not been binded to a font");
+    return m_Encoding != nullptr && typeid(*m_Encoding) == typeid(PdfDynamicEncodingMap);
 }
 
 char32_t PdfEncoding::GetCodePoint(const PdfCharCode& codeUnit) const
 {
     auto& map = GetToUnicodeMapSafe();
-    vector<char32_t> codePoints;
+    CodePointSpan codePoints;
     if (!map.TryGetCodePoints(codeUnit, codePoints)
-        || codePoints.size() != 1)
+        || codePoints.GetSize() != 1)
     {
         return U'\0';
     }
 
-    return codePoints[0];
+    return *codePoints;
 }
 
 char32_t PdfEncoding::GetCodePoint(unsigned charCode) const
 {
     auto& map = GetToUnicodeMapSafe();
     auto& limits = map.GetLimits();
-    vector<char32_t> codePoints;
+    CodePointSpan codePoints;
     for (unsigned char i = limits.MinCodeSize; i <= limits.MaxCodeSize; i++)
     {
         if (map.TryGetCodePoints({ charCode, i }, codePoints)
-            && codePoints.size() == 1)
+            && codePoints.GetSize() == 1)
         {
-            return codePoints[0];
+            return *codePoints;
         }
     }
 
@@ -489,7 +525,7 @@ const PdfEncodingMap& PdfEncoding::GetToUnicodeMapSafe() const
     return *ret;
 }
 
-const PdfEncodingMapConstPtr PdfEncoding::GetToUnicodeMapPtr() const
+PdfEncodingMapConstPtr PdfEncoding::GetToUnicodeMapPtr() const
 {
     if (m_ToUnicode != nullptr)
         return m_ToUnicode;
@@ -540,17 +576,17 @@ PdfCharCode fetchFallbackCharCode(string_view::iterator& it, const string_view::
     return { code, i };
 }
 
-void PdfEncoding::writeCIDMapping(PdfObject& cmapObj, const PdfFont& font, const string_view& fontName) const
+void PdfEncoding::writeCIDMapping(PdfObject& cmapObj, const PdfFont& font, const PdfCIDSystemInfo& cidInfo) const
 {
     // CMap specification is in Adobe technical node #5014
     auto& cmapDict = cmapObj.GetDictionary();
-    string cmapName = (string)fontName;
-    if (font.IsSubsettingEnabled())
-        cmapName.append("-subset");
+
+    string cmapName = "CMap-";
+    cmapName.append(cidInfo.Ordering.GetString());
 
     // Table 120: Additional entries in a CMap stream dictionary
-    cmapDict.AddKey(PdfName::KeyType, PdfName("CMap"));
-    cmapDict.AddKey("CMapName", PdfName(cmapName));
+    cmapDict.AddKey("Type"_n, "CMap"_n);
+    cmapDict.AddKey("CMapName"_n, PdfName(cmapName));
 
     charbuff temp;
     auto& stream = cmapObj.GetOrCreateStream();
@@ -560,98 +596,45 @@ void PdfEncoding::writeCIDMapping(PdfObject& cmapObj, const PdfFont& font, const
         "12 dict begin\n"
         "begincmap\n"
         "/CIDSystemInfo <<\n"
-        "   /Registry (" CMAP_REGISTRY_NAME  ")\n"
+        "   /Registry ({})\n"
         "   /Ordering ({})\n"
-        "   /Supplement 0\n"
+        "   /Supplement {}\n"
         ">> def\n"
         "/CMapName /{} def\n"
         "/CMapType 1 def\n"     // As defined in Adobe Technical Notes #5099
-        "1 begincodespacerange\n", fontName, cmapName);
+        , cidInfo.Registry.GetString(), cidInfo.Ordering.GetString(),
+        cidInfo.Supplement, cmapName);
+    auto wmode = m_Encoding->GetWModeSafe();
+    if (wmode != PdfWModeKind::Horizontal)
+    {
+        utls::FormatTo(temp, "/WMode {} def\n", (unsigned)wmode);
+        output.Write(temp);
+    }
     output.Write(temp);
 
-    if (font.IsSubsettingEnabled())
+    unique_ptr<PdfEncodingMap> replCIDEncodingMap;
+    if (font.TryGetSubstituteCIDEncoding(replCIDEncodingMap))
     {
-        struct Limit
-        {
-            PdfCharCode FirstCode;
-            PdfCharCode LastCode;
-        };
-
-        unordered_map<unsigned char, Limit> ranges;
-        auto& usedGids = font.GetUsedGIDs();
-        for (auto& pair : usedGids)
-        {
-            auto& codeUnit = pair.second.Unit;
-            auto found = ranges.find(codeUnit.CodeSpaceSize);
-            if (found == ranges.end())
-            {
-                ranges[codeUnit.CodeSpaceSize] = Limit{ codeUnit, codeUnit };
-            }
-            else
-            {
-                auto& limit = found->second;
-                if (codeUnit.Code < limit.FirstCode.Code)
-                    limit.FirstCode = codeUnit;
-
-                if (codeUnit.Code > limit.LastCode.Code)
-                    limit.LastCode = codeUnit;
-            }
-        }
-
-        bool first = true;
-        for (auto& pair : ranges)
-        {
-            if (first)
-                first = false;
-            else
-                output.Write("\n");
-
-            auto& range = pair.second;
-            range.FirstCode.WriteHexTo(temp);
-            output.Write(temp);
-            range.LastCode.WriteHexTo(temp);
-            output.Write(temp);
-        }
+        replCIDEncodingMap->AppendCodeSpaceRange(output, temp);
+        replCIDEncodingMap->AppendCIDMappingEntries(output, font, temp);
     }
     else
     {
         m_Encoding->AppendCodeSpaceRange(output, temp);
-    }
-
-    output.Write("\nendcodespacerange\n");
-
-    if (font.IsSubsettingEnabled())
-    {
-        auto& usedGids = font.GetUsedGIDs();
-        output.Write(std::to_string(usedGids.size()));
-        output.Write(" begincidchar\n");
-        string code;
-        for (auto& pair : usedGids)
-        {
-            auto& cid = pair.second;
-            cid.Unit.WriteHexTo(code);
-            output.Write(code);
-            output.Write(" ");
-            output.Write(std::to_string(cid.Id));
-            output.Write("\n");;
-        }
-        output.Write("endcidchar\n");
-    }
-    else
-    {
         m_Encoding->AppendCIDMappingEntries(output, font, temp);
     }
 
     output.Write(
         "endcmap\n"
-        "CMapName currentdict / CMap defineresource pop\n"
+        "CMapName currentdict /CMap defineresource pop\n"
         "end\n"
         "end");
 }
 
-void PdfEncoding::writeToUnicodeCMap(PdfObject& cmapObj) const
+void PdfEncoding::writeToUnicodeCMap(PdfObject& cmapObj, const PdfFont& font) const
 {
-    // NOTE: We definetely want a valid Unicode map at this point
+    (void)font;
+    // NOTE: We definitely want a valid Unicode map at this point
     charbuff temp;
     auto& toUnicode = GetToUnicodeMap();
     auto& stream = cmapObj.GetOrCreateStream();
@@ -669,22 +652,20 @@ void PdfEncoding::writeToUnicodeCMap(PdfObject& cmapObj) const
         "   /Supplement 0\n"
         ">> def\n"
         "/CMapName /Adobe-Identity-UCS def\n"
-        "/CMapType 2 def\n"     // As defined in Adobe Technical Notes #5099
-        "1 begincodespacerange\n");
+        "/CMapType 2 def\n");     // As defined in Adobe Technical Notes #5099
     toUnicode.AppendCodeSpaceRange(output, temp);
-    output.Write("\nendcodespacerange\n");
 
     toUnicode.AppendToUnicodeEntries(output, temp);
     output.Write(
-        "\nendcmap\n"
-        "CMapName currentdict / CMap defineresource pop\n"
+        "endcmap\n"
+        "CMapName currentdict /CMap defineresource pop\n"
         "end\n"
         "end");
 }
 
-size_t PdfEncoding::GetNextId()
+unsigned PdfEncoding::GetNextId()
 {
-    static atomic<size_t> s_nextid = CustomEncodingStartId;
+    static atomic<unsigned> s_nextid = CustomEncodingStartId;
     return s_nextid++;
 }
 
@@ -702,7 +683,13 @@ bool PdfStringScanContext::IsEndOfString() const
     return m_it == m_end;
 }
 
-bool PdfStringScanContext::TryScan(PdfCID& cid, string& utf8str, vector<codepoint>& codepoints)
+bool PdfStringScanContext::TryScan(PdfCID& cid, string& utf8str, CodePointSpan& codepoints)
+{
+    vector<unsigned> positions;
+    return TryScan(cid, utf8str, positions, codepoints);
+}
+
+bool PdfStringScanContext::TryScan(PdfCID & cid, string &utf8str, vector<unsigned>&positions, CodePointSpan & codepoints)
 {
     bool success = true;
     if (!m_encoding->TryGetNextCID(m_it, m_end, cid))
@@ -712,15 +699,19 @@ bool PdfStringScanContext::TryScan(PdfCID& cid, string& utf8str, vector<codepoin
         success = false;
     }
 
-    if (m_toUnicode->TryGetCodePoints(cid.Unit, codepoints))
+    if (m_toUnicode->TryGetCodePoints(cid, codepoints))
     {
-        for (size_t i = 0; i < codepoints.size(); i++)
+        auto view = codepoints.view();
+        unsigned prevPos = (unsigned)utf8str.length();
+        for (size_t i = 0; i < view.size(); i++)
         {
-            char32_t codePoint = codepoints[i];
+            char32_t codePoint = view[i];
             if (codePoint != U'\0' && utf8::internal::is_code_point_valid(codePoint))
             {
                 // Validate codepoints to insert
-                utf8::unchecked::append((uint32_t)codepoints[i], std::back_inserter(utf8str));
+                utf8::unchecked::append((uint32_t)view[i], std::back_inserter(utf8str));
+                positions.push_back(prevPos);
+                prevPos = (unsigned)utf8str.length();
             }
         }
     }
@@ -731,3 +722,60 @@ bool PdfStringScanContext::TryScan(PdfCID& cid, string& utf8str, vector<codepoin
 
     return success;
 }
+
+PdfDynamicEncodingMap::PdfDynamicEncodingMap(shared_ptr<PdfCharCodeMap> map)
+    : PdfEncodingMapBase(std::move(map), PdfEncodingMapType::CMap) {
+}
+
+void PdfDynamicEncodingMap::AppendCodeSpaceRange(OutputStream& stream, charbuff& temp) const
+{
+    vector<unsigned char> usedCodeSpaceSizes;
+    for (auto& pair : m_charMap->GetMappings())
+        pushCodeRangeSize(usedCodeSpaceSizes, pair.first.CodeSpaceSize);
+
+    for (auto& range : m_charMap->GetRanges())
+        pushCodeRangeSize(usedCodeSpaceSizes, range.SrcCodeLo.CodeSpaceSize);
+
+    unsigned size = 0;
+    for (auto& usedCodeSpaceSize : usedCodeSpaceSizes)
+    {
+        vector<utls::FSSUTFRange> ranges = utls::GetFSSUTFRanges(usedCodeSpaceSize);
+        size += (unsigned)ranges.size();
+    }
+
+    stream.Write(std::to_string(size));
+    stream.Write(" begincodespacerange\n");
+
+    bool first = true;
+    for (auto& usedCodeSpaceSize : usedCodeSpaceSizes)
+    {
+        vector<utls::FSSUTFRange> ranges = utls::GetFSSUTFRanges(usedCodeSpaceSize);
+
+        for (auto& range : ranges)
+        {
+            if (first)
+                first = false;
+            else
+                stream.Write("\n");
+
+            PdfCharCode firstCode(range.FirstCode);
+            PdfCharCode lastCode(range.LastCode);
+
+            firstCode.WriteHexTo(temp);
+            stream.Write(temp);
+            lastCode.WriteHexTo(temp);
+            stream.Write(temp);
+        }
+    }
+
+    stream.Write("\nendcodespacerange\n");
+}
+
+void pushCodeRangeSize(vector<unsigned char>& codeRangeSizes, unsigned char codeRangeSize)
+{
+    auto found = std::find(codeRangeSizes.begin(), codeRangeSizes.end(), codeRangeSize);
+    if (found != codeRangeSizes.end())
+        return;
+
+    codeRangeSizes.push_back(codeRangeSize);
+}
