@@ -10,18 +10,29 @@
 #include "PdfDocument.h"
 #include "PdfArray.h"
 #include "PdfDictionary.h"
-#include "PdfEncrypt.h"
-#include "PdfStreamedObjectStream.h"
+#include "PdfStatefulEncrypt.h"
 #include "PdfMemoryObjectStream.h"
+
 #include <podofo/auxiliary/StreamDevice.h>
+#include <podofo/private/PdfStreamedObjectStream.h>
 
 using namespace std;
 using namespace PoDoFo;
 
-PdfObject PdfObject::Null;
+const PdfObject PdfObject::Null = PdfObject(nullptr);
 
 PdfObject::PdfObject()
-    : PdfObject(PdfDictionary(), PdfReference(), false) { }
+    : m_Variant(new PdfDictionary()), m_IsDirty(false), m_IsImmutable(false)
+{
+    initObject();
+    m_Variant.GetDictionaryUnsafe().SetOwner(*this);
+}
+
+PdfObject::PdfObject(nullptr_t)
+    : m_IsDirty(false), m_IsImmutable(false)
+{
+    initObject();
+}
 
 PdfObject::~PdfObject() { }
 
@@ -29,76 +40,76 @@ PdfObject::PdfObject(const PdfVariant& var)
     : PdfObject(PdfVariant(var), PdfReference(), false) { }
 
 PdfObject::PdfObject(PdfVariant&& var) noexcept
-    : m_Variant(std::move(var)), m_IsDirty(false)
+    : m_Variant(std::move(var)), m_IsDirty(false), m_IsImmutable(false)
 {
     initObject();
     SetVariantOwner();
 }
 
 PdfObject::PdfObject(bool b)
-    : m_Variant(b), m_IsDirty(false)
+    : m_Variant(b), m_IsDirty(false), m_IsImmutable(false)
 {
     initObject();
 }
 
 PdfObject::PdfObject(int64_t l)
-    : m_Variant(l), m_IsDirty(false)
+    : m_Variant(l), m_IsDirty(false), m_IsImmutable(false)
 {
     initObject();
 }
 
 PdfObject::PdfObject(double d)
-    : m_Variant(d), m_IsDirty(false)
+    : m_Variant(d), m_IsDirty(false), m_IsImmutable(false)
 {
     initObject();
 }
 
 PdfObject::PdfObject(const PdfString& str)
-    : m_Variant(str), m_IsDirty(false)
+    : m_Variant(str), m_IsDirty(false), m_IsImmutable(false)
 {
     initObject();
 }
 PdfObject::PdfObject(const PdfName& name)
-    : m_Variant(name), m_IsDirty(false)
+    : m_Variant(name), m_IsDirty(false), m_IsImmutable(false)
 {
     initObject();
 }
 
 PdfObject::PdfObject(const PdfReference& ref)
-    : m_Variant(ref), m_IsDirty(false)
+    : m_Variant(ref), m_IsDirty(false), m_IsImmutable(false)
 {
     initObject();
 }
 
 PdfObject::PdfObject(const PdfArray& arr)
-    : m_Variant(arr), m_IsDirty(false)
+    : m_Variant(arr), m_IsDirty(false), m_IsImmutable(false)
 {
     initObject();
     m_Variant.GetArray().SetOwner(*this);
 }
 
 PdfObject::PdfObject(PdfArray&& arr) noexcept
-    : m_Variant(std::move(arr)), m_IsDirty(false)
+    : m_Variant(std::move(arr)), m_IsDirty(false), m_IsImmutable(false)
 {
     initObject();
     m_Variant.GetArray().SetOwner(*this);
 }
 
 PdfObject::PdfObject(const PdfDictionary& dict)
-    : m_Variant(dict), m_IsDirty(false)
+    : m_Variant(dict), m_IsDirty(false), m_IsImmutable(false)
 {
     initObject();
-    m_Variant.GetDictionary().SetOwner(*this);
+    m_Variant.GetDictionaryUnsafe().SetOwner(*this);
 }
 
 PdfObject::PdfObject(PdfDictionary&& dict) noexcept
-    : m_Variant(std::move(dict)), m_IsDirty(false)
+    : m_Variant(std::move(dict)), m_IsDirty(false), m_IsImmutable(false)
 {
     initObject();
-    m_Variant.GetDictionary().SetOwner(*this);
+    m_Variant.GetDictionaryUnsafe().SetOwner(*this);
 }
 
-// NOTE: Don't copy parent document/container/indirect refernce.
+// NOTE: Don't copy parent document/container/indirect reference.
 // Copied objects must be always detached. Ownership will be set
 // automatically elsewhere
 PdfObject::PdfObject(const PdfObject& rhs)
@@ -110,13 +121,27 @@ PdfObject::PdfObject(const PdfObject& rhs)
 // NOTE: Don't move parent document/container. Copied objects must be
 // always detached. Ownership will be set automatically elsewhere.
 // Also don't move reference
-PdfObject::PdfObject(PdfObject&& rhs) noexcept :
-    m_Variant(std::move(rhs.m_Variant)),
-    m_IsDirty(false)
+PdfObject::PdfObject(PdfObject&& rhs) noexcept
+    : m_Document(nullptr), m_Parent(nullptr), m_IsDirty(false), m_IsImmutable(false)
+{
+    moveFrom(std::move(rhs));
+    rhs.SetDirty();
+}
+
+// NOTE: Dirty objects are those who are supposed to be serialized
+// or deserialized.
+PdfObject::PdfObject(PdfVariant&& var, const PdfReference& indirectReference, bool isDirty)
+    : m_Variant(std::move(var)), m_IndirectReference(indirectReference), m_IsDirty(isDirty), m_IsImmutable(false)
 {
     initObject();
     SetVariantOwner();
-    moveStreamFrom(rhs);
+}
+
+PdfObject::PdfObject(PdfArray* arr)
+    : m_Variant(arr), m_IsDirty(false), m_IsImmutable(false)
+{
+    initObject();
+    m_Variant.GetDictionaryUnsafe().SetOwner(*this);
 }
 
 const PdfObjectStream* PdfObject::GetStream() const
@@ -131,15 +156,6 @@ PdfObjectStream* PdfObject::GetStream()
     return m_Stream.get();
 }
 
-// NOTE: Dirty objects are those who are supposed to be serialized
-// or deserialized.
-PdfObject::PdfObject(PdfVariant&& var, const PdfReference& indirectReference, bool isDirty)
-    : m_Variant(std::move(var)), m_IndirectReference(indirectReference), m_IsDirty(isDirty)
-{
-    initObject();
-    SetVariantOwner();
-}
-
 void PdfObject::ForceCreateStream()
 {
     DelayedLoadStream();
@@ -163,14 +179,14 @@ void PdfObject::DelayedLoad() const
     if (m_IsDelayedLoadDone)
         return;
 
-    const_cast<PdfObject&>(*this).DelayedLoadImpl();
+    const_cast<PdfObject&>(*this).delayedLoad();
     m_IsDelayedLoadDone = true;
     const_cast<PdfObject&>(*this).SetVariantOwner();
 }
 
-void PdfObject::DelayedLoadImpl()
+void PdfObject::delayedLoad()
 {
-    // Default implementation of virtual void DelayedLoadImpl() throws, since delayed
+    // Default implementation of virtual void delayedLoad() throws, since delayed
     // loading should not be enabled except by types that support it.
     PODOFO_RAISE_ERROR(PdfErrorCode::InternalLogic);
 }
@@ -181,10 +197,10 @@ void PdfObject::SetVariantOwner()
     switch (dataType)
     {
         case PdfDataType::Dictionary:
-            m_Variant.GetDictionary().SetOwner(*this);
+            m_Variant.GetDictionaryUnsafe().SetOwner(*this);
             break;
         case PdfDataType::Array:
-            m_Variant.GetArray().SetOwner(*this);
+            m_Variant.GetArrayUnsafe().SetOwner(*this);
             break;
         default:
             break;
@@ -205,31 +221,27 @@ void PdfObject::initObject()
     m_IsDelayedLoadStreamDone = true;
 }
 
-void PdfObject::Write(OutputStream& device, PdfWriteFlags writeMode,
-    const PdfEncrypt* encrypt_, charbuff& buffer) const
+void PdfObject::Write(OutputStream& stream, PdfWriteFlags writeMode,
+    const PdfStatefulEncrypt* encrypt, charbuff& buffer) const
 {
-    DelayedLoad();
     DelayedLoadStream();
+    write(stream, true, writeMode, encrypt, buffer);
+}
 
-    PdfStatefulEncrypt encrypt;
-    if (encrypt_ != nullptr)
-        encrypt = PdfStatefulEncrypt(*encrypt_, m_IndirectReference);
+void PdfObject::WriteFinal(OutputStream& stream, PdfWriteFlags writeMode, const PdfStatefulEncrypt* encrypt, charbuff& buffer)
+{
+    DelayedLoadStream();
+    write(stream, false, writeMode, encrypt, buffer);
 
+    // After writing we can reset the dirty flag
+    ResetDirty();
+}
+
+void PdfObject::write(OutputStream& stream, bool skipLengthFix,
+    PdfWriteFlags writeMode, const PdfStatefulEncrypt* encrypt, charbuff& buffer) const
+{
     if (m_IndirectReference.IsIndirect())
-    {
-        if ((writeMode & PdfWriteFlags::Clean) == PdfWriteFlags::None
-            && (writeMode & PdfWriteFlags::NoPDFAPreserve) != PdfWriteFlags::None)
-        {
-            utls::FormatTo(buffer, "{} {} obj", m_IndirectReference.ObjectNumber(), m_IndirectReference.GenerationNumber());
-            device.Write(buffer);
-        }
-        else
-        {
-            // PDF/A compliance requires all objects to be written in a clean way
-            utls::FormatTo(buffer, "{} {} obj\n", m_IndirectReference.ObjectNumber(), m_IndirectReference.GenerationNumber());
-            device.Write(buffer);
-        }
-    }
+        WriteHeader(stream, writeMode, buffer);
 
     if (m_Stream != nullptr)
     {
@@ -244,40 +256,53 @@ void PdfObject::Write(OutputStream& device, PdfWriteFlags writeMode,
                 || m_IndirectReference != metadataObj->GetIndirectReference()))
         {
             PdfObject object;
-            auto& stream = object.GetOrCreateStream();
+            auto& objStream = object.GetOrCreateStream();
             {
-                auto output = stream.GetOutputStream({ PdfFilterType::FlateDecode });
+                auto output = objStream.GetOutputStream({ PdfFilterType::FlateDecode });
                 auto input = m_Stream->GetInputStream();
                 input.CopyTo(output);
             }
 
-            m_Stream->MoveFrom(stream);
+            m_Stream->MoveFrom(objStream);
         }
 
         // Set length if it's not handled by the underlying provider
-        if (!m_Stream->GetProvider().IsLengthHandled())
+        if (!skipLengthFix)
         {
             size_t length = m_Stream->GetLength();
-            if (encrypt.HasEncrypt())
-                length = encrypt.CalculateStreamLength(length);
+            if (encrypt != nullptr)
+                length = encrypt->CalculateStreamLength(length);
 
             // Add the key without triggering SetDirty
-            const_cast<PdfObject&>(*this).m_Variant.GetDictionary()
-                .AddKey(PdfName::KeyLength, static_cast<int64_t>(length), true);
+            const_cast<PdfObject&>(*this).m_Variant.GetDictionaryUnsafe()
+                .AddKeyNoDirtySet("Length"_n, PdfVariant(static_cast<int64_t>(length)));
         }
     }
 
-    m_Variant.Write(device, writeMode, encrypt, buffer);
-    device.Write('\n');
+    m_Variant.Write(stream, writeMode, encrypt, buffer);
+    stream.Write('\n');
 
     if (m_Stream != nullptr)
-        m_Stream->Write(device, encrypt);
+        m_Stream->Write(stream, encrypt);
 
     if (m_IndirectReference.IsIndirect())
-        device.Write("endobj\n");
+        stream.Write("endobj\n");
+}
 
-    // After write we ca reset the dirty flag
-    const_cast<PdfObject&>(*this).ResetDirty();
+void PdfObject::WriteHeader(OutputStream& stream, PdfWriteFlags writeMode, charbuff& buffer) const
+{
+    if ((writeMode & PdfWriteFlags::Clean) != PdfWriteFlags::None
+        || (writeMode & PdfWriteFlags::PdfAPreserve) != PdfWriteFlags::None)
+    {
+        // PDF/A compliance requires all objects to be written in a clean way
+        utls::FormatTo(buffer, "{} {} obj\n", m_IndirectReference.ObjectNumber(), m_IndirectReference.GenerationNumber());
+        stream.Write(buffer);
+    }
+    else
+    {
+        utls::FormatTo(buffer, "{} {} obj", m_IndirectReference.ObjectNumber(), m_IndirectReference.GenerationNumber());
+        stream.Write(buffer);
+    }
 }
 
 PdfObjectStream& PdfObject::GetOrCreateStream()
@@ -286,6 +311,18 @@ PdfObjectStream& PdfObject::GetOrCreateStream()
     return getOrCreateStream();
 }
 
+void PdfObject::RemoveStream()
+{
+    DelayedLoad();
+    // Unconditionally set the stream as already loaded,
+    // then just remove it
+    m_IsDelayedLoadStreamDone = true;
+    bool hasStream = m_Stream != nullptr || removeStream();
+    m_Stream = nullptr;
+    if (hasStream)
+        SetDirty();
+}
+
 const PdfObjectStream& PdfObject::MustGetStream() const
 {
     DelayedLoadStream();
@@ -309,10 +346,16 @@ bool PdfObject::IsIndirect() const
     return m_IndirectReference.IsIndirect();
 }
 
+bool PdfObject::TryUnload()
+{
+    // Do nothing on base PdfObject class
+    return false;
+}
+
 bool PdfObject::HasStream() const
 {
-    DelayedLoadStream();
-    return m_Stream != nullptr;
+    DelayedLoad();
+    return m_Stream != nullptr || HasStreamToParse();
 }
 
 PdfObjectStream& PdfObject::getOrCreateStream()
@@ -356,7 +399,7 @@ void PdfObject::delayedLoadStream() const
     if (m_IsDelayedLoadStreamDone)
         return;
 
-    const_cast<PdfObject&>(*this).DelayedLoadStreamImpl();
+    const_cast<PdfObject&>(*this).delayedLoadStream();
     m_IsDelayedLoadStreamDone = true;
 }
 
@@ -372,8 +415,9 @@ PdfObject& PdfObject::operator=(const PdfObject& rhs)
 
 PdfObject& PdfObject::operator=(PdfObject&& rhs) noexcept
 {
-    moveFrom(rhs);
+    moveFrom(std::move(rhs));
     SetDirty();
+    rhs.SetDirty();
     return *this;
 }
 
@@ -385,7 +429,7 @@ void PdfObject::copyStreamFrom(const PdfObject& obj)
     if (obj.m_Stream != nullptr)
     {
         auto& stream = getOrCreateStream();
-        stream = *obj.m_Stream;
+        stream.CopyFrom(*obj.m_Stream);
     }
 }
 
@@ -393,6 +437,8 @@ void PdfObject::moveStreamFrom(PdfObject& obj)
 {
     obj.DelayedLoadStream();
     m_Stream = std::move(obj.m_Stream);
+    if (m_Stream != nullptr)
+        m_Stream->SetParent(*this);
 }
 
 void PdfObject::EnableDelayedLoading()
@@ -405,21 +451,50 @@ void PdfObject::EnableDelayedLoadingStream()
     m_IsDelayedLoadStreamDone = false;
 }
 
-void PdfObject::DelayedLoadStreamImpl()
+void PdfObject::SetRevised()
+{
+    // Do nothing on base PdfObject class
+}
+
+void PdfObject::delayedLoadStream()
 {
     // Default implementation throws, since delayed loading of
-    // steams should not be enabled except by types that support it.
+    // streams should not be enabled except by types that support it.
     PODOFO_RAISE_ERROR(PdfErrorCode::InternalLogic);
 }
 
-void PdfObject::Assign(const PdfObject& rhs)
+bool PdfObject::removeStream()
 {
-    if (&rhs == this)
-        return;
+    // Do nothing for regular object
+    return false;
+}
 
+bool PdfObject::HasStreamToParse() const
+{
+    return false;
+}
+
+void PdfObject::AssignNoDirtySet(const PdfObject& rhs)
+{
+    PODOFO_ASSERT(&rhs != this);
     assign(rhs);
 }
 
+void PdfObject::AssignNoDirtySet(PdfObject&& rhs)
+{
+    PODOFO_ASSERT(&rhs != this);
+    moveFrom(std::move(rhs));
+}
+
+void PdfObject::AssignNoDirtySet(PdfVariant&& rhs)
+{
+    m_Variant = std::move(rhs);
+    m_IsDelayedLoadDone = true;
+    SetVariantOwner();
+    m_Stream = nullptr;
+    m_IsDelayedLoadStreamDone = true;
+}
+
 void PdfObject::SetParent(PdfDataContainer& parent)
 {
     m_Parent = &parent;
@@ -427,6 +502,12 @@ void PdfObject::SetParent(PdfDataContainer& parent)
     SetDocument(document);
 }
 
+void PdfObject::assertMutable() const
+{
+    if (m_IsImmutable)
+        PODOFO_RAISE_ERROR(PdfErrorCode::ChangeOnImmutable);
+}
+
 // NOTE: Don't copy parent document/container and indirect reference.
 // Objects being assigned always keep current ownership
 void PdfObject::assign(const PdfObject& rhs)
@@ -441,7 +522,7 @@ void PdfObject::assign(const PdfObject& rhs)
 
 // NOTE: Don't move parent document/container and indirect reference.
 // Objects being assigned always keep current ownership
-void PdfObject::moveFrom(PdfObject& rhs)
+void PdfObject::moveFrom(PdfObject&& rhs)
 {
     rhs.DelayedLoad();
     m_Variant = std::move(rhs.m_Variant);
@@ -460,10 +541,10 @@ void PdfObject::ResetDirty()
         // Arrays and Dictionaries
         // handle dirty status by themselves
         case PdfDataType::Array:
-            static_cast<PdfDataContainer&>(m_Variant.GetArray()).ResetDirty();
+            m_Variant.GetArrayUnsafe().ResetDirty();
             break;
         case PdfDataType::Dictionary:
-            static_cast<PdfDataContainer&>(m_Variant.GetDictionary()).ResetDirty();
+            m_Variant.GetDictionaryUnsafe().ResetDirty();
             break;
         case PdfDataType::Bool:
         case PdfDataType::Number:
@@ -491,7 +572,7 @@ void PdfObject::SetDirty()
     else if (m_Parent != nullptr)
     {
         // Reset parent if not indirect. Resetting will stop at
-        // first indirect anchestor
+        // first indirect ancestor
         m_Parent->SetDirty();
     }
 }
@@ -499,6 +580,7 @@ void PdfObject::SetDirty()
 void PdfObject::setDirty()
 {
     m_IsDirty = true;
+    SetRevised();
 }
 
 void PdfObject::resetDirty()
@@ -532,19 +614,35 @@ PdfDataType PdfObject::GetDataType() const
     return m_Variant.GetDataType();
 }
 
-string PdfObject::ToString() const
+string PdfObject::ToString(PdfWriteFlags writeFlags) const
 {
     string ret;
-    ToString(ret);
+    ToString(ret, writeFlags);
     return ret;
 }
 
-void PdfObject::ToString(string& ret) const
+void PdfObject::ToString(string& ret, PdfWriteFlags writeFlags) const
 {
+    DelayedLoadStream();
     ret.clear();
+    switch (m_Variant.GetDataType())
+    {
+        case PdfDataType::Null:
+        case PdfDataType::Bool:
+        case PdfDataType::Number:
+        case PdfDataType::Real:
+        case PdfDataType::Reference:
+            // We enforce the literals to not be spaced
+            writeFlags |= PdfWriteFlags::NoInlineLiteral;
+            break;
+        default:
+            // Do nothing
+            break;
+    }
+
     StringStreamDevice device(ret);
     charbuff buffer;
-    Write(device, PdfWriteFlags::None, nullptr, buffer);
+    write(device, true, writeFlags, nullptr, buffer);
 }
 
 bool PdfObject::GetBool() const
@@ -667,6 +765,22 @@ bool PdfObject::TryGetArray(PdfArray*& arr)
     return m_Variant.TryGetArray(arr);
 }
 
+bool PdfObject::TryGetArray(PdfArray& arr) const
+{
+    DelayedLoad();
+    const PdfArray* val;
+    if (m_Variant.TryGetArray(val))
+    {
+        arr = *val;
+        return true;
+    }
+    else
+    {
+        arr.Clear();
+        return false;
+    }
+}
+
 const PdfDictionary& PdfObject::GetDictionary() const
 {
     DelayedLoad();
@@ -691,6 +805,22 @@ bool PdfObject::TryGetDictionary(PdfDictionary*& dict)
     return m_Variant.TryGetDictionary(dict);
 }
 
+bool PdfObject::TryGetDictionary(PdfDictionary& dict) const
+{
+    DelayedLoad();
+    const PdfDictionary* val;
+    if (m_Variant.TryGetDictionary(val))
+    {
+        dict = *val;
+        return true;
+    }
+    else
+    {
+        dict.Clear();
+        return false;
+    }
+}
+
 PdfReference PdfObject::GetReference() const
 {
     DelayedLoad();
@@ -705,6 +835,7 @@ bool PdfObject::TryGetReference(PdfReference& ref) const
 
 void PdfObject::SetBool(bool b)
 {
+    assertMutable();
     DelayedLoad();
     m_Variant.SetBool(b);
     SetDirty();
@@ -712,6 +843,7 @@ void PdfObject::SetBool(bool b)
 
 void PdfObject::SetNumber(int64_t l)
 {
+    assertMutable();
     DelayedLoad();
     m_Variant.SetNumber(l);
     SetDirty();
@@ -719,6 +851,7 @@ void PdfObject::SetNumber(int64_t l)
 
 void PdfObject::SetReal(double d)
 {
+    assertMutable();
     DelayedLoad();
     m_Variant.SetReal(d);
     SetDirty();
@@ -726,6 +859,7 @@ void PdfObject::SetReal(double d)
 
 void PdfObject::SetName(const PdfName& name)
 {
+    assertMutable();
     DelayedLoad();
     m_Variant.SetName(name);
     SetDirty();
@@ -733,6 +867,7 @@ void PdfObject::SetName(const PdfName& name)
 
 void PdfObject::SetString(const PdfString& str)
 {
+    assertMutable();
     DelayedLoad();
     m_Variant.SetString(str);
     SetDirty();
@@ -740,12 +875,25 @@ void PdfObject::SetString(const PdfString& str)
 
 void PdfObject::SetReference(const PdfReference& ref)
 {
+    assertMutable();
     DelayedLoad();
     m_Variant.SetReference(ref);
     SetDirty();
 }
 
-const char* PdfObject::GetDataTypeString() const
+void PdfObject::SetNumberNoDirtySet(int64_t l)
+{
+    PODOFO_ASSERT(m_IsDelayedLoadDone);
+    m_Variant.SetNumber(l);
+}
+
+void PdfObject::SetImmutable()
+{
+    PODOFO_ASSERT(m_IsDelayedLoadDone);
+    m_IsImmutable = true;
+}
+
+string_view PdfObject::GetDataTypeString() const
 {
     DelayedLoad();
     return m_Variant.GetDataTypeString();
