@@ -15,14 +15,15 @@
 #include "PdfFont.h"
 #include "PdfFontMetrics.h"
 #include "PdfImage.h"
+#include "PdfDocument.h"
 
 using namespace std;
 using namespace PoDoFo;
 
 static string expandTabs(const string_view& str, unsigned tabWidth, unsigned tabCount);
 
-PdfPainter::PdfPainter(PdfPainterFlags flags) :
-    m_flags(flags),
+PdfPainter::PdfPainter() :
+    m_flags(PdfPainterFlags::None),
     m_painterStatus(StatusDefault),
     m_textStackCount(0),
     GraphicsState(*this, m_StateStack.Current->GraphicsState),
@@ -42,21 +43,27 @@ PdfPainter::~PdfPainter() noexcept(false)
     }
     catch (...)
     {
-        if (!std::uncaught_exceptions())
+        if (std::uncaught_exceptions() == 0)
             throw;
     }
 }
 
-void PdfPainter::SetCanvas(PdfCanvas& canvas)
+void PdfPainter::SetCanvas(PdfCanvas& canvas, PdfPainterFlags flags)
 {
-    // Ignore setting the same canvas twice
     if (m_canvas == &canvas)
+    {
+        if (flags != m_flags && m_objStream != nullptr)
+            PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InternalLogic, "Invalid setting the same canvas with different painter flags");
+
+        // Ignore setting the same canvas twice
         return;
+    }
 
     finishDrawing();
     reset();
     canvas.EnsureResourcesCreated();
     m_canvas = &canvas;
+    m_flags = flags;
     m_objStream = nullptr;
 }
 
@@ -124,46 +131,14 @@ void PdfPainter::finishDrawing()
 
 void PdfPainter::reset()
 {
+    m_flags = PdfPainterFlags::None;
+    m_painterStatus = PainterStatus::StatusDefault;
     m_StateStack.Clear();
-    m_stream.Clear();
+    m_textStackCount = 0;
     m_objStream = nullptr;
     m_canvas = nullptr;
-}
-
-void PdfPainter::SetStrokingShadingPattern(const PdfShadingPattern& pattern)
-{
-    checkStream();
-    checkStatus(StatusDefault);
-    this->addToPageResources("Pattern", pattern.GetIdentifier(), pattern.GetObject());
-    PoDoFo::WriteOperator_CS(m_stream, PdfColorSpace::Pattern);
-    PoDoFo::WriteOperator_SCN(m_stream, pattern.GetIdentifier().GetString());
-}
-
-void PdfPainter::SetShadingPattern(const PdfShadingPattern& pattern)
-{
-    checkStream();
-    checkStatus(StatusDefault);
-    this->addToPageResources("Pattern", pattern.GetIdentifier(), pattern.GetObject());
-    PoDoFo::WriteOperator_cs(m_stream, PdfColorSpace::Pattern);
-    PoDoFo::WriteOperator_scn(m_stream, pattern.GetIdentifier().GetString());
-}
-
-void PdfPainter::SetStrokingTilingPattern(const PdfTilingPattern& pattern)
-{
-    checkStream();
-    checkStatus(StatusDefault);
-    this->addToPageResources("Pattern", pattern.GetIdentifier(), pattern.GetObject());
-    PoDoFo::WriteOperator_CS(m_stream, PdfColorSpace::Pattern);
-    PoDoFo::WriteOperator_SCN(m_stream, pattern.GetIdentifier().GetString());
-}
-
-void PdfPainter::SetTilingPattern(const PdfTilingPattern& pattern)
-{
-    checkStream();
-    checkStatus(StatusDefault);
-    this->addToPageResources("Pattern", pattern.GetIdentifier(), pattern.GetObject());
-    PoDoFo::WriteOperator_cs(m_stream, PdfColorSpace::Pattern);
-    PoDoFo::WriteOperator_scn(m_stream, pattern.GetIdentifier().GetString());
+    m_stream.Clear();
+    m_resNameCache.clear();
 }
 
 void PdfPainter::SetStrokeStyle(PdfStrokeStyle strokeStyle, bool inverted, double scale, bool subtractJoinCap)
@@ -356,49 +331,51 @@ void PdfPainter::DrawText(const string_view& str, double x, double y,
     checkStatus(StatusDefault);
     checkFont();
 
+    vector<array<double, 4>> linesToDraw;
+    save();
     PoDoFo::WriteOperator_BT(m_stream);
     writeTextState();
     drawText(str, x, y,
         (style & PdfDrawTextStyle::Underline) != PdfDrawTextStyle::Regular,
-        (style & PdfDrawTextStyle::StrikeThrough) != PdfDrawTextStyle::Regular);
+        (style & PdfDrawTextStyle::StrikeThrough) != PdfDrawTextStyle::Regular, linesToDraw);
     PoDoFo::WriteOperator_ET(m_stream);
+    drawLines(linesToDraw);
+    restore();
 }
 
-void PdfPainter::drawText(const string_view& str, double x, double y, bool isUnderline, bool isStrikeThrough)
+void PdfPainter::drawText(const string_view& str, double x, double y,
+    bool isUnderline, bool isStrikeThrough, vector<array<double, 4>>& linesToDraw)
 {
-    PoDoFo::WriteOperator_Td(m_stream, x, y);
-
     auto& textState = m_StateStack.Current->TextState;
     auto& font = *textState.Font;
     auto expStr = this->expandTabs(str);
 
     if (isUnderline || isStrikeThrough)
     {
-        this->save();
-
         // Draw underline
         this->setLineWidth(font.GetUnderlineThickness(textState));
         if (isUnderline)
         {
-            this->DrawLine(x,
+            linesToDraw.push_back({ x,
                 y + font.GetUnderlinePosition(textState),
                 x + font.GetStringLength(expStr, textState),
-                y + font.GetUnderlinePosition(textState));
+                y + font.GetUnderlinePosition(textState)
+            });
         }
 
         // Draw strikethrough
-        this->setLineWidth(font.GetStrikeThroughThickness(textState));
         if (isStrikeThrough)
         {
-            this->DrawLine(x,
+            linesToDraw.push_back({ x,
                 y + font.GetStrikeThroughPosition(textState),
                 x + font.GetStringLength(expStr, textState),
-                y + font.GetStrikeThroughPosition(textState));
+                y + font.GetStrikeThroughPosition(textState)
+            });
         }
-
-        this->restore();
     }
 
+    PoDoFo::WriteOperator_Td(m_stream, x, y);
+
     PoDoFo::WriteOperator_Tj(m_stream, font.GetEncoding().ConvertToEncoded(str),
         !font.GetEncoding().IsSimpleEncoding());
 }
@@ -421,7 +398,7 @@ void PdfPainter::DrawTextMultiLine(const string_view& str, double x, double y, d
 
     drawMultiLineText(str, x, y, width, height,
         params.HorizontalAlignment, params.VerticalAlignment,
-        params.Clip, params.SkipSpaces, params.Style);
+        params.SkipClip, params.PreserveTrailingSpaces, params.Style);
 }
 
 void PdfPainter::DrawTextAligned(const string_view& str, double x, double y, double width,
@@ -434,28 +411,32 @@ void PdfPainter::DrawTextAligned(const string_view& str, double x, double y, dou
     checkStatus(StatusDefault | StatusTextObject);
     checkFont();
 
+    save();
     PoDoFo::WriteOperator_BT(m_stream);
     writeTextState();
-    drawTextAligned(str, x, y, width, hAlignment, style);
+    vector<array<double, 4>> linesToDraw;
+    drawTextAligned(str, x, y, width, hAlignment, style, linesToDraw);
     PoDoFo::WriteOperator_ET(m_stream);
+    drawLines(linesToDraw);
+    restore();
 }
 
 void PdfPainter::drawMultiLineText(const string_view& str, double x, double y, double width, double height,
-    PdfHorizontalAlignment hAlignment, PdfVerticalAlignment vAlignment, bool clip, bool skipSpaces,
+    PdfHorizontalAlignment hAlignment, PdfVerticalAlignment vAlignment, bool skipClip, bool preserveTrailingSpaces,
     PdfDrawTextStyle style)
 {
     auto& textState = m_StateStack.Current->TextState;
     auto& font = *textState.Font;
 
     this->save();
-    if (clip)
+    if (!skipClip)
         this->SetClipRect(x, y, width, height);
 
     auto expanded = this->expandTabs(str);
 
     PoDoFo::WriteOperator_BT(m_stream);
     writeTextState();
-    vector<string> lines = getMultiLineTextAsLines(expanded, width, skipSpaces);
+    vector<string> lines = m_StateStack.Current->TextState.SplitTextAsLines(str, width, preserveTrailingSpaces);
     double lineGap = font.GetLineSpacing(textState) - font.GetAscent(textState) + font.GetDescent(textState);
     // Do vertical alignment
     switch (vAlignment)
@@ -473,195 +454,34 @@ void PdfPainter::drawMultiLineText(const string_view& str, double x, double y, d
     }
 
     y -= font.GetAscent(textState) + lineGap / 2;
+    vector<array<double, 4>> linesToDraw;
     for (auto& line : lines)
     {
         if (line.length() != 0)
-            this->drawTextAligned(line, x, y, width, hAlignment, style);
+            this->drawTextAligned(line, x, y, width, hAlignment, style, linesToDraw);
 
         x = 0;
+        switch (hAlignment)
+        {
+            default:
+            case PdfHorizontalAlignment::Left:
+                break;
+            case PdfHorizontalAlignment::Center:
+                x = -(width - textState.Font->GetStringLength(line, textState)) / 2.0;
+                break;
+            case PdfHorizontalAlignment::Right:
+                x = -(width - textState.Font->GetStringLength(line, textState));
+                break;
+        }
         y = -font.GetLineSpacing(textState);
     }
     PoDoFo::WriteOperator_ET(m_stream);
+    drawLines(linesToDraw);
     this->restore();
 }
 
-vector<string> PdfPainter::getMultiLineTextAsLines(const string_view& str, double width, bool skipSpaces)
-{
-    if (width <= 0) // nonsense arguments
-        return vector<string>();
-
-    if (str.length() == 0) // empty string
-        return vector<string>(1, (string)str);
-
-    auto& textState = m_StateStack.Current->TextState;
-    auto& font = *textState.Font;
-
-    bool startOfWord = true;
-    double curWidthOfLine = 0;
-    vector<string> lines;
-
-    // do simple word wrapping
-    auto it = str.begin();
-    auto end = str.end();
-    auto lineBegin = it;
-    auto prevIt = it;
-    auto startOfCurrentWord = it;
-    while (it != end)
-    {
-        char32_t ch = (char32_t)utf8::next(it, end);
-        if (utls::IsNewLineLikeChar(ch)) // hard-break!
-        {
-            lines.push_back((string)str.substr(lineBegin - str.begin(), prevIt - lineBegin));
-
-            lineBegin = it; // skip the line feed
-            startOfWord = true;
-            curWidthOfLine = 0;
-        }
-        else if (utls::IsSpaceLikeChar(ch))
-        {
-            if (curWidthOfLine > width)
-            {
-                // The previous word does not fit in the current line.
-                // -> Move it to the next one.
-                if (startOfCurrentWord > lineBegin)
-                {
-                    lines.push_back((string)str.substr(lineBegin - str.begin(), startOfCurrentWord - lineBegin));
-                }
-                else
-                {
-                    lines.push_back((string)str.substr(lineBegin - str.begin(), prevIt - lineBegin));
-                    if (skipSpaces)
-                    {
-                        // Skip all spaces at the end of the line
-                        while (it != end)
-                        {
-                            ch = (char32_t)utf8::next(it, end);
-                            if (!utls::IsSpaceLikeChar(ch))
-                                break;
-                        }
-
-                        startOfCurrentWord = it;
-                    }
-                    else
-                    {
-                        startOfCurrentWord = prevIt;
-                    }
-                    startOfWord = true;
-                }
-                lineBegin = startOfCurrentWord;
-
-                if (!startOfWord)
-                {
-                    curWidthOfLine = font.GetStringLength(
-                        str.substr(startOfCurrentWord - str.begin(), prevIt - startOfCurrentWord),
-                        textState);
-                }
-                else
-                {
-                    curWidthOfLine = 0;
-                }
-            }
-            else if ((curWidthOfLine + font.GetCharLength(ch, textState)) > width)
-            {
-                lines.push_back((string)str.substr(lineBegin - str.begin(), prevIt - lineBegin));
-                if (skipSpaces)
-                {
-                    // Skip all spaces at the end of the line
-                    while (it != end)
-                    {
-                        ch = (char32_t)utf8::next(it, end);
-                        if (!utls::IsSpaceLikeChar(ch))
-                            break;
-                    }
-
-                    startOfCurrentWord = it;
-                }
-                else
-                {
-                    startOfCurrentWord = prevIt;
-                }
-                lineBegin = startOfCurrentWord;
-                startOfWord = true;
-                curWidthOfLine = 0;
-            }
-            else
-            {
-                curWidthOfLine += font.GetCharLength(ch, textState);
-            }
-
-            startOfWord = true;
-        }
-        else
-        {
-            if (startOfWord)
-            {
-                startOfCurrentWord = prevIt;
-                startOfWord = false;
-            }
-            //else do nothing
-
-            if ((curWidthOfLine + font.GetCharLength(ch, textState)) > width)
-            {
-                if (lineBegin == startOfCurrentWord)
-                {
-                    // This word takes up the whole line.
-                    // Put as much as possible on this line.
-                    if (lineBegin == prevIt)
-                    {
-                        lines.push_back((string)str.substr(prevIt - str.begin(), it - prevIt));
-                        lineBegin = it;
-                        startOfCurrentWord = it;
-                        curWidthOfLine = 0;
-                    }
-                    else
-                    {
-                        lines.push_back((string)str.substr(lineBegin - str.begin(), prevIt - lineBegin));
-                        lineBegin = prevIt;
-                        startOfCurrentWord = prevIt;
-                        curWidthOfLine = font.GetCharLength(ch, textState);
-                    }
-                }
-                else
-                {
-                    // The current word does not fit in the current line.
-                    // -> Move it to the next one.
-                    lines.push_back((string)str.substr(lineBegin - str.begin(), startOfCurrentWord - lineBegin));
-                    lineBegin = startOfCurrentWord;
-                    curWidthOfLine = font.GetStringLength((string)str.substr(startOfCurrentWord - str.begin(), it - startOfCurrentWord), textState);
-                }
-            }
-            else
-            {
-                curWidthOfLine += font.GetCharLength(ch, textState);
-            }
-        }
-
-        prevIt = it;
-    }
-
-    if ((prevIt - lineBegin) > 0)
-    {
-        if (curWidthOfLine > width && startOfCurrentWord > lineBegin)
-        {
-            // The previous word does not fit in the current line.
-            // -> Move it to the next one.
-            lines.push_back((string)str.substr(lineBegin - str.begin(), startOfCurrentWord - lineBegin));
-            lineBegin = startOfCurrentWord;
-        }
-        //else do nothing
-
-        if (prevIt - lineBegin > 0)
-        {
-            lines.push_back((string)str.substr(lineBegin - str.begin(), prevIt - lineBegin));
-        }
-        //else do nothing
-    }
-
-    return lines;
-}
-
 void PdfPainter::drawTextAligned(const string_view& str, double x, double y, double width,
-    PdfHorizontalAlignment hAlignment, PdfDrawTextStyle style)
+    PdfHorizontalAlignment hAlignment, PdfDrawTextStyle style, vector<array<double, 4>>& linesToDraw)
 {
     auto& textState = m_StateStack.Current->TextState;
     switch (hAlignment)
@@ -679,7 +499,8 @@ void PdfPainter::drawTextAligned(const string_view& str, double x, double y, dou
 
     this->drawText(str, x, y,
         (style & PdfDrawTextStyle::Underline) != PdfDrawTextStyle::Regular,
-        (style & PdfDrawTextStyle::StrikeThrough) != PdfDrawTextStyle::Regular);
+        (style & PdfDrawTextStyle::StrikeThrough) != PdfDrawTextStyle::Regular,
+        linesToDraw);
 }
 
 void PdfPainter::DrawImage(const PdfImage& obj, double x, double y, double scaleX, double scaleY)
@@ -692,14 +513,9 @@ void PdfPainter::DrawImage(const PdfImage& obj, double x, double y, double scale
 void PdfPainter::DrawXObject(const PdfXObject& obj, double x, double y, double scaleX, double scaleY)
 {
     checkStream();
-
-    // use OriginalReference() as the XObject might have been written to disk
-    // already and is not in memory anymore in this case.
-    this->addToPageResources("XObject", obj.GetIdentifier(), obj.GetObject());
-
     PoDoFo::WriteOperator_q(m_stream);
     PoDoFo::WriteOperator_cm(m_stream, scaleX, 0, 0, scaleY, x, y);
-    PoDoFo::WriteOperator_Do(m_stream, obj.GetIdentifier().GetString());
+    PoDoFo::WriteOperator_Do(m_stream, tryAddResource(obj.GetObject(), PdfResourceType::XObject));
     PoDoFo::WriteOperator_Q(m_stream);
 }
 
@@ -770,12 +586,10 @@ void PdfPainter::restore()
     TextState.SetState(current.TextState);
 }
 
-void PdfPainter::SetExtGState(const PdfExtGState& inGState)
+void PdfPainter::SetExtGState(const PdfExtGState& extGState)
 {
     checkStream();
-    checkStatus(StatusDefault);
-    this->addToPageResources("ExtGState", inGState.GetIdentifier(), inGState.GetObject());
-    PoDoFo::WriteOperator_gs(m_stream, inGState.GetIdentifier().GetString());
+    PoDoFo::WriteOperator_gs(m_stream, tryAddResource(extGState.GetObject(), PdfResourceType::ExtGState));
 }
 
 // TODO: Validate when marked content can be put
@@ -883,118 +697,178 @@ void PdfPainter::SetRenderingIntent(const string_view& intent)
     PoDoFo::WriteOperator_ri(m_stream, intent);
 }
 
-void PdfPainter::SetFillColor(const PdfColor& color)
+void PdfPainter::SetNonStrokingColor(const PdfColor& color)
 {
     checkStream();
     switch (color.GetColorSpace())
     {
-        default:
-        case PdfColorSpace::DeviceRGB:
+        case PdfColorSpaceType::DeviceRGB:
         {
             PoDoFo::WriteOperator_rg(m_stream, color.GetRed(), color.GetGreen(), color.GetBlue());
             break;
         }
-        case PdfColorSpace::DeviceCMYK:
+        case PdfColorSpaceType::DeviceCMYK:
         {
             PoDoFo::WriteOperator_k(m_stream, color.GetCyan(), color.GetMagenta(), color.GetYellow(), color.GetBlack());
             break;
         }
-        case PdfColorSpace::DeviceGray:
+        case PdfColorSpaceType::DeviceGray:
         {
             PoDoFo::WriteOperator_g(m_stream, color.GetGrayScale());
             break;
         }
-        case PdfColorSpace::Separation:
+        default:
         {
-            m_canvas->GetOrCreateResources().AddColorResource(color);
-            vector<double> components = { color.GetDensity() };
-            PoDoFo::WriteOperator_cs(m_stream, PdfName(color.GetName()).GetEscapedName());
-            PoDoFo::WriteOperator_scn(m_stream, components);
-            break;
-        }
-        case PdfColorSpace::Lab:
-        {
-            m_canvas->GetOrCreateResources().AddColorResource(color);
-            vector<double> components = { color.GetCieL(),color.GetCieA(), color.GetCieB() };
-            PoDoFo::WriteOperator_cs(m_stream, "ColorSpaceCieLab");
-            PoDoFo::WriteOperator_scn(m_stream, components);
-            break;
-        }
-        case PdfColorSpace::Unknown:
-        case PdfColorSpace::Indexed:
-        {
-            PODOFO_RAISE_ERROR(PdfErrorCode::CannotConvertColor);
+            PODOFO_RAISE_ERROR_INFO(PdfErrorCode::CannotConvertColor, "Unsupported color space");
         }
     }
 }
 
-void PdfPainter::SetStrokeColor(const PdfColor& color)
+void PdfPainter::SetStrokingColor(const PdfColor& color)
 {
     checkStream();
     switch (color.GetColorSpace())
     {
-        default:
-        case PdfColorSpace::DeviceRGB:
+        case PdfColorSpaceType::DeviceRGB:
         {
             PoDoFo::WriteOperator_RG(m_stream, color.GetRed(), color.GetGreen(), color.GetBlue());
             break;
         }
-        case PdfColorSpace::DeviceCMYK:
+        case PdfColorSpaceType::DeviceCMYK:
         {
             PoDoFo::WriteOperator_K(m_stream, color.GetCyan(), color.GetMagenta(), color.GetYellow(), color.GetBlack());
             break;
         }
-        case PdfColorSpace::DeviceGray:
+        case PdfColorSpaceType::DeviceGray:
         {
             PoDoFo::WriteOperator_G(m_stream, color.GetGrayScale());
             break;
         }
-        case PdfColorSpace::Separation:
+        default:
         {
-            m_canvas->GetOrCreateResources().AddColorResource(color);
-            vector<double> components = { color.GetDensity() };
-            PoDoFo::WriteOperator_CS(m_stream, PdfName(color.GetName()).GetEscapedName());
-            PoDoFo::WriteOperator_SCN(m_stream, components);
-            break;
-        }
-        case PdfColorSpace::Lab:
-        {
-            m_canvas->GetOrCreateResources().AddColorResource(color);
-            vector<double> components = { color.GetCieL(),color.GetCieA(), color.GetCieB() };
-            PoDoFo::WriteOperator_CS(m_stream, "ColorSpaceCieLab");
-            PoDoFo::WriteOperator_SCN(m_stream, components);
-            break;
-        }
-        case PdfColorSpace::Unknown:
-        case PdfColorSpace::Indexed:
-        {
-            PODOFO_RAISE_ERROR(PdfErrorCode::CannotConvertColor);
+            PODOFO_RAISE_ERROR_INFO(PdfErrorCode::CannotConvertColor, "Unsupported color space");
         }
     }
 }
 
-void PdfPainter::SetFont(const PdfFont* font, double fontSize)
+void PdfPainter::SetNonStrokingColor(const PdfColorRaw& color, const PdfColorSpaceFilter& colorSpace)
+{
+    checkStream();
+    PoDoFo::WriteOperator_scn(m_stream, cspan<double>(color.data(), colorSpace.GetColorComponentCount()));
+}
+
+void PdfPainter::SetStrokingColor(const PdfColorRaw& color, const PdfColorSpaceFilter& colorSpace)
+{
+    checkStream();
+    PoDoFo::WriteOperator_SCN(m_stream, cspan<double>(color.data(), colorSpace.GetColorComponentCount()));
+}
+
+void PdfPainter::SetNonStrokingColorSpace(const PdfVariant& expVar)
+{
+    checkStream();
+    switch (expVar.GetDataType())
+    {
+        case PdfDataType::Name:
+        {
+            PoDoFo::WriteOperator_cs(m_stream, expVar.GetName());
+            break;
+        }
+        case PdfDataType::Reference:
+        {
+            PoDoFo::WriteOperator_cs(m_stream, tryAddResource(expVar.GetReference(), PdfResourceType::ColorSpace));
+            break;
+        }
+        default:
+            PODOFO_RAISE_ERROR_INFO(PdfErrorCode::UnsupportedFilter, "Unsupported setting the colorspace without an exorpt object");
+    }
+}
+
+void PdfPainter::SetStrokingColorSpace(const PdfVariant& expVar)
+{
+    checkStream();
+    switch (expVar.GetDataType())
+    {
+        case PdfDataType::Name:
+        {
+            PoDoFo::WriteOperator_CS(m_stream, expVar.GetName());
+            break;
+        }
+        case PdfDataType::Reference:
+        {
+            PoDoFo::WriteOperator_CS(m_stream, tryAddResource(expVar.GetReference(), PdfResourceType::ColorSpace));
+            break;
+        }
+        default:
+            PODOFO_RAISE_ERROR_INFO(PdfErrorCode::UnsupportedFilter, "Unsupported setting the colorspace without an exorpt object");
+    }
+}
+
+void PdfPainter::SetStrokingPattern(const PdfPattern& pattern, const PdfColorRaw* color, const PdfColorSpaceFilter* colorSpace)
+{
+    checkStream();
+    if (color == nullptr)
+        PoDoFo::WriteOperator_SCN(m_stream, tryAddResource(pattern.GetObject(), PdfResourceType::Pattern));
+    else
+        PoDoFo::WriteOperator_SCN(m_stream, cspan<double>(color->data(), colorSpace->GetColorComponentCount()), tryAddResource(pattern.GetObject(), PdfResourceType::Pattern));
+}
+
+void PdfPainter::SetNonStrokingPattern(const PdfPattern& pattern, const PdfColorRaw* color, const PdfColorSpaceFilter* colorSpace)
+{
+    checkStream();
+    if (color == nullptr)
+        PoDoFo::WriteOperator_scn(m_stream, tryAddResource(pattern.GetObject(), PdfResourceType::Pattern));
+    else
+        PoDoFo::WriteOperator_scn(m_stream, cspan<double>(color->data(), colorSpace->GetColorComponentCount()), tryAddResource(pattern.GetObject(), PdfResourceType::Pattern));
+}
+
+void PdfPainter::SetShadingDictionary(const PdfShadingDictionary& shading)
+{
+    checkStream();
+    PoDoFo::WriteOperator_sh(m_stream, tryAddResource(shading.GetObject(), PdfResourceType::Shading));
+}
+
+PdfName PdfPainter::tryAddResource(const PdfObject& obj, PdfResourceType type)
+{
+    return tryAddResource(obj.GetIndirectReference(), type);
+}
+
+PdfName PdfPainter::tryAddResource(const PdfReference& ref, PdfResourceType type)
+{
+    auto found = m_resNameCache.find(ref);
+    if (found == m_resNameCache.end())
+    {
+        auto name = m_canvas->GetOrCreateResources().AddResource(type, ref);
+        m_resNameCache[ref] = name;
+        return name;
+    }
+
+    return found->second;
+}
+
+void PdfPainter::drawLines(const vector<array<double, 4>>& lines)
+{
+    for (auto& line : lines)
+        this->DrawLine(line[0], line[1], line[2], line[3]);
+}
+
+void PdfPainter::SetFont(const PdfFont& font, double fontSize)
 {
-    if (font == nullptr)
-        return;
-
     checkStream();
-    this->addToPageResources("Font", font->GetIdentifier(), font->GetObject());
     if (m_painterStatus == StatusTextObject)
         setFont(font, fontSize);
 }
 
-void PdfPainter::setFont(const PdfFont* font, double fontSize)
+void PdfPainter::setFont(const PdfFont& font, double fontSize)
 {
     auto& textState = m_StateStack.Current->EmittedTextState;
-    if (textState.Font == font
+    if (textState.Font == &font
         && textState.FontSize == fontSize)
     {
         return;
     }
 
-    PoDoFo::WriteOperator_Tf(m_stream, font->GetIdentifier().GetString(), fontSize);
-    textState.Font = font;
+    PoDoFo::WriteOperator_Tf(m_stream, tryAddResource(font.GetObject(), PdfResourceType::Font), fontSize);
+    textState.Font = &font;
     textState.FontSize = fontSize;
 }
 
@@ -1056,6 +930,13 @@ void PdfPainter::SetTextRenderingMode(PdfTextRenderingMode value)
         setTextRenderingMode(value);
 }
 
+void PdfPainter::SetTextMatrix(const Matrix& matrix)
+{
+    checkStream();
+    if (m_painterStatus == StatusTextObject)
+        setTextMatrix(matrix);
+}
+
 void PdfPainter::setTextRenderingMode(PdfTextRenderingMode value)
 {
     auto& textState = m_StateStack.Current->EmittedTextState;
@@ -1066,11 +947,21 @@ void PdfPainter::setTextRenderingMode(PdfTextRenderingMode value)
     textState.RenderingMode = value;
 }
 
+void PdfPainter::setTextMatrix(const Matrix& value)
+{
+    auto& textState = m_StateStack.Current->EmittedTextState;
+    if (textState.Matrix == value)
+        return;
+
+    PoDoFo::WriteOperator_Tm(m_stream, value[0], value[1], value[2], value[3], value[4], value[5]);
+    textState.Matrix = value;
+}
+
 void PdfPainter::writeTextState()
 {
     auto& textState = m_StateStack.Current->TextState;
     if (textState.Font != nullptr)
-        setFont(textState.Font, textState.FontSize);
+        setFont(*textState.Font, textState.FontSize);
 
     if (textState.FontScale != 1)
         setFontScale(textState.FontScale);
@@ -1083,14 +974,9 @@ void PdfPainter::writeTextState()
 
     if (textState.RenderingMode != PdfTextRenderingMode::Fill)
         setTextRenderingMode(textState.RenderingMode);
-}
 
-void PdfPainter::addToPageResources(const PdfName& type, const PdfName& identifier, const PdfObject& obj)
-{
-    if (m_canvas == nullptr)
-        PODOFO_RAISE_ERROR(PdfErrorCode::InvalidHandle);
-
-    m_canvas->GetOrCreateResources().AddResource(type, identifier, obj);
+    if (textState.Matrix != Matrix::Identity)
+        setTextMatrix(textState.Matrix);
 }
 
 string PdfPainter::expandTabs(const string_view& str) const
@@ -1124,7 +1010,7 @@ void PdfPainter::checkStream()
         return;
 
     PODOFO_RAISE_LOGIC_IF(m_canvas == nullptr, "Call SetCanvas() first before doing drawing operations");
-    m_objStream = &m_canvas->GetStreamForAppending((PdfStreamAppendFlags)(m_flags & (~PdfPainterFlags::NoSaveRestore)));
+    m_objStream = &m_canvas->GetOrCreateContentsStream((PdfStreamAppendFlags)(m_flags & (~PdfPainterFlags::NoSaveRestore)));
 }
 
 void PdfPainter::openPath(double x, double y)
@@ -1198,13 +1084,10 @@ void PdfPainterTextObject::End()
 PdfGraphicsStateWrapper::PdfGraphicsStateWrapper(PdfPainter& painter, PdfGraphicsState& state)
     : m_painter(&painter), m_state(&state) { }
 
-void PdfGraphicsStateWrapper::SetCurrentMatrix(const Matrix& matrix)
+void PdfGraphicsStateWrapper::ConcatenateTransformationMatrix(const Matrix& matrix)
 {
-    if (m_state->CTM == matrix)
-        return;
-
-    m_state->CTM = matrix;
-    m_painter->SetTransformationMatrix(m_state->CTM);
+    m_state->CTM = matrix * m_state->CTM;
+    m_painter->SetTransformationMatrix(matrix);
 }
 
 void PdfGraphicsStateWrapper::SetLineWidth(double lineWidth)
@@ -1252,71 +1135,208 @@ void PdfGraphicsStateWrapper::SetRenderingIntent(const string_view& intent)
     m_painter->SetRenderingIntent(m_state->RenderingIntent);
 }
 
-void PdfGraphicsStateWrapper::SetFillColor(const PdfColor& color)
+void PdfGraphicsStateWrapper::SetNonStrokingColorSpace(PdfColorSpaceInitializer&& colorSpace)
 {
-    if (m_state->FillColor == color)
+    if (m_state->NonStrokingColorSpaceFilter == colorSpace.GetFilterPtr())
         return;
 
-    m_state->FillColor = color;
-    m_painter->SetFillColor(m_state->FillColor);
+    PdfVariant expVar;
+    m_state->NonStrokingColorSpaceFilter = colorSpace.Take(expVar);
+    m_painter->SetNonStrokingColorSpace(expVar);
 }
 
-void PdfGraphicsStateWrapper::SetStrokeColor(const PdfColor& color)
+void PdfGraphicsStateWrapper::SetStrokingColorSpace(PdfColorSpaceInitializer&& colorSpace)
 {
-    if (m_state->StrokeColor == color)
+    if (m_state->StrokingColorSpaceFilter == colorSpace.GetFilterPtr())
         return;
 
-    m_state->StrokeColor = color;
-    m_painter->SetStrokeColor(m_state->StrokeColor);
+    PdfVariant expVar;
+    m_state->StrokingColorSpaceFilter = colorSpace.Take(expVar);
+    m_painter->SetStrokingColorSpace(expVar);
+}
+
+void PdfGraphicsStateWrapper::SetNonStrokingColor(const PdfColor& color)
+{
+    if (m_state->NonStrokingColorSpaceFilter->GetType() != color.GetColorSpace())
+        m_state->NonStrokingColorSpaceFilter = PdfColorSpaceFilterFactory::GetTrivialFilterPtr(color.GetColorSpace());
+
+    if (m_state->NonStrokingColor == color.GetRawColor())
+        return;
+
+    m_state->NonStrokingColor = color.GetRawColor();
+    m_painter->SetNonStrokingColor(color);
+}
+
+void PdfGraphicsStateWrapper::SetStrokingColor(const PdfColor& color)
+{
+    if (m_state->StrokingColorSpaceFilter->GetType() != color.GetColorSpace())
+        m_state->StrokingColorSpaceFilter = PdfColorSpaceFilterFactory::GetTrivialFilterPtr(color.GetColorSpace());
+
+    if (m_state->StrokingColor == color.GetRawColor())
+        return;
+
+    m_state->StrokingColor = color.GetRawColor();
+    m_painter->SetStrokingColor(color);
+}
+
+void PdfGraphicsStateWrapper::SetNonStrokingColor(const PdfColorRaw& color)
+{
+    if (m_state->NonStrokingColorSpaceFilter->GetType() == PdfColorSpaceType::Pattern)
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidInput, "Found a pattern non stroking color space set");
+
+    if (m_state->NonStrokingColor == color)
+        return;
+
+    m_state->NonStrokingColor = color;
+    m_painter->SetNonStrokingColor(color, *m_state->NonStrokingColorSpaceFilter);
+}
+
+void PdfGraphicsStateWrapper::SetStrokingColor(const PdfColorRaw& color)
+{
+    if (m_state->StrokingColorSpaceFilter->GetType() == PdfColorSpaceType::Pattern)
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidInput, "Found a pattern stroking color space set");
+
+    if (m_state->StrokingColor == color)
+        return;
+
+    m_state->StrokingColor = color;
+    m_painter->SetStrokingColor(color, *m_state->StrokingColorSpaceFilter);
+}
+
+void PdfGraphicsStateWrapper::SetExtGState(const PdfExtGState& extGState)
+{
+    if (m_state->ExtGState.get() == &extGState.GetDefinition())
+        return;
+
+    m_state->ExtGState = extGState.GetDefinitionPtr();
+    m_painter->SetExtGState(extGState);
+}
+
+void PdfGraphicsStateWrapper::SetStrokingUncolouredTilingPattern(const PdfUncolouredTilingPattern& pattern, const PdfColorRaw& color)
+{
+    if (m_state->StrokingColorSpaceFilter->GetType() != PdfColorSpaceType::Pattern)
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidInput, "Stroking color space should be pattern");
+
+    if (m_state->StrokingPattern.get() == &pattern.GetDefinition() && m_state->StrokingColor == color)
+        return;
+
+    m_state->StrokingPattern = pattern.GetDefinitionPtr();
+    m_state->StrokingColor = color;
+    m_painter->SetStrokingPattern(pattern, &color, m_state->StrokingColorSpaceFilter.get());
+}
+
+void PdfGraphicsStateWrapper::SetNonStrokingUncolouredTilingPattern(const PdfUncolouredTilingPattern& pattern, const PdfColorRaw& color)
+{
+    if (m_state->NonStrokingColorSpaceFilter->GetType() != PdfColorSpaceType::Pattern)
+        PODOFO_RAISE_ERROR_INFO(PdfErrorCode::InvalidInput, "Non stroking color space should be pattern");
+
+    if (m_state->NonStrokingPattern.get() == &pattern.GetDefinition() && m_state->NonStrokingColor == color)
+        return;
+
+    m_state->NonStrokingPattern = pattern.GetDefinitionPtr();
+    m_state->NonStrokingColor = color;
+    m_painter->SetNonStrokingPattern(pattern, &color, m_state->NonStrokingColorSpaceFilter.get());
+}
+
+void PdfGraphicsStateWrapper::SetStrokingPattern(const PdfPattern& pattern)
+{
+    if (m_state->StrokingPattern.get() == &pattern.GetDefinition())
+        return;
+
+    if (m_state->StrokingColorSpaceFilter->GetType() != PdfColorSpaceType::Pattern
+        || static_cast<const PdfColorSpaceFilterPattern&>(*m_state->StrokingColorSpaceFilter).GetUnderlyingColorSpace().GetType() != PdfColorSpaceType::Unknown)
+    {
+        m_state->StrokingColorSpaceFilter = PdfColorSpaceFilterFactory::GetParameterLessPatternInstancePtr();
+        m_painter->SetStrokingColorSpace("Pattern"_n);
+    }
+
+    m_state->StrokingPattern = pattern.GetDefinitionPtr();
+    m_state->StrokingColor = { };
+    m_painter->SetStrokingPattern(pattern, nullptr, nullptr);
+}
+
+void PdfGraphicsStateWrapper::SetNonStrokingPattern(const PdfPattern& pattern)
+{
+    if (m_state->NonStrokingPattern.get() == &pattern.GetDefinition())
+        return;
+
+    if (m_state->NonStrokingColorSpaceFilter->GetType() != PdfColorSpaceType::Pattern
+        || static_cast<const PdfColorSpaceFilterPattern&>(*m_state->NonStrokingColorSpaceFilter).GetUnderlyingColorSpace().GetType() != PdfColorSpaceType::Unknown)
+    {
+        m_state->NonStrokingColorSpaceFilter = PdfColorSpaceFilterFactory::GetParameterLessPatternInstancePtr();
+        m_painter->SetNonStrokingColorSpace("Pattern"_n);
+    }
+
+    m_state->NonStrokingPattern = pattern.GetDefinitionPtr();
+    m_state->NonStrokingColor = { };
+    m_painter->SetNonStrokingPattern(pattern, nullptr, nullptr);
+}
+
+void PdfGraphicsStateWrapper::SetShadingDictionary(const PdfShadingDictionary& shading)
+{
+    if (m_state->Shading.get() == &shading.GetDefinition())
+        return;
+
+    m_state->Shading = shading.GetDefinitionPtr();
+    m_painter->SetShadingDictionary(shading);
 }
 
 PdfTextStateWrapper::PdfTextStateWrapper(PdfPainter& painter, PdfTextState& state)
-    : m_painter(&painter), m_state(&state) { }
+    : m_painter(&painter), m_State(&state) { }
 
 void PdfTextStateWrapper::SetFont(const PdfFont& font, double fontSize)
 {
-    if (m_state->Font == &font && m_state->FontSize == fontSize)
+    if (m_State->Font == &font && m_State->FontSize == fontSize)
         return;
 
-    m_state->Font = &font;
-    m_state->FontSize = fontSize;
-    m_painter->SetFont(m_state->Font, m_state->FontSize);
+    m_State->Font = &font;
+    m_State->FontSize = fontSize;
+    m_painter->SetFont(*m_State->Font, m_State->FontSize);
 }
 
 void PdfTextStateWrapper::SetFontScale(double scale)
 {
-    if (m_state->FontScale == scale)
+    if (m_State->FontScale == scale)
         return;
 
-    m_state->FontScale = scale;
-    m_painter->SetFontScale(m_state->FontScale);
+    m_State->FontScale = scale;
+    m_painter->SetFontScale(m_State->FontScale);
 }
 
 void PdfTextStateWrapper::SetCharSpacing(double charSpacing)
 {
-    if (m_state->CharSpacing == charSpacing)
+    if (m_State->CharSpacing == charSpacing)
         return;
 
-    m_state->CharSpacing = charSpacing;
-    m_painter->SetCharSpacing(m_state->CharSpacing);
+    m_State->CharSpacing = charSpacing;
+    m_painter->SetCharSpacing(m_State->CharSpacing);
 }
 
 void PdfTextStateWrapper::SetWordSpacing(double wordSpacing)
 {
-    if (m_state->WordSpacing == wordSpacing)
+    if (m_State->WordSpacing == wordSpacing)
         return;
 
-    m_state->WordSpacing = wordSpacing;
-    m_painter->SetWordSpacing(m_state->WordSpacing);
+    m_State->WordSpacing = wordSpacing;
+    m_painter->SetWordSpacing(m_State->WordSpacing);
 }
 
 void PdfTextStateWrapper::SetRenderingMode(PdfTextRenderingMode mode)
 {
-    if (m_state->RenderingMode == mode)
+    if (m_State->RenderingMode == mode)
         return;
 
-    m_state->RenderingMode = mode;
-    m_painter->SetTextRenderingMode(m_state->RenderingMode);
+    m_State->RenderingMode = mode;
+    m_painter->SetTextRenderingMode(m_State->RenderingMode);
+}
+
+void PdfTextStateWrapper::SetMatrix(const Matrix& matrix)
+{
+    if (m_State->Matrix == matrix)
+        return;
+
+    m_State->Matrix = matrix;
+    m_painter->SetTextMatrix(m_State->Matrix);
 }
 
 void PdfPainter::drawRectangle(double x, double y, double width, double height, PdfPathDrawMode mode, double roundX, double roundY)
@@ -1394,3 +1414,10 @@ string expandTabs(const string_view& str, unsigned tabWidth, unsigned tabCount)
 
     return ret;
 }
+
+PdfPainterState::PdfPainterState()
+{
+    // Reset font size(s)
+    TextState.FontSize = -1;
+    EmittedTextState.FontSize = -1;
+}
