@@ -7,130 +7,108 @@
 #include <podofo/private/PdfDeclarationsPrivate.h>
 #include "PdfVariant.h"
 
+#include <podofo/auxiliary/StreamDevice.h>
+#include <podofo/private/PdfParserObject.h>
+
 #include "PdfArray.h"
 #include "PdfData.h"
 #include "PdfDictionary.h"
-#include "PdfParserObject.h"
-#include <podofo/auxiliary/StreamDevice.h>
 
 using namespace PoDoFo;
 using namespace std;
 
 constexpr unsigned short DefaultPrecision = 6;
 
-PdfVariant PdfVariant::Null;
-
-PdfVariant::PdfVariant(PdfDataType type)
-    : m_Data{ }, m_DataType(type) { }
+const PdfVariant PdfVariant::Null;
 
 PdfVariant::PdfVariant()
-    : PdfVariant(PdfDataType::Null) { }
+    : m_Null() { }
 
 PdfVariant::PdfVariant(bool value)
-    : PdfVariant(PdfDataType::Bool)
-{
-    m_Data.Bool = value;
-}
+    : m_Bool(value) { }
 
 PdfVariant::PdfVariant(int64_t value)
-    : PdfVariant(PdfDataType::Number)
-{
-    m_Data.Number = value;
-}
+    : m_Number(value) { }
 
 PdfVariant::PdfVariant(double value)
-    : PdfVariant(PdfDataType::Real)
-{
-    m_Data.Real = value;
-}
+    : m_Real(value) { }
 
 PdfVariant::PdfVariant(const PdfString& str)
-    : PdfVariant(PdfDataType::String)
-{
-    m_Data.Data = new PdfString(str);
-}
+    : m_String(str) { }
 
 PdfVariant::PdfVariant(const PdfName& name)
-    : PdfVariant(PdfDataType::Name)
-{
-    m_Data.Data = new PdfName(name);
-}
+    : m_Name(name) { }
 
 PdfVariant::PdfVariant(const PdfReference& ref)
-    : PdfVariant(PdfDataType::Reference)
-{
-    m_Data.Reference = ref;
-}
+    : m_Reference(ref) { }
 
 PdfVariant::PdfVariant(const PdfArray& arr)
-    : PdfVariant(PdfDataType::Array)
-{
-    m_Data.Data = new PdfArray(arr);
-}
+    : m_Array(new PdfArray(arr)) { }
 
 PdfVariant::PdfVariant(PdfArray&& arr) noexcept
-    : PdfVariant(PdfDataType::Array)
-{
-    m_Data.Data = new PdfArray(std::move(arr));
-}
+    : m_Array(new PdfArray(std::move(arr))) { }
 
 PdfVariant::PdfVariant(const PdfDictionary& dict)
-    : PdfVariant(PdfDataType::Dictionary)
-{
-    m_Data.Data = new PdfDictionary(dict);
-}
+    : m_Dictionary(new PdfDictionary(dict)) { }
 
 PdfVariant::PdfVariant(PdfDictionary&& dict) noexcept
-    : PdfVariant(PdfDataType::Dictionary)
-{
-    m_Data.Data = new PdfDictionary(std::move(dict));
-}
+    : m_Dictionary(new PdfDictionary(std::move(dict))) { }
 
 PdfVariant::PdfVariant(const PdfData& data)
-    : PdfVariant(PdfDataType::RawData)
-{
-    m_Data.Data = new PdfData(data);
-}
+    : m_Data(new PdfData(data)) { }
 
 PdfVariant::PdfVariant(PdfData&& data) noexcept
-    : PdfVariant(PdfDataType::RawData)
-{
-    m_Data.Data = new PdfData(std::move(data));
-}
-
+    : m_Data(new PdfData(std::move(data))) { }
 
 PdfVariant::PdfVariant(const PdfVariant& rhs)
-    : m_Data{ }
 {
     assign(rhs);
 }
 
 PdfVariant::PdfVariant(PdfVariant&& rhs) noexcept
-    : m_Data(rhs.m_Data), m_DataType(rhs.m_DataType)
 {
-    rhs.m_Data = { };
-    rhs.m_DataType = PdfDataType::Null;
+    moveFrom(std::move(rhs));
+}
+
+PdfVariant::PdfVariant(PdfDictionary* dict)
+    : m_Dictionary(dict)
+{
+}
+
+PdfVariant::PdfVariant(PdfArray* arr)
+    : m_Array(arr)
+{
 }
 
 PdfVariant::~PdfVariant()
 {
-    clear();
-}
-
-void PdfVariant::clear()
-{
-    switch (m_DataType)
+    switch (GetDataType())
     {
         case PdfDataType::Array:
-        case PdfDataType::Dictionary:
-        case PdfDataType::Name:
-        case PdfDataType::String:
-        case PdfDataType::RawData:
         {
-            delete m_Data.Data;
+            delete m_Array.Value;
+            break;
+        }
+        case PdfDataType::Dictionary:
+        {
+            delete m_Dictionary.Value;
+            break;
+        }
+        case PdfDataType::RawData:
+        {
+            delete m_Data.Value;
+            break;
+        }
+        case PdfDataType::Name:
+        {
+            m_Name.~PdfName();
+            break;
+        }
+        case PdfDataType::String:
+        {
+            m_String.~PdfString();
             break;
         }
-
         case PdfDataType::Reference:
         case PdfDataType::Bool:
         case PdfDataType::Null:
@@ -143,16 +121,16 @@ void PdfVariant::clear()
 }
 
 void PdfVariant::Write(OutputStream& device, PdfWriteFlags writeMode,
-    const PdfStatefulEncrypt& encrypt, charbuff& buffer) const
+    const PdfStatefulEncrypt* encrypt, charbuff& buffer) const
 {
-    switch (m_DataType)
+    switch (GetDataType())
     {
         case PdfDataType::Bool:
         {
             if ((writeMode & PdfWriteFlags::NoInlineLiteral) == PdfWriteFlags::None)
                 device.Write(' '); // Write space before true or false
 
-            if (m_Data.Bool)
+            if (m_Bool.Value)
                 device.Write("true");
             else
                 device.Write("false");
@@ -163,7 +141,7 @@ void PdfVariant::Write(OutputStream& device, PdfWriteFlags writeMode,
             if ((writeMode & PdfWriteFlags::NoInlineLiteral) == PdfWriteFlags::None)
                 device.Write(' '); // Write space before numbers
 
-            utls::FormatTo(buffer, "{}", m_Data.Number);
+            utls::FormatTo(buffer, "{}", m_Number.Value);
             device.Write(buffer);
             break;
         }
@@ -172,23 +150,40 @@ void PdfVariant::Write(OutputStream& device, PdfWriteFlags writeMode,
             if ((writeMode & PdfWriteFlags::NoInlineLiteral) == PdfWriteFlags::None)
                 device.Write(' '); // Write space before numbers
 
-            utls::FormatTo(buffer, m_Data.Real, DefaultPrecision);
+            utls::FormatTo(buffer, m_Real.Value, DefaultPrecision);
             device.Write(buffer);
             break;
         }
         case PdfDataType::Reference:
-            m_Data.Reference.Write(device, writeMode, buffer);
+        {
+            m_Reference.Write(device, writeMode, encrypt, buffer);
             break;
+        }
         case PdfDataType::String:
-        case PdfDataType::Name:
-        case PdfDataType::Array:
-        case PdfDataType::Dictionary:
-        case PdfDataType::RawData:
-            if (m_Data.Data == nullptr)
-                PODOFO_RAISE_ERROR(PdfErrorCode::InvalidHandle);
-
-            m_Data.Data->Write(device, writeMode, encrypt, buffer);
+        {
+            m_String.Write(device, writeMode, encrypt, buffer);
             break;
+        }
+        case PdfDataType::Name:
+        {
+            m_Name.Write(device, writeMode, encrypt, buffer);
+            break;
+        }
+        case PdfDataType::Array:
+        {
+            m_Array.Value->Write(device, writeMode, encrypt, buffer);
+            break;
+        }
+        case PdfDataType::Dictionary:
+        {
+            m_Dictionary.Value->Write(device, writeMode, encrypt, buffer);
+            break;
+        }
+        case PdfDataType::RawData:
+        {
+            m_Data.Value->Write(device, writeMode, encrypt, buffer);
+            break;
+        }
         case PdfDataType::Null:
         {
             if ((writeMode & PdfWriteFlags::NoInlineLiteral) == PdfWriteFlags::None)
@@ -206,127 +201,207 @@ void PdfVariant::Write(OutputStream& device, PdfWriteFlags writeMode,
     };
 }
 
-string PdfVariant::ToString() const
+string PdfVariant::ToString(PdfWriteFlags writeFlags) const
 {
     string ret;
-    ToString(ret);
+    ToString(ret, writeFlags);
     return ret;
 }
 
-void PdfVariant::ToString(string& str) const
+void PdfVariant::ToString(string& str, PdfWriteFlags writeFlags) const
 {
     str.clear();
-    PdfWriteFlags writeFlags;
-    switch (m_DataType)
+    switch (GetDataType())
     {
         case PdfDataType::Null:
         case PdfDataType::Bool:
         case PdfDataType::Number:
         case PdfDataType::Real:
         case PdfDataType::Reference:
-            // By default we want the literals to not be spaced
-            writeFlags = PdfWriteFlags::NoInlineLiteral;
+            // We enforce the literals to not be spaced
+            writeFlags |= PdfWriteFlags::NoInlineLiteral;
             break;
         default:
-            writeFlags = PdfWriteFlags::None;
+            // Do nothing
             break;
     }
 
     charbuff buffer;
     StringStreamDevice device(str);
-    this->Write(device, writeFlags, PdfStatefulEncrypt(), buffer);
+    this->Write(device, writeFlags, nullptr, buffer);
 }
 
 PdfVariant& PdfVariant::operator=(const PdfVariant& rhs)
 {
-    clear();
+    this->~PdfVariant();
     assign(rhs);
     return *this;
 }
 
 PdfVariant& PdfVariant::operator=(PdfVariant&& rhs) noexcept
 {
-    clear();
-    m_DataType = rhs.m_DataType;
-    m_Data = rhs.m_Data;
-    rhs.m_DataType = PdfDataType::Null;
-    rhs.m_Data = { };
+    this->~PdfVariant();
+    moveFrom(std::move(rhs));
     return *this;
 }
 
 void PdfVariant::assign(const PdfVariant& rhs)
 {
-    m_DataType = rhs.m_DataType;
-    switch (m_DataType)
+    switch (rhs.GetDataType())
     {
         case PdfDataType::Array:
         {
-            m_Data.Data = new PdfArray(*static_cast<const PdfArray*>(rhs.m_Data.Data));
+            new(&m_Array)PrimitiveMember(new PdfArray(*rhs.m_Array.Value));
             break;
         }
         case PdfDataType::Dictionary:
         {
-            m_Data.Data = new PdfDictionary(*static_cast<const PdfDictionary*>(rhs.m_Data.Data));
+            new(&m_Dictionary)PrimitiveMember(new PdfDictionary(*rhs.m_Dictionary.Value));
+            break;
+        }
+        case PdfDataType::RawData:
+        {
+            new(&m_Data)PrimitiveMember(new PdfData(*rhs.m_Data.Value));
             break;
         }
         case PdfDataType::Name:
         {
-            m_Data.Data = new PdfName(*static_cast<const PdfName*>(rhs.m_Data.Data));
+            new(&m_Name)PdfName(rhs.m_Name);
             break;
         }
         case PdfDataType::String:
         {
-            m_Data.Data = new PdfString(*static_cast<const PdfString*>(rhs.m_Data.Data));
-            break;
-        }
-
-        case PdfDataType::RawData:
-        {
-            m_Data.Data = new PdfData((*static_cast<const PdfData*>(rhs.m_Data.Data)));
+            new(&m_String)PdfString(rhs.m_String);
             break;
         }
         case PdfDataType::Reference:
-        case PdfDataType::Bool:
-        case PdfDataType::Null:
-        case PdfDataType::Number:
-        case PdfDataType::Real:
-            m_Data = rhs.m_Data;
+        {
+            new(&m_Reference)PdfReference(rhs.m_Reference);
             break;
-
+        }
+        case PdfDataType::Bool:
+        {
+            new(&m_Bool)PrimitiveMember(rhs.m_Bool.Value);
+            break;
+        }
+        case PdfDataType::Number:
+        {
+            new(&m_Number)PrimitiveMember(rhs.m_Number.Value);
+            break;
+        }
+        case PdfDataType::Real:
+        {
+            new(&m_Real)PrimitiveMember(rhs.m_Real.Value);
+            break;
+        }
+        case PdfDataType::Null:
+        {
+            new(&m_Null)NullMember();
+            break;
+        }
         case PdfDataType::Unknown:
         default:
-            break;
-    };
+            PODOFO_RAISE_ERROR(PdfErrorCode::InvalidEnumValue);
+    }
 }
 
-const char* PdfVariant::GetDataTypeString() const
+void PdfVariant::moveFrom(PdfVariant&& rhs)
 {
-    switch (m_DataType)
+    switch (rhs.GetDataType())
+    {
+        case PdfDataType::Array:
+        {
+            new(&m_Array)PrimitiveMember(rhs.m_Array);
+            break;
+        }
+        case PdfDataType::Dictionary:
+        {
+            new(&m_Dictionary)PrimitiveMember(rhs.m_Dictionary);
+            break;
+        }
+        case PdfDataType::RawData:
+        {
+            new(&m_Data)PrimitiveMember(rhs.m_Data);
+            break;
+        }
+        case PdfDataType::Name:
+        {
+            new(&m_Name)PdfName(std::move(rhs.m_Name));
+            break;
+        }
+        case PdfDataType::String:
+        {
+            new(&m_String)PdfString(std::move(rhs.m_String));
+            break;
+        }
+        case PdfDataType::Reference:
+        {
+            new(&m_Reference)PdfReference(rhs.m_Reference);
+            break;
+        }
+        case PdfDataType::Bool:
+        {
+            new(&m_Bool)PrimitiveMember(rhs.m_Bool.Value);
+            break;
+        }
+        case PdfDataType::Number:
+        {
+            new(&m_Number)PrimitiveMember(rhs.m_Number.Value);
+            break;
+        }
+        case PdfDataType::Real:
+        {
+            new(&m_Real)PrimitiveMember(rhs.m_Real.Value);
+            break;
+        }
+        case PdfDataType::Null:
+        {
+            new(&m_Null)NullMember();
+            break;
+        }
+        case PdfDataType::Unknown:
+        default:
+            PODOFO_RAISE_ERROR(PdfErrorCode::InvalidEnumValue);
+    }
+
+    // Reset rhs to null type
+    new(&rhs.m_Null)NullMember();
+}
+
+void PdfVariant::Reset()
+{
+    this->~PdfVariant();
+    new(&m_Null)NullMember();
+}
+
+string_view PdfVariant::GetDataTypeString() const
+{
+    switch (GetDataType())
     {
         case PdfDataType::Bool:
-            return "Bool";
+            return "Bool"sv;
         case PdfDataType::Number:
-            return "Number";
+            return "Number"sv;
         case PdfDataType::Real:
-            return "Real";
+            return "Real"sv;
         case PdfDataType::String:
-            return "String";
+            return "String"sv;
         case PdfDataType::Name:
-            return "Name";
+            return "Name"sv;
         case PdfDataType::Array:
-            return "Array";
+            return "Array"sv;
         case PdfDataType::Dictionary:
-            return "Dictionary";
+            return "Dictionary"sv;
         case PdfDataType::Null:
-            return "Null";
+            return "Null"sv;
         case PdfDataType::Reference:
-            return "Reference";
+            return "Reference"sv;
         case PdfDataType::RawData:
-            return "RawData";
+            return "RawData"sv;
         case PdfDataType::Unknown:
-            return "Unknown";
+            return "Unknown"sv;
         default:
-            return "INVALID_TYPE_ENUM";
+            PODOFO_RAISE_ERROR(PdfErrorCode::InvalidEnumValue);
     }
 }
 
@@ -335,13 +410,13 @@ bool PdfVariant::operator==(const PdfVariant& rhs) const
     if (this == &rhs)
         return true;
 
-    switch (m_DataType)
+    switch (GetDataType())
     {
         case PdfDataType::Bool:
         {
             bool value;
             if (rhs.TryGetBool(value))
-                return m_Data.Bool == value;
+                return m_Bool.Value == value;
             else
                 return false;
         }
@@ -349,7 +424,7 @@ bool PdfVariant::operator==(const PdfVariant& rhs) const
         {
             int64_t value;
             if (rhs.TryGetNumber(value))
-                return m_Data.Number == value;
+                return m_Number.Value == value;
             else
                 return false;
         }
@@ -358,7 +433,7 @@ bool PdfVariant::operator==(const PdfVariant& rhs) const
             // NOTE: Real type equality semantics is strict
             double value;
             if (rhs.TryGetRealStrict(value))
-                return m_Data.Real == value;
+                return m_Real.Value == value;
             else
                 return false;
         }
@@ -366,7 +441,7 @@ bool PdfVariant::operator==(const PdfVariant& rhs) const
         {
             PdfReference value;
             if (rhs.TryGetReference(value))
-                return m_Data.Reference == value;
+                return m_Reference == value;
             else
                 return false;
         }
@@ -374,7 +449,7 @@ bool PdfVariant::operator==(const PdfVariant& rhs) const
         {
             const PdfString* value;
             if (rhs.tryGetString(value))
-                return *(PdfString*)m_Data.Data == *value;
+                return m_String == *value;
             else
                 return false;
         }
@@ -382,7 +457,7 @@ bool PdfVariant::operator==(const PdfVariant& rhs) const
         {
             const PdfName* value;
             if (rhs.tryGetName(value))
-                return *(PdfName*)m_Data.Data == *value;
+                return m_Name == *value;
             else
                 return false;
         }
@@ -390,7 +465,7 @@ bool PdfVariant::operator==(const PdfVariant& rhs) const
         {
             const PdfArray* value;
             if (rhs.TryGetArray(value))
-                return *(PdfArray*)m_Data.Data == *value;
+                return *m_Array.Value == *value;
             else
                 return false;
         }
@@ -398,18 +473,17 @@ bool PdfVariant::operator==(const PdfVariant& rhs) const
         {
             const PdfDictionary* value;
             if (rhs.TryGetDictionary(value))
-                return *(PdfDictionary*)m_Data.Data == *value;
+                return *m_Dictionary.Value == *value;
             else
                 return false;
         }
         case PdfDataType::RawData:
             PODOFO_RAISE_ERROR_INFO(PdfErrorCode::NotImplemented, "Equality not yet implemented for RawData");
         case PdfDataType::Null:
-            return m_DataType == PdfDataType::Null;
+            return rhs.GetDataType() == PdfDataType::Null;
         case PdfDataType::Unknown:
-            return false;
         default:
-            PODOFO_RAISE_ERROR(PdfErrorCode::NotImplemented);
+            PODOFO_RAISE_ERROR(PdfErrorCode::InvalidEnumValue);
     }
 }
 
@@ -418,13 +492,13 @@ bool PdfVariant::operator!=(const PdfVariant& rhs) const
     if (this != &rhs)
         return true;
 
-    switch (m_DataType)
+    switch (GetDataType())
     {
         case PdfDataType::Bool:
         {
             bool value;
             if (rhs.TryGetBool(value))
-                return m_Data.Bool != value;
+                return m_Bool.Value != value;
             else
                 return true;
         }
@@ -432,7 +506,7 @@ bool PdfVariant::operator!=(const PdfVariant& rhs) const
         {
             int64_t value;
             if (rhs.TryGetNumber(value))
-                return m_Data.Number != value;
+                return m_Number.Value != value;
             else
                 return true;
         }
@@ -441,7 +515,7 @@ bool PdfVariant::operator!=(const PdfVariant& rhs) const
             // NOTE: Real type equality semantics is strict
             double value;
             if (rhs.TryGetRealStrict(value))
-                return m_Data.Real != value;
+                return m_Real.Value != value;
             else
                 return true;
         }
@@ -449,7 +523,7 @@ bool PdfVariant::operator!=(const PdfVariant& rhs) const
         {
             PdfReference value;
             if (rhs.TryGetReference(value))
-                return m_Data.Reference != value;
+                return m_Reference != value;
             else
                 return true;
         }
@@ -457,7 +531,7 @@ bool PdfVariant::operator!=(const PdfVariant& rhs) const
         {
             const PdfString* value;
             if (rhs.tryGetString(value))
-                return *(PdfString*)m_Data.Data != *value;
+                return m_String != *value;
             else
                 return true;
         }
@@ -465,7 +539,7 @@ bool PdfVariant::operator!=(const PdfVariant& rhs) const
         {
             const PdfName* value;
             if (rhs.tryGetName(value))
-                return *(PdfName*)m_Data.Data != *value;
+                return m_Name != *value;
             else
                 return true;
         }
@@ -473,7 +547,7 @@ bool PdfVariant::operator!=(const PdfVariant& rhs) const
         {
             const PdfArray* value;
             if (rhs.TryGetArray(value))
-                return *(PdfArray*)m_Data.Data != *value;
+                return *m_Array.Value != *value;
             else
                 return true;
         }
@@ -481,21 +555,25 @@ bool PdfVariant::operator!=(const PdfVariant& rhs) const
         {
             const PdfDictionary* value;
             if (rhs.TryGetDictionary(value))
-                return *(PdfDictionary*)m_Data.Data != *value;
+                return *m_Dictionary.Value != *value;
             else
                 return true;
         }
         case PdfDataType::RawData:
             PODOFO_RAISE_ERROR_INFO(PdfErrorCode::NotImplemented, "Disequality not yet implemented for RawData");
         case PdfDataType::Null:
-            return m_DataType != PdfDataType::Null;
+            return rhs.GetDataType() != PdfDataType::Null;
         case PdfDataType::Unknown:
-            return true;
         default:
-            PODOFO_RAISE_ERROR(PdfErrorCode::NotImplemented);
+            PODOFO_RAISE_ERROR(PdfErrorCode::InvalidEnumValue);
     }
 }
 
+PdfDataType PdfVariant::GetDataType() const
+{
+    return reinterpret_cast<const PdfDataMember&>(*this).GetDataType();
+}
+
 bool PdfVariant::GetBool() const
 {
     bool ret;
@@ -507,13 +585,13 @@ bool PdfVariant::GetBool() const
 
 bool PdfVariant::TryGetBool(bool& value) const
 {
-    if (m_DataType != PdfDataType::Bool)
+    if (GetDataType() != PdfDataType::Bool)
     {
         value = false;
         return false;
     }
 
-    value = m_Data.Bool;
+    value = m_Bool.Value;
     return true;
 }
 
@@ -528,19 +606,22 @@ int64_t PdfVariant::GetNumberLenient() const
 
 bool PdfVariant::TryGetNumberLenient(int64_t& value) const
 {
-    if (!(m_DataType == PdfDataType::Number
-        || m_DataType == PdfDataType::Real))
+    auto type = GetDataType();
+    if (type == PdfDataType::Number)
+    {
+        value = m_Number.Value;
+        return true;
+    }
+    else if (type == PdfDataType::Real)
+    {
+        value = static_cast<int64_t>(std::round(m_Real.Value));
+        return true;
+    }
+    else
     {
         value = 0;
         return false;
     }
-
-    if (m_DataType == PdfDataType::Real)
-        value = static_cast<int64_t>(std::round(m_Data.Real));
-    else
-        value = m_Data.Number;
-
-    return true;
 }
 
 int64_t PdfVariant::GetNumber() const
@@ -549,18 +630,18 @@ int64_t PdfVariant::GetNumber() const
     if (!TryGetNumber(ret))
         PODOFO_RAISE_ERROR(PdfErrorCode::InvalidDataType);
 
-    return m_Data.Number;
+    return m_Number.Value;
 }
 
 bool PdfVariant::TryGetNumber(int64_t& value) const
 {
-    if (m_DataType != PdfDataType::Number)
+    if (GetDataType() != PdfDataType::Number)
     {
         value = 0;
         return false;
     }
 
-    value = m_Data.Number;
+    value = m_Number.Value;
     return true;
 }
 
@@ -575,19 +656,22 @@ double PdfVariant::GetReal() const
 
 bool PdfVariant::TryGetReal(double& value) const
 {
-    if (!(m_DataType == PdfDataType::Real
-        || m_DataType == PdfDataType::Number))
+    auto type = GetDataType();
+    if (type == PdfDataType::Real)
+    {
+        value = m_Real.Value;
+        return true;
+    }
+    else if (type == PdfDataType::Number)
+    {
+        value = static_cast<double>(m_Number.Value);
+        return true;
+    }
+    else
     {
         value = 0;
         return false;
     }
-
-    if (m_DataType == PdfDataType::Number)
-        value = static_cast<double>(m_Data.Number);
-    else
-        value = m_Data.Real;
-
-    return true;
 }
 
 double PdfVariant::GetRealStrict() const
@@ -596,18 +680,18 @@ double PdfVariant::GetRealStrict() const
     if (!TryGetRealStrict(ret))
         PODOFO_RAISE_ERROR(PdfErrorCode::InvalidDataType);
 
-    return m_Data.Real;
+    return m_Real.Value;
 }
 
 bool PdfVariant::TryGetRealStrict(double& value) const
 {
-    if (m_DataType != PdfDataType::Real)
+    if (GetDataType() != PdfDataType::Real)
     {
         value = 0;
         return false;
     }
 
-    value = m_Data.Real;
+    value = m_Real.Value;
     return true;
 }
 
@@ -640,13 +724,13 @@ bool PdfVariant::TryGetString(const PdfString*& str) const
 
 bool PdfVariant::tryGetString(const PdfString*& str) const
 {
-    if (m_DataType != PdfDataType::String)
+    if (GetDataType() != PdfDataType::String)
     {
         str = nullptr;
         return false;
     }
 
-    str = (PdfString*)m_Data.Data;
+    str = &m_String;
     return true;
 }
 
@@ -679,13 +763,13 @@ bool PdfVariant::TryGetName(const PdfName*& name) const
 
 bool PdfVariant::tryGetName(const PdfName*& name) const
 {
-    if (m_DataType != PdfDataType::Name)
+    if (GetDataType() != PdfDataType::Name)
     {
         name = nullptr;
         return false;
     }
 
-    name = (PdfName*)m_Data.Data;
+    name = &m_Name;
     return true;
 }
 
@@ -700,13 +784,13 @@ PdfReference PdfVariant::GetReference() const
 
 bool PdfVariant::TryGetReference(PdfReference& ref) const
 {
-    if (m_DataType != PdfDataType::Reference)
+    if (GetDataType() != PdfDataType::Reference)
     {
         ref = PdfReference();
         return false;
     }
 
-    ref = m_Data.Reference;
+    ref = m_Reference;
     return true;
 }
 
@@ -768,139 +852,158 @@ bool PdfVariant::TryGetDictionary(PdfDictionary*& dict)
 
 bool PdfVariant::tryGetDictionary(PdfDictionary*& dict) const
 {
-    if (m_DataType != PdfDataType::Dictionary)
+    if (GetDataType() != PdfDataType::Dictionary)
     {
         dict = nullptr;
         return false;
     }
 
-    dict = (PdfDictionary*)m_Data.Data;
+    dict = (PdfDictionary*)m_Dictionary.Value;
     return true;
 }
 
 bool PdfVariant::tryGetArray(PdfArray*& arr) const
 {
-    if (m_DataType != PdfDataType::Array)
+    if (GetDataType() != PdfDataType::Array)
     {
         arr = nullptr;
         return false;
     }
 
-    arr = (PdfArray*)m_Data.Data;
+    arr = (PdfArray*)m_Array.Value;
     return true;
 }
 
+PdfReference PdfVariant::GetReferenceUnsafe() const
+{
+    return m_Reference;
+}
+
+const PdfDictionary& PdfVariant::GetDictionaryUnsafe() const
+{
+    return *(const PdfDictionary*)m_Dictionary.Value;
+}
+
+const PdfArray& PdfVariant::GetArrayUnsafe() const
+{
+    return *(const PdfArray*)m_Array.Value;
+}
+
+PdfDictionary& PdfVariant::GetDictionaryUnsafe()
+{
+    return *(PdfDictionary*)m_Dictionary.Value;
+}
+
+PdfArray& PdfVariant::GetArrayUnsafe()
+{
+    return *(PdfArray*)m_Array.Value;
+}
+
 void PdfVariant::SetBool(bool value)
 {
-    if (m_DataType != PdfDataType::Bool)
+    if (GetDataType() != PdfDataType::Bool)
         PODOFO_RAISE_ERROR(PdfErrorCode::InvalidDataType);
 
-    m_Data.Bool = value;
+    m_Bool = value;
 }
 
 void PdfVariant::SetNumber(int64_t value)
 {
-    if (!(m_DataType == PdfDataType::Number
-        || m_DataType == PdfDataType::Real))
-    {
+    if (GetDataType() != PdfDataType::Number)
         PODOFO_RAISE_ERROR(PdfErrorCode::InvalidDataType);
-    }
 
-    if (m_DataType == PdfDataType::Real)
-        m_Data.Real = static_cast<double>(value);
-    else
-        m_Data.Number = value;
+    m_Number = value;
 }
 
 void PdfVariant::SetReal(double value)
 {
-    if (!(m_DataType == PdfDataType::Real
-        || m_DataType == PdfDataType::Number))
-    {
+    if (GetDataType() != PdfDataType::Real)
         PODOFO_RAISE_ERROR(PdfErrorCode::InvalidDataType);
-    }
 
-    if (m_DataType == PdfDataType::Number)
-        m_Data.Number = static_cast<int64_t>(std::round(value));
-    else
-        m_Data.Real = value;
+    m_Real = value;
 }
 
 void PdfVariant::SetName(const PdfName& name)
 {
-    if (m_DataType != PdfDataType::Name)
+    if (GetDataType() != PdfDataType::Name)
         PODOFO_RAISE_ERROR(PdfErrorCode::InvalidDataType);
 
-    *((PdfName*)m_Data.Data) = name;
+    m_Name = name;
 }
 
 void PdfVariant::SetString(const PdfString& str)
 {
-    if (m_DataType != PdfDataType::String)
+    if (GetDataType() != PdfDataType::String)
         PODOFO_RAISE_ERROR(PdfErrorCode::InvalidDataType);
 
-    *((PdfString*)m_Data.Data) = str;
+    m_String = str;
 }
 
 void PdfVariant::SetReference(const PdfReference& ref)
 {
-    if (m_DataType != PdfDataType::Reference)
+    if (GetDataType() != PdfDataType::Reference)
         PODOFO_RAISE_ERROR(PdfErrorCode::InvalidDataType);
 
-    m_Data.Reference = ref;
+    m_Reference = ref;
 }
 
 bool PdfVariant::IsBool() const
 {
-    return m_DataType == PdfDataType::Bool;
+    return GetDataType() == PdfDataType::Bool;
 }
 
 bool PdfVariant::IsNumber() const
 {
-    return m_DataType == PdfDataType::Number;
+    return GetDataType() == PdfDataType::Number;
 }
 
 bool PdfVariant::IsRealStrict() const
 {
-    return m_DataType == PdfDataType::Real;
+    return GetDataType() == PdfDataType::Real;
 }
 
 bool PdfVariant::IsNumberOrReal() const
 {
-    return m_DataType == PdfDataType::Number || m_DataType == PdfDataType::Real;
+    auto type = GetDataType();
+    return type == PdfDataType::Number || type == PdfDataType::Real;
 }
 
 bool PdfVariant::IsString() const
 {
-    return m_DataType == PdfDataType::String;
+    return GetDataType() == PdfDataType::String;
 }
 
 bool PdfVariant::IsName() const
 {
-    return m_DataType == PdfDataType::Name;
+    return GetDataType() == PdfDataType::Name;
 }
 
 bool PdfVariant::IsArray() const
 {
-    return m_DataType == PdfDataType::Array;
+    return GetDataType() == PdfDataType::Array;
 }
 
 bool PdfVariant::IsDictionary() const
 {
-    return m_DataType == PdfDataType::Dictionary;
+    return GetDataType() == PdfDataType::Dictionary;
 }
 
 bool PdfVariant::IsRawData() const
 {
-    return m_DataType == PdfDataType::RawData;
+    return GetDataType() == PdfDataType::RawData;
 }
 
 bool PdfVariant::IsNull() const
 {
-    return m_DataType == PdfDataType::Null;
+    return GetDataType() == PdfDataType::Null;
 }
 
 bool PdfVariant::IsReference() const
 {
-    return m_DataType == PdfDataType::Reference;
+    return GetDataType() == PdfDataType::Reference;
+}
+
+PdfVariant::NullMember::NullMember()
+    : PdfDataMember(PdfDataType::Null)
+{
 }
