@@ -17,6 +17,7 @@
 #include <cinttypes>
 #include <climits>
 
+#include <typeinfo>
 #include <stdexcept>
 #include <limits>
 #include <algorithm>
@@ -24,15 +25,28 @@
 
 #include "Format.h"
 #include "numbers_compat.h"
-
- // Macro to define friendship with test classes.
- // Must be defined before base declarations
-#define PODOFO_UNIT_TEST(classname) friend class classname
+#include "charconv_compat.h"
 
 #include <podofo/main/PdfDeclarations.h>
 
+// Redefine empty PODOFO_PRIVATE_FRIEND to specify actual
+// friendship with private identifiers (class or methods)
+#undef PODOFO_PRIVATE_FRIEND
+#define PODOFO_PRIVATE_FRIEND(identifier) friend identifier
+
+// Redefine PODOFO_3RDPARTY_API_ENABLED to enable 3rd party
+// API interoperability
+#undef PODOFO_3RDPARTY_INTEROP_ENABLED
+#define PODOFO_3RDPARTY_INTEROP_ENABLED 1
+
+#include <podofo/auxiliary/Rect.h>
+#include <podofo/optional/PdfConvert.h>
+#include <podofo/optional/PdfUtils.h>
+#include <podofo/main/PdfName.h>
+#include "PdfXRefEntry.h"
+
 #ifdef _WIN32
-// Microsft itself assumes little endian
+// Microsoft itself assumes little endian
 // https://github.com/microsoft/STL/blob/b11945b73fc1139d3cf1115907717813930cedbf/stl/inc/bit#L336
 #define PODOFO_IS_LITTLE_ENDIAN
 #else // Unix
@@ -88,7 +102,7 @@
 
 // This is a do nothing macro that can be used to define
 // an invariant property without actually checking for it,
-// not even in DEBUG build. It's user responsability to
+// not even in DEBUG build. It's user responsibility to
 // ensure it's actually satisfied
 #define PODOFO_INVARIANT(x)
 
@@ -115,13 +129,13 @@
    *
    * Add frame to error callastack
    */
-#define PODOFO_PUSH_FRAME(err) err.AddToCallStack(__FILE__, __LINE__)
+#define PODOFO_PUSH_FRAME(err) AddToCallStack(err, __FILE__, __LINE__, { })
 
    /** \def PODOFO_PUSH_FRAME_INFO(err, msg)
     *
     * Add frame to error callastack with msg information
     */
-#define PODOFO_PUSH_FRAME_INFO(err, msg, ...) err.AddToCallStack(__FILE__, __LINE__, COMMON_FORMAT(msg, ##__VA_ARGS__))
+#define PODOFO_PUSH_FRAME_INFO(err, msg, ...) AddToCallStack(err, __FILE__, __LINE__, COMMON_FORMAT(msg, ##__VA_ARGS__))
 
     /** \def PODOFO_PUSH_FRAME(err, msg)
      *
@@ -137,36 +151,26 @@ namespace PoDoFo
 {
     class OutputStream;
     class InputStream;
+    class PdfPage;
+    class PdfDictionaryElement;
 
     constexpr double DEG2RAD = std::numbers::pi / 180;
     constexpr double RAD2DEG = 180 / std::numbers::pi;
 
+    /**
+     * Transform the given raw rect accordingly to the page rotation
+     */
+    Rect TransformCornersPage(const Corners& rect, const PdfPage& page);
+
     PdfVersion GetPdfVersion(const std::string_view& str);
 
-    std::string_view GetPdfVersionName(PdfVersion version);
+    const PdfName& GetPdfVersionName(PdfVersion version);
 
-    /**
-     *  Convert a name into a colorspace enum.
-     *
-     *  \param name name representing a colorspace such as DeviceGray
-     *  \returns colorspace enum or PdfColorSpace_Unknown if name is unknown
-     */
-    PdfColorSpace NameToColorSpaceRaw(const std::string_view& name);
-
-    /*
-     *  Convert a colorspace enum value into a name such as DeviceRGB
-     *
-     *  \param colorSpace a colorspace
-     *  \returns a name
-     */
-    std::string_view ColorSpaceToNameRaw(PdfColorSpace colorSpace);
-
-    std::string_view AnnotationTypeToName(PdfAnnotationType type);
-    PdfAnnotationType NameToAnnotationType(const std::string_view& str);
+    bool IsAccessibiltyProfile(PdfALevel pdfaLevel);
 
     /** Normalize base font name, removing known bold/italic/subset prefixes/suffixes
      */
-    std::string NormalizeFontName(const std::string_view& fontName);
+    std::string ExtractBaseFontName(const std::string_view& fontName, bool skipTrimSubset = false);
 
     /** Extract base font name, removing known bold/italic/subset prefixes/suffixes
      * \returns normalized font name
@@ -174,6 +178,13 @@ namespace PoDoFo
     std::string ExtractFontHints(const std::string_view& fontName,
         bool& isItalic, bool& isBold);
 
+    /**
+     * Get the end index of a subset prefix (eg. "AAAAAA+")
+     */
+    unsigned char GetSubsetPrefixLength(const std::string_view& fontName);
+
+    void CreateObjectStructElement(PdfDictionaryElement& elem, PdfPage& page, const PdfName& elementType);
+
     std::vector<std::string> ToPdfKeywordsList(const std::string_view& str);
     std::string ToPdfKeywordsString(const cspan<std::string>&keywords);
 
@@ -194,6 +205,34 @@ namespace PoDoFo
     {
         LogMessage(logSeverity, COMMON_FORMAT(msg, args...));
     }
+
+    char XRefEntryTypeToChar(PdfXRefEntryType type);
+    PdfXRefEntryType XRefEntryTypeFromChar(char c);
+
+    void AddToCallStack(PdfError& err, std::string filepath, unsigned line, std::string information);
+
+    /** Get the operands count of the operator
+     * \returns count the number of operand, -1 means variadic number of operands
+     */
+    int GetOperandCount(PdfOperator op);
+
+    /** Get the operands count of the operator
+     * \param count the number of operand, -1 means variadic number of operands
+     */
+    bool TryGetOperandCount(PdfOperator op, int& count);
+
+    /** Helper type to serialize 3 byte integers
+     */
+    struct uint24_t final
+    {
+        uint24_t();
+
+        explicit uint24_t(unsigned value);
+
+        operator unsigned() const;
+    private:
+        uint8_t value[3];
+    };
 }
 
 /**
@@ -216,17 +255,22 @@ namespace utls
         RecursionGuard();
         ~RecursionGuard();
 
-        // set maximum recursion depth (set to 0 to disable recursion check)
-        static void SetMaxRecursionDepth(unsigned maxRecursionDepth);
-
-        static unsigned GetMaxRecursionDepth();
-
     private:
         void Enter();
         void Exit();
     };
 
-    void SerializeEncodedString(PoDoFo::OutputStream& stream, const std::string_view& encoded, bool wantHex);
+    // https://stackoverflow.com/a/14369745/213871
+    inline double RoundTo(double value, double precision)
+    {
+        return std::round(value / precision) * precision;
+    }
+
+    /** Normalize the coordinate so the first corner is left-bottom, and the second right-top
+     */
+    void NormalizeCoordinates(double& x1, double& y1, double& x2, double& y2);
+
+    void SerializeEncodedString(PoDoFo::OutputStream& stream, const std::string_view& encoded, bool wantHex, bool skipDelimiters = false);
 
     /**
      * Check if multiplying two numbers will overflow. This is crucial when calculating buffer sizes that are the product of two numbers/
@@ -236,6 +280,8 @@ namespace utls
 
     const std::locale& GetInvariantLocale();
 
+    std::string_view GetEnvironmentVariable(const std::string_view& name);
+
     bool IsValidUtf8String(const std::string_view& str);
 
     bool IsStringDelimiter(char32_t ch);
@@ -286,6 +332,8 @@ namespace utls
     // Write the char to the supplied buffer as hexadecimal code
     void WriteCharHexTo(char buf[2], char ch);
 
+    std::string GetCharHexString(const PoDoFo::bufferview& buff);
+
     // Append the unicode code point to a big endian encoded utf16 string
     void WriteUtf16BETo(std::u16string& str, char32_t codePoint);
 
@@ -317,10 +365,36 @@ namespace utls
 
     void FormatTo(std::string& str, double value, unsigned short precision);
 
+    template <typename T, typename = std::enable_if_t<std::is_integral_v<T>>>
+    inline bool TryParse(const std::string_view& str, T& val, int base = 10)
+    {
+        if (std::from_chars(str.data(), str.data() + str.size(), val, base).ec == std::errc())
+            return true;
+        else
+            return false;
+    }
+
+    template <typename T, typename = std::enable_if_t<std::is_floating_point_v<T>>>
+    inline bool TryParse(const std::string_view& str, T& val, std::chars_format fmt = std::chars_format::fixed)
+    {
+        if (std::from_chars(str.data(), str.data() + str.size(), val, fmt).ec == std::errc())
+            return true;
+        else
+            return false;
+    }
+
+    template <typename T, typename... Args>
+    std::array<T, sizeof...(Args) + 1> MakeArray(T&& first, Args&&... rest)
+    {
+        return { std::forward<T>(first), std::forward<Args>(rest)... };
+    }
+
     std::string ToLower(const std::string_view& str);
 
     std::string Trim(const std::string_view& str, char ch);
 
+    void Replace(std::string& str, const std::string_view& from, const std::string_view& to);
+
     // https://stackoverflow.com/a/38140932/213871
     inline void hash_combine(std::size_t& seed)
     {
@@ -341,6 +415,18 @@ namespace utls
     // Returns pow(2, size * 8) - 1
     unsigned GetCharCodeMaxValue(unsigned char codeSize);
 
+    // Use the FSS-UTF encoding (early name for UTF-8 variable byte encoding)
+    // https://www.unicode.org/L2/Historical/wg20-n193-fss-utf.pdf
+    unsigned FSSUTFEncode(unsigned codePoint);
+
+    struct FSSUTFRange
+    {
+        unsigned FirstCode;
+        unsigned LastCode;
+    };
+
+    std::vector<FSSUTFRange> GetFSSUTFRanges(unsigned char codeSize);
+
     template<typename T>
     void move(T& in, T& out)
     {
@@ -400,8 +486,8 @@ namespace utls
     size_t FileSize(const std::string_view& filename);
 
     void CopyTo(std::ostream& dst, std::istream& src);
-    void ReadTo(PoDoFo::charbuff& str, const std::string_view& filepath);
-    void ReadTo(PoDoFo::charbuff& str, std::istream& stream);
+    void ReadTo(PoDoFo::charbuff& str, const std::string_view& filepath, size_t maxReadSize = std::numeric_limits<size_t>::max());
+    void ReadTo(PoDoFo::charbuff& str, std::istream& stream, size_t maxReadSize = std::numeric_limits<size_t>::max());
     void WriteTo(const std::string_view& filepath, const PoDoFo::bufferview& view);
     void WriteTo(std::ostream& stream, const PoDoFo::bufferview& view);
 
@@ -420,24 +506,31 @@ namespace utls
     std::ofstream open_ofstream(const std::string_view& filename, std::ios_base::openmode mode);
     std::fstream open_fstream(const std::string_view& filename, std::ios_base::openmode mode);
 
-    // NOTE: Never use this function unless you really need a C FILE descriptor,
+    // NOTE: Never use this function unless you really want a C FILE descriptor,
     // as in PdfImage.cpp . For all the other I/O, use an STL stream
     FILE* fopen(const std::string_view& view, const std::string_view& mode);
 
+    ssize_t ftell(FILE* file);
+    ssize_t fseek(FILE* file, ssize_t offset, int origin);
+
     void WriteUInt32BE(PoDoFo::OutputStream& output, uint32_t value);
     void WriteInt32BE(PoDoFo::OutputStream& output, int32_t value);
+    void WriteUInt24BE(PoDoFo::OutputStream& output, PoDoFo::uint24_t value);
     void WriteUInt16BE(PoDoFo::OutputStream& output, uint16_t value);
     void WriteInt16BE(PoDoFo::OutputStream& output, int16_t value);
     void WriteUInt32BE(char* buf, uint32_t value);
+    void WriteUInt24BE(char* buf, PoDoFo::uint24_t value);
     void WriteInt32BE(char* buf, int32_t value);
     void WriteUInt16BE(char* buf, uint16_t value);
     void WriteInt16BE(char* buf, int16_t value);
     void ReadUInt32BE(PoDoFo::InputStream& input, uint32_t& value);
     void ReadInt32BE(PoDoFo::InputStream& input, int32_t& value);
+    void ReadUInt24BE(PoDoFo::InputStream& input, PoDoFo::uint24_t& value);
     void ReadUInt16BE(PoDoFo::InputStream& input, uint16_t& value);
     void ReadInt16BE(PoDoFo::InputStream& input, int16_t& value);
     void ReadUInt32BE(const char* buf, uint32_t& value);
     void ReadInt32BE(const char* buf, int32_t& value);
+    void ReadUInt24BE(const char* buf, PoDoFo::uint24_t& value);
     void ReadUInt16BE(const char* buf, uint16_t& value);
     void ReadInt16BE(const char* buf, int16_t& value);
 
@@ -508,8 +601,25 @@ namespace utls
         return (int64_t)__builtin_bswap64((uint64_t)n);
     }
 #endif
+
+    inline PoDoFo::uint24_t ByteSwap(PoDoFo::uint24_t n)
+    {
+        PoDoFo::uint24_t ret;
+        // NOTE: The following is safe as uint24_t is internally a uint8_t array
+        auto in = (const uint8_t*)&n;
+        auto out = (uint8_t*)&ret;
+        out[0] = in[2];
+        out[1] = in[1];
+        out[2] = in[0];
+        return ret;
+    }
 #pragma endregion // Byte Swap
 
+    // Normalize a page rotation to [0, 90, 180, 270]
+    int NormalizePageRotation(double angle);
+
+    // Normalize a value to the circular input range [start,end)
+    double NormalizeCircularRange(double value, double start, double end);
 }
 
 /**
