@@ -51,12 +51,12 @@ private:
     const char* m_colorName;
 };
 
-TEST_CASE("testDefaultConstructor")
+TEST_CASE("TestDefaultConstructor")
 {
     PdfColor color;
 
     REQUIRE(color.IsGrayScale());
-    REQUIRE(color.GetColorSpace() == PdfColorSpace::DeviceGray);
+    REQUIRE(color.GetColorSpace() == PdfColorSpaceType::DeviceGray);
     REQUIRE(color.GetGrayScale() == 0);
     REQUIRE(color.ConvertToGrayScale().GetGrayScale() == 0);
     REQUIRE(color.ConvertToRGB() == PdfColor(0, 0, 0));
@@ -68,12 +68,6 @@ TEST_CASE("testDefaultConstructor")
 
     REQUIRE(!color.IsRGB());
     REQUIRE(!color.IsCMYK());
-    REQUIRE(!color.IsSeparation());
-    REQUIRE(!color.IsCieLab());
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetAlternateColorSpace(),
-        PdfErrorCode::InternalLogic);
 
     ASSERT_THROW_WITH_ERROR_CODE(
         color.GetRed(),
@@ -102,29 +96,9 @@ TEST_CASE("testDefaultConstructor")
     ASSERT_THROW_WITH_ERROR_CODE(
         color.GetBlack(),
         PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetName(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetDensity(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCieL(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCieA(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCieB(),
-        PdfErrorCode::InternalLogic);
 }
 
-TEST_CASE("testGreyConstructor")
+TEST_CASE("TestGreyConstructor")
 {
     const double GRAY_VALUE = 0.123;
     PdfColor color(GRAY_VALUE);
@@ -132,13 +106,7 @@ TEST_CASE("testGreyConstructor")
     REQUIRE(color.IsGrayScale());
     REQUIRE(!color.IsRGB());
     REQUIRE(!color.IsCMYK());
-    REQUIRE(!color.IsSeparation());
-    REQUIRE(!color.IsCieLab());
-    REQUIRE(color.GetColorSpace() == PdfColorSpace::DeviceGray);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetAlternateColorSpace(),
-        PdfErrorCode::InternalLogic);
+    REQUIRE(color.GetColorSpace() == PdfColorSpaceType::DeviceGray);
 
     REQUIRE(color.GetGrayScale() == GRAY_VALUE);
 
@@ -170,26 +138,6 @@ TEST_CASE("testGreyConstructor")
         color.GetBlack(),
         PdfErrorCode::InternalLogic);
 
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetName(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetDensity(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCieL(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCieA(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCieB(),
-        PdfErrorCode::InternalLogic);
-
     REQUIRE(color.ConvertToGrayScale() == color);
 
     REQUIRE(color.ConvertToRGB() == PdfColor(GRAY_VALUE, GRAY_VALUE, GRAY_VALUE));
@@ -202,7 +150,7 @@ TEST_CASE("testGreyConstructor")
 
 }
 
-TEST_CASE("testGrayConstructorInvalid")
+TEST_CASE("TestGrayConstructorInvalid")
 {
     {
         const double GRAY_VALUE = 1.01;
@@ -219,7 +167,7 @@ TEST_CASE("testGrayConstructorInvalid")
     }
 }
 
-TEST_CASE("testRGBConstructor")
+TEST_CASE("TestRGBConstructor")
 {
     const double R_VALUE = 0.023;
     const double G_VALUE = 0.345;
@@ -229,13 +177,7 @@ TEST_CASE("testRGBConstructor")
     REQUIRE(!color.IsGrayScale());
     REQUIRE(color.IsRGB());
     REQUIRE(!color.IsCMYK());
-    REQUIRE(!color.IsSeparation());
-    REQUIRE(!color.IsCieLab());
-    REQUIRE(color.GetColorSpace() == PdfColorSpace::DeviceRGB);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetAlternateColorSpace(),
-        PdfErrorCode::InternalLogic);
+    REQUIRE(color.GetColorSpace() == PdfColorSpaceType::DeviceRGB);
 
     ASSERT_THROW_WITH_ERROR_CODE(
         color.GetGrayScale(),
@@ -261,26 +203,6 @@ TEST_CASE("testRGBConstructor")
         color.GetBlack(),
         PdfErrorCode::InternalLogic);
 
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetName(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetDensity(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCieL(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCieA(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCieB(),
-        PdfErrorCode::InternalLogic);
-
     REQUIRE(color.ConvertToGrayScale() == PdfColor(0.299 * R_VALUE + 0.587 * G_VALUE + 0.114 * B_VALUE));
 
     REQUIRE(color.ConvertToRGB() == PdfColor(R_VALUE, G_VALUE, B_VALUE));
@@ -301,7 +223,7 @@ TEST_CASE("testRGBConstructor")
     REQUIRE(COLOR_ARRAY[2] == PdfObject(B_VALUE));
 }
 
-TEST_CASE("testRGBConstructorInvalid")
+TEST_CASE("TestRGBConstructorInvalid")
 {
     {
         const double R_VALUE = 1.023;
@@ -359,7 +281,7 @@ TEST_CASE("testRGBConstructorInvalid")
 
 }
 
-TEST_CASE("testCMYKConstructor")
+TEST_CASE("TestCMYKConstructor")
 {
     const double C_VALUE = 0.1;
     const double M_VALUE = 0.2;
@@ -370,13 +292,7 @@ TEST_CASE("testCMYKConstructor")
     REQUIRE(!color.IsGrayScale());
     REQUIRE(!color.IsRGB());
     REQUIRE(color.IsCMYK());
-    REQUIRE(!color.IsSeparation());
-    REQUIRE(!color.IsCieLab());
-    REQUIRE(color.GetColorSpace() == PdfColorSpace::DeviceCMYK);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetAlternateColorSpace(),
-        PdfErrorCode::InternalLogic);
+    REQUIRE(color.GetColorSpace() == PdfColorSpaceType::DeviceCMYK);
 
     ASSERT_THROW_WITH_ERROR_CODE(
         color.GetGrayScale(),
@@ -399,26 +315,6 @@ TEST_CASE("testCMYKConstructor")
     REQUIRE(color.GetYellow() == Y_VALUE);
     REQUIRE(color.GetBlack() == B_VALUE);
 
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetName(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetDensity(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCieL(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCieA(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCieB(),
-        PdfErrorCode::InternalLogic);
-
     REQUIRE(color.ConvertToRGB().ConvertToGrayScale() == color.ConvertToGrayScale());
 
     {
@@ -426,7 +322,12 @@ TEST_CASE("testCMYKConstructor")
         double dGreen = M_VALUE * (1.0 - B_VALUE) + B_VALUE;
         double dBlue = Y_VALUE * (1.0 - B_VALUE) + B_VALUE;
 
-        REQUIRE(color.ConvertToRGB() == PdfColor(1.0 - dRed, 1.0 - dGreen, 1.0 - dBlue));
+        auto rgb1 = color.ConvertToRGB();
+        PdfColor rgb2(1.0 - dRed, 1.0 - dGreen, 1.0 - dBlue);
+
+        ASSERT_EQUAL(rgb1.GetRed(), rgb2.GetRed());
+        ASSERT_EQUAL(rgb1.GetGreen(), rgb2.GetGreen());
+        ASSERT_EQUAL(rgb1.GetBlue(), rgb2.GetBlue());
     }
 
     REQUIRE(color.ConvertToCMYK() == PdfColor(C_VALUE, M_VALUE, Y_VALUE, B_VALUE));
@@ -439,7 +340,7 @@ TEST_CASE("testCMYKConstructor")
     REQUIRE(COLOR_ARRAY[3] == PdfObject(B_VALUE));
 }
 
-TEST_CASE("testCMYKConstructorInvalid")
+TEST_CASE("TestCMYKConstructorInvalid")
 {
     {
         const double C_VALUE = 1.1;
@@ -523,7 +424,7 @@ TEST_CASE("testCMYKConstructorInvalid")
 
 }
 
-TEST_CASE("testCopyConstructor")
+TEST_CASE("TestCopyConstructor")
 {
     {
         const double GRAY_VALUE = 0.123;
@@ -533,9 +434,7 @@ TEST_CASE("testCopyConstructor")
         REQUIRE(color.IsGrayScale());
         REQUIRE(!color.IsRGB());
         REQUIRE(!color.IsCMYK());
-        REQUIRE(!color.IsSeparation());
-        REQUIRE(!color.IsCieLab());
-        REQUIRE(color.GetColorSpace() == PdfColorSpace::DeviceGray);
+        REQUIRE(color.GetColorSpace() == PdfColorSpaceType::DeviceGray);
 
         REQUIRE(color.GetGrayScale() == GRAY_VALUE);
     }
@@ -550,9 +449,7 @@ TEST_CASE("testCopyConstructor")
         REQUIRE(!color.IsGrayScale());
         REQUIRE(color.IsRGB());
         REQUIRE(!color.IsCMYK());
-        REQUIRE(!color.IsSeparation());
-        REQUIRE(!color.IsCieLab());
-        REQUIRE(color.GetColorSpace() == PdfColorSpace::DeviceRGB);
+        REQUIRE(color.GetColorSpace() == PdfColorSpaceType::DeviceRGB);
 
         REQUIRE(color.GetRed() == R_VALUE);
         REQUIRE(color.GetGreen() == G_VALUE);
@@ -570,9 +467,7 @@ TEST_CASE("testCopyConstructor")
         REQUIRE(!color.IsGrayScale());
         REQUIRE(!color.IsRGB());
         REQUIRE(color.IsCMYK());
-        REQUIRE(!color.IsSeparation());
-        REQUIRE(!color.IsCieLab());
-        REQUIRE(color.GetColorSpace() == PdfColorSpace::DeviceCMYK);
+        REQUIRE(color.GetColorSpace() == PdfColorSpaceType::DeviceCMYK);
 
         REQUIRE(color.GetCyan() == C_VALUE);
         REQUIRE(color.GetMagenta() == M_VALUE);
@@ -581,7 +476,7 @@ TEST_CASE("testCopyConstructor")
     }
 }
 
-TEST_CASE("testAssignmentOperator")
+TEST_CASE("TestAssignmentOperator")
 {
     {
         const double GRAY_VALUE = 0.123;
@@ -592,9 +487,7 @@ TEST_CASE("testAssignmentOperator")
         REQUIRE(color.IsGrayScale());
         REQUIRE(!color.IsRGB());
         REQUIRE(!color.IsCMYK());
-        REQUIRE(!color.IsSeparation());
-        REQUIRE(!color.IsCieLab());
-        REQUIRE(color.GetColorSpace() == PdfColorSpace::DeviceGray);
+        REQUIRE(color.GetColorSpace() == PdfColorSpaceType::DeviceGray);
 
         REQUIRE(color.GetGrayScale() == GRAY_VALUE);
     }
@@ -610,9 +503,7 @@ TEST_CASE("testAssignmentOperator")
         REQUIRE(!color.IsGrayScale());
         REQUIRE(color.IsRGB());
         REQUIRE(!color.IsCMYK());
-        REQUIRE(!color.IsSeparation());
-        REQUIRE(!color.IsCieLab());
-        REQUIRE(color.GetColorSpace() == PdfColorSpace::DeviceRGB);
+        REQUIRE(color.GetColorSpace() == PdfColorSpaceType::DeviceRGB);
 
         REQUIRE(color.GetRed() == R_VALUE);
         REQUIRE(color.GetGreen() == G_VALUE);
@@ -631,9 +522,7 @@ TEST_CASE("testAssignmentOperator")
         REQUIRE(!color.IsGrayScale());
         REQUIRE(!color.IsRGB());
         REQUIRE(color.IsCMYK());
-        REQUIRE(!color.IsSeparation());
-        REQUIRE(!color.IsCieLab());
-        REQUIRE(color.GetColorSpace() == PdfColorSpace::DeviceCMYK);
+        REQUIRE(color.GetColorSpace() == PdfColorSpaceType::DeviceCMYK);
 
         REQUIRE(color.GetCyan() == C_VALUE);
         REQUIRE(color.GetMagenta() == M_VALUE);
@@ -642,7 +531,7 @@ TEST_CASE("testAssignmentOperator")
     }
 }
 
-TEST_CASE("testEqualsOperator")
+TEST_CASE("TestEqualsOperator")
 {
     //Grey test
     { //Positive
@@ -826,10 +715,10 @@ TEST_CASE("testEqualsOperator")
 
 }
 
-TEST_CASE("testHexNames")
+TEST_CASE("TestHexNames")
 {
     {
-        PdfColor rgb = PdfColor::FromString("#FF0AEF");
+        PdfColor rgb = PdfColor::CreateFromString("#FF0AEF");
         REQUIRE(rgb.IsRGB());
         REQUIRE(static_cast<int>(rgb.GetRed() * 255.0) == 0xFF);
         REQUIRE(static_cast<int>(rgb.GetGreen() * 255.0) == 0x0A);
@@ -837,7 +726,7 @@ TEST_CASE("testHexNames")
     }
 
     {
-        PdfColor rgb = PdfColor::FromString("#012345");
+        PdfColor rgb = PdfColor::CreateFromString("#012345");
         REQUIRE(rgb.IsRGB());
         REQUIRE(static_cast<int>(rgb.GetRed() * 255.0) == 0x01);
         REQUIRE(static_cast<int>(rgb.GetGreen() * 255.0) == 0x23);
@@ -845,7 +734,7 @@ TEST_CASE("testHexNames")
     }
 
     {
-        PdfColor rgb = PdfColor::FromString("#ABCDEF");
+        PdfColor rgb = PdfColor::CreateFromString("#ABCDEF");
         REQUIRE(rgb.IsRGB());
         REQUIRE(static_cast<int>(rgb.GetRed() * 255.0) == 0xAB);
         REQUIRE(static_cast<int>(rgb.GetGreen() * 255.0) == 0xCD);
@@ -853,7 +742,7 @@ TEST_CASE("testHexNames")
     }
 
     {
-        PdfColor rgb = PdfColor::FromString("#abcdef");
+        PdfColor rgb = PdfColor::CreateFromString("#abcdef");
         REQUIRE(rgb.IsRGB());
         REQUIRE(static_cast<int>(rgb.GetRed() * 255.0) == 0xAB);
         REQUIRE(static_cast<int>(rgb.GetGreen() * 255.0) == 0xCD);
@@ -861,28 +750,28 @@ TEST_CASE("testHexNames")
     }
 
     {
-        PdfColor invalidColour = PdfColor::FromString("#01");
+        PdfColor invalidColour = PdfColor::CreateFromString("#01");
         REQUIRE(invalidColour == PdfColor());
     }
 
     {
-        PdfColor invalidColour = PdfColor::FromString("#123456789");
+        PdfColor invalidColour = PdfColor::CreateFromString("#123456789");
         REQUIRE(invalidColour == PdfColor());
     }
 
     {
         ASSERT_THROW_WITH_ERROR_CODE(
-            PdfColor::FromString("#12345g"),
+            PdfColor::CreateFromString("#12345g"),
             PdfErrorCode::CannotConvertColor);
     }
 
     {
         ASSERT_THROW_WITH_ERROR_CODE(
-            PdfColor::FromString("#1234g5"),
+            PdfColor::CreateFromString("#1234g5"),
             PdfErrorCode::CannotConvertColor);
     }
 
-    PdfColor cmyk = PdfColor::FromString("#ABCDEF01");
+    PdfColor cmyk = PdfColor::CreateFromString("#ABCDEF01");
     REQUIRE(cmyk.IsCMYK());
     REQUIRE(static_cast<int>(cmyk.GetCyan() * 255.0) == 0xAB);
     REQUIRE(static_cast<int>(cmyk.GetMagenta() * 255.0) == 0xCD);
@@ -890,58 +779,58 @@ TEST_CASE("testHexNames")
     REQUIRE(static_cast<int>(cmyk.GetBlack() * 255.0) == 0x01);
 }
 
-TEST_CASE("testNamesGeneral")
+TEST_CASE("TestNamesGeneral")
 {
-    PdfColor aliceBlue = PdfColor::FromString("aliceblue");
-    REQUIRE(aliceBlue == PdfColor::FromString("#F0F8FF"));
+    PdfColor aliceBlue = PdfColor::CreateFromString("aliceblue");
+    REQUIRE(aliceBlue == PdfColor::CreateFromString("#F0F8FF"));
     REQUIRE(aliceBlue.GetRed() == static_cast<double>(0xF0) / 255.0);
     REQUIRE(aliceBlue.GetGreen() == static_cast<double>(0xF8) / 255.0);
     REQUIRE(aliceBlue.GetBlue() == static_cast<double>(0xFF) / 255.0);
 
-    PdfColor lime = PdfColor::FromString("lime");
+    PdfColor lime = PdfColor::CreateFromString("lime");
     REQUIRE(lime == PdfColor(0.000, 1.000, 0.000));
 
-    PdfColor yellowGreen = PdfColor::FromString("yellowgreen");
-    REQUIRE(yellowGreen == PdfColor::FromString("#9ACD32"));
+    PdfColor yellowGreen = PdfColor::CreateFromString("yellowgreen");
+    REQUIRE(yellowGreen == PdfColor::CreateFromString("#9ACD32"));
 
     {
         // Test a not existing color
-        PdfColor notExist = PdfColor::FromString("asfaf9q341");
+        PdfColor notExist = PdfColor::CreateFromString("asfaf9q341");
         REQUIRE(notExist == PdfColor());
     }
 
     {
         // Test a not existing color
-        PdfColor notExist = PdfColor::FromString("A");
+        PdfColor notExist = PdfColor::CreateFromString("A");
         REQUIRE(notExist == PdfColor());
     }
 
     {
         // Test a not existing color
-        PdfColor notExist = PdfColor::FromString("");
+        PdfColor notExist = PdfColor::CreateFromString("");
         REQUIRE(notExist == PdfColor());
     }
 
     {
         // Test a not existing color
-        PdfColor notExist = PdfColor::FromString("yellowgree");
+        PdfColor notExist = PdfColor::CreateFromString("yellowgree");
         REQUIRE(notExist == PdfColor());
     }
 
     {
         // Test a not existing color
-        PdfColor notExist = PdfColor::FromString("yellowgreem");
+        PdfColor notExist = PdfColor::CreateFromString("yellowgreem");
         REQUIRE(notExist == PdfColor());
     }
 
     {
         // Test a not existing color
-        PdfColor notExist = PdfColor::FromString("yellowgreen ");
+        PdfColor notExist = PdfColor::CreateFromString("yellowgreen ");
         REQUIRE(notExist == PdfColor());
     }
 }
 
-TEST_CASE("testNamesOneByOne")
+TEST_CASE("TestNamesOneByOne")
 {
     //Copied and adjusted from http://cvsweb.xfree86.org/cvsweb/xc/programs/rgb/rgb.txt?rev=1.2
     const TestColor TABLE_OF_TEST_COLORS[] =
@@ -1611,7 +1500,7 @@ TEST_CASE("testNamesOneByOne")
     {
         const TestColor& TEST_COLOR(TABLE_OF_TEST_COLORS[i]);
 
-        const PdfColor COLOR_FROM_NAME(PdfColor::FromString(TEST_COLOR.getColorName()));
+        const PdfColor COLOR_FROM_NAME(PdfColor::CreateFromString(TEST_COLOR.getColorName()));
         const PdfColor EXPECTED_COLOR(TEST_COLOR.getR(), TEST_COLOR.getG(), TEST_COLOR.getB());
 
         if (PdfColor() == COLOR_FROM_NAME)
@@ -1626,7 +1515,7 @@ TEST_CASE("testNamesOneByOne")
     }
 }
 
-TEST_CASE("testColorGreyConstructor")
+TEST_CASE("TestColorGreyConstructor")
 {
     const double GRAY_VALUE = 0.123;
     PdfColor color(GRAY_VALUE);
@@ -1634,14 +1523,12 @@ TEST_CASE("testColorGreyConstructor")
     REQUIRE(color.IsGrayScale());
     REQUIRE(!color.IsRGB());
     REQUIRE(!color.IsCMYK());
-    REQUIRE(!color.IsSeparation());
-    REQUIRE(!color.IsCieLab());
-    REQUIRE(color.GetColorSpace() == PdfColorSpace::DeviceGray);
+    REQUIRE(color.GetColorSpace() == PdfColorSpaceType::DeviceGray);
 
     REQUIRE(color.GetGrayScale() == GRAY_VALUE);
 }
 
-TEST_CASE("testColorRGBConstructor")
+TEST_CASE("TestColorRGBConstructor")
 {
     const double R_VALUE = 0.023;
     const double G_VALUE = 0.345;
@@ -1651,16 +1538,14 @@ TEST_CASE("testColorRGBConstructor")
     REQUIRE(!color.IsGrayScale());
     REQUIRE(color.IsRGB());
     REQUIRE(!color.IsCMYK());
-    REQUIRE(!color.IsSeparation());
-    REQUIRE(!color.IsCieLab());
-    REQUIRE(color.GetColorSpace() == PdfColorSpace::DeviceRGB);
+    REQUIRE(color.GetColorSpace() == PdfColorSpaceType::DeviceRGB);
 
     REQUIRE(color.GetRed() == R_VALUE);
     REQUIRE(color.GetGreen() == G_VALUE);
     REQUIRE(color.GetBlue() == B_VALUE);
 }
 
-TEST_CASE("testColorCMYKConstructor")
+TEST_CASE("TestColorCMYKConstructor")
 {
     const double C_VALUE = 0.1;
     const double M_VALUE = 0.2;
@@ -1671,9 +1556,7 @@ TEST_CASE("testColorCMYKConstructor")
     REQUIRE(!color.IsGrayScale());
     REQUIRE(!color.IsRGB());
     REQUIRE(color.IsCMYK());
-    REQUIRE(!color.IsSeparation());
-    REQUIRE(!color.IsCieLab());
-    REQUIRE(color.GetColorSpace() == PdfColorSpace::DeviceCMYK);
+    REQUIRE(color.GetColorSpace() == PdfColorSpaceType::DeviceCMYK);
 
     REQUIRE(color.GetCyan() == C_VALUE);
     REQUIRE(color.GetMagenta() == M_VALUE);
@@ -1681,496 +1564,7 @@ TEST_CASE("testColorCMYKConstructor")
     REQUIRE(color.GetBlack() == B_VALUE);
 }
 
-TEST_CASE("testColorSeparationAllConstructor")
-{
-    auto color = PdfColor::CreateSeparationAll();
-
-    REQUIRE(!color.IsGrayScale());
-    REQUIRE(!color.IsRGB());
-    REQUIRE(!color.IsCMYK());
-    REQUIRE(color.IsSeparation());
-    REQUIRE(!color.IsCieLab());
-    REQUIRE(color.GetColorSpace() == PdfColorSpace::Separation);
-    REQUIRE(color.GetAlternateColorSpace() == PdfColorSpace::DeviceCMYK);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetGrayScale(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetRed(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetGreen(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetBlue(),
-        PdfErrorCode::InternalLogic);
-
-    REQUIRE(color.GetCyan() == 1.0);
-    REQUIRE(color.GetMagenta() == 1.0);
-    REQUIRE(color.GetYellow() == 1.0);
-    REQUIRE(color.GetBlack() == 1.0);
-
-    REQUIRE(color.GetName() == "All");
-    REQUIRE(color.GetDensity() == 1.0);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCieL(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCieA(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCieB(),
-        PdfErrorCode::InternalLogic);
-
-    REQUIRE(color.ConvertToGrayScale() == PdfColor(0.0, 0.0, 0.0));
-    REQUIRE(color.ConvertToRGB() == PdfColor(0.0, 0.0, 0.0));
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.ConvertToCMYK(),
-        PdfErrorCode::CannotConvertColor);
-
-    const PdfArray COLOR_ARRAY = color.ToArray();
-    REQUIRE(COLOR_ARRAY.GetSize() == 1);
-    REQUIRE(COLOR_ARRAY[0] == PdfObject(1.0));
-}
-
-TEST_CASE("testColorSeparationNoneConstructor")
-{
-    auto color = PdfColor::CreateSeparationNone();
-
-    REQUIRE(!color.IsGrayScale());
-    REQUIRE(!color.IsRGB());
-    REQUIRE(!color.IsCMYK());
-    REQUIRE(color.IsSeparation());
-    REQUIRE(!color.IsCieLab());
-    REQUIRE(color.GetColorSpace() == PdfColorSpace::Separation);
-    REQUIRE(color.GetAlternateColorSpace() == PdfColorSpace::DeviceCMYK);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetGrayScale(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetRed(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetGreen(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetBlue(),
-        PdfErrorCode::InternalLogic);
-
-    REQUIRE(color.GetCyan() == 0.0);
-    REQUIRE(color.GetMagenta() == 0.0);
-    REQUIRE(color.GetYellow() == 0.0);
-    REQUIRE(color.GetBlack() == 0.0);
-
-    REQUIRE(color.GetName() == "None");
-    REQUIRE(color.GetDensity() == 0.0);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCieL(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCieA(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCieB(),
-        PdfErrorCode::InternalLogic);
-
-    REQUIRE(color.ConvertToGrayScale() == PdfColor(1.0, 1.0, 1.0));
-    REQUIRE(color.ConvertToRGB() == PdfColor(1.0, 1.0, 1.0));
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.ConvertToCMYK(),
-        PdfErrorCode::CannotConvertColor);
-
-    const PdfArray COLOR_ARRAY = color.ToArray();
-    REQUIRE(COLOR_ARRAY.GetSize() == 1);
-    REQUIRE(COLOR_ARRAY[0] == PdfObject(0.0));
-}
-
-TEST_CASE("testColorSeparationConstructor")
-{
-    { //alternate color is Greyscale
-        PdfColor ALTERNATE_COLOR(0.1234);
-        const double DENSITY = 0.523456;
-        const string_view NAME("Hello");
-        auto color = PdfColor::CreateSeparation(NAME, DENSITY, ALTERNATE_COLOR);
-
-        REQUIRE(!color.IsGrayScale());
-        REQUIRE(!color.IsRGB());
-        REQUIRE(!color.IsCMYK());
-        REQUIRE(color.IsSeparation());
-        REQUIRE(!color.IsCieLab());
-        REQUIRE(color.GetColorSpace() == PdfColorSpace::Separation);
-        REQUIRE(color.GetAlternateColorSpace() == PdfColorSpace::DeviceGray);
-
-        REQUIRE(ALTERNATE_COLOR.GetGrayScale() == color.GetGrayScale());
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetRed(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetGreen(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetBlue(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetCyan(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetMagenta(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetYellow(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetBlack(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetCieL(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetCieA(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetCieB(),
-            PdfErrorCode::InternalLogic);
-
-        REQUIRE(color.GetName() == NAME);
-        REQUIRE(color.GetDensity() == DENSITY);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.ConvertToGrayScale(),
-            PdfErrorCode::NotImplemented);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.ConvertToRGB(),
-            PdfErrorCode::NotImplemented);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.ConvertToCMYK(),
-            PdfErrorCode::CannotConvertColor);
-
-        const PdfArray COLOR_ARRAY = color.ToArray();
-        REQUIRE(COLOR_ARRAY.GetSize() == 1);
-        REQUIRE(COLOR_ARRAY[0] == PdfObject(DENSITY));
-    }
-
-    { //alternate color is RGB
-        const double R_VALUE = 0.023;
-        const double G_VALUE = 0.345;
-        const double B_VALUE = 0.678;
-        const PdfColor ALTERNATE_COLOR(R_VALUE, G_VALUE, B_VALUE);
-        const double DENSITY = 0.523456;
-        const string_view NAME("Hello");
-        auto color = PdfColor::CreateSeparation(NAME, DENSITY, ALTERNATE_COLOR);
-
-        REQUIRE(!color.IsGrayScale());
-        REQUIRE(!color.IsRGB());
-        REQUIRE(!color.IsCMYK());
-        REQUIRE(color.IsSeparation());
-        REQUIRE(!color.IsCieLab());
-        REQUIRE(color.GetColorSpace() == PdfColorSpace::Separation);
-        REQUIRE(color.GetAlternateColorSpace() == PdfColorSpace::DeviceRGB);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetGrayScale(),
-            PdfErrorCode::InternalLogic);
-
-        REQUIRE(ALTERNATE_COLOR.GetRed() == color.GetRed());
-        REQUIRE(ALTERNATE_COLOR.GetGreen() == color.GetGreen());
-        REQUIRE(ALTERNATE_COLOR.GetBlue() == color.GetBlue());
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetCyan(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetMagenta(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetYellow(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetBlack(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetCieL(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetCieA(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetCieB(),
-            PdfErrorCode::InternalLogic);
-
-        REQUIRE(color.GetName() == NAME);
-        REQUIRE(color.GetDensity() == DENSITY);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.ConvertToGrayScale(),
-            PdfErrorCode::NotImplemented);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.ConvertToRGB(),
-            PdfErrorCode::NotImplemented);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.ConvertToCMYK(),
-            PdfErrorCode::CannotConvertColor);
-
-        const PdfArray COLOR_ARRAY = color.ToArray();
-        REQUIRE(COLOR_ARRAY.GetSize() == 1);
-        REQUIRE(COLOR_ARRAY[0] == PdfObject(DENSITY));
-    }
-
-    { //alternate color is CMYK
-        const double C_VALUE = 0.023;
-        const double M_VALUE = 0.345;
-        const double Y_VALUE = 0.678;
-        const double K_VALUE = 0.18;
-        const PdfColor ALTERNATE_COLOR(C_VALUE, M_VALUE, Y_VALUE, K_VALUE);
-        const double DENSITY = 0.123456;
-        const string_view NAME("Hello");
-        auto color = PdfColor::CreateSeparation(NAME, DENSITY, ALTERNATE_COLOR);
-
-        REQUIRE(!color.IsGrayScale());
-        REQUIRE(!color.IsRGB());
-        REQUIRE(!color.IsCMYK());
-        REQUIRE(color.IsSeparation());
-        REQUIRE(!color.IsCieLab());
-        REQUIRE(color.GetColorSpace() == PdfColorSpace::Separation);
-        REQUIRE(color.GetAlternateColorSpace() == PdfColorSpace::DeviceCMYK);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetGrayScale(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetRed(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetGreen(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetBlue(),
-            PdfErrorCode::InternalLogic);
-
-        REQUIRE(ALTERNATE_COLOR.GetCyan() == color.GetCyan());
-        REQUIRE(ALTERNATE_COLOR.GetMagenta() == color.GetMagenta());
-        REQUIRE(ALTERNATE_COLOR.GetYellow() == color.GetYellow());
-        REQUIRE(ALTERNATE_COLOR.GetBlack() == color.GetBlack());
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetCieL(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetCieA(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetCieB(),
-            PdfErrorCode::InternalLogic);
-
-        REQUIRE(color.GetName() == NAME);
-        REQUIRE(color.GetDensity() == DENSITY);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.ConvertToCMYK(),
-            PdfErrorCode::CannotConvertColor);
-
-        const PdfArray COLOR_ARRAY = color.ToArray();
-        REQUIRE(COLOR_ARRAY.GetSize() == 1);
-        REQUIRE(COLOR_ARRAY[0] == PdfObject(DENSITY));
-    }
-
-    { //alternate color is CieLab
-        const double dCieL = 0.023;
-        const double dCieA = 0.345;
-        const double dCieB = 0.678;
-        auto ALTERNATE_COLOR = PdfColor::CreateCieLab(dCieL, dCieA, dCieB);
-        const double DENSITY = 0.523456;
-        const string_view NAME("Hello");
-        auto color = PdfColor::CreateSeparation(NAME, DENSITY, ALTERNATE_COLOR);
-
-        REQUIRE(!color.IsGrayScale());
-        REQUIRE(!color.IsRGB());
-        REQUIRE(!color.IsCMYK());
-        REQUIRE(color.IsSeparation());
-        REQUIRE(!color.IsCieLab());
-        REQUIRE(color.GetColorSpace() == PdfColorSpace::Separation);
-        REQUIRE(color.GetAlternateColorSpace() == PdfColorSpace::Lab);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetGrayScale(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetRed(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetGreen(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetBlue(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetCyan(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetMagenta(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetYellow(),
-            PdfErrorCode::InternalLogic);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.GetBlack(),
-            PdfErrorCode::InternalLogic);
-
-        REQUIRE(ALTERNATE_COLOR.GetCieL() == color.GetCieL());
-        REQUIRE(ALTERNATE_COLOR.GetCieA() == color.GetCieA());
-        REQUIRE(ALTERNATE_COLOR.GetCieB() == color.GetCieB());
-
-        REQUIRE(color.GetName() == NAME);
-        REQUIRE(color.GetDensity() == DENSITY);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.ConvertToGrayScale(),
-            PdfErrorCode::NotImplemented);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.ConvertToRGB(),
-            PdfErrorCode::NotImplemented);
-
-        ASSERT_THROW_WITH_ERROR_CODE(
-            color.ConvertToCMYK(),
-            PdfErrorCode::CannotConvertColor);
-
-        const PdfArray COLOR_ARRAY = color.ToArray();
-        REQUIRE(COLOR_ARRAY.GetSize() == 1);
-        REQUIRE(COLOR_ARRAY[0] == PdfObject(DENSITY));
-    }
-}
-
-TEST_CASE("testColorCieLabConstructor")
-{
-    const double dCieL = 0.023;
-    const double dCieA = 0.345;
-    const double dCieB = 0.678;
-    auto color = PdfColor::CreateCieLab(dCieL, dCieA, dCieB);
-
-    REQUIRE(!color.IsGrayScale());
-    REQUIRE(!color.IsRGB());
-    REQUIRE(!color.IsCMYK());
-    REQUIRE(!color.IsSeparation());
-    REQUIRE(color.IsCieLab());
-    REQUIRE(color.GetColorSpace() == PdfColorSpace::Lab);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetAlternateColorSpace(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetGrayScale(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetRed(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetGreen(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetBlue(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetCyan(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetMagenta(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetYellow(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetBlack(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetName(),
-        PdfErrorCode::InternalLogic);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.GetDensity(),
-        PdfErrorCode::InternalLogic);
-
-    REQUIRE(color.GetCieL() == dCieL);
-    REQUIRE(color.GetCieA() == dCieA);
-    REQUIRE(color.GetCieB() == dCieB);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.ConvertToGrayScale(),
-        PdfErrorCode::CannotConvertColor);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.ConvertToRGB(),
-        PdfErrorCode::CannotConvertColor);
-
-    ASSERT_THROW_WITH_ERROR_CODE(
-        color.ConvertToCMYK(),
-        PdfErrorCode::CannotConvertColor);
-
-    const PdfArray COLOR_ARRAY = color.ToArray();
-    REQUIRE(COLOR_ARRAY.GetSize() == 3);
-    REQUIRE(PdfObject(dCieL) == COLOR_ARRAY[0]);
-    REQUIRE(PdfObject(dCieA) == COLOR_ARRAY[1]);
-    REQUIRE(PdfObject(dCieB) == COLOR_ARRAY[2]);
-}
-
-TEST_CASE("testRGBtoCMYKConversions")
+TEST_CASE("TestRGBtoCMYKConversions")
 {
     using TPairOfColors = std::pair<PdfColor, PdfColor>;
     using TMapOfColors = std::map<std::string, TPairOfColors>;
@@ -2190,7 +1584,7 @@ TEST_CASE("testRGBtoCMYKConversions")
         ++iter)
     {
         const std::string COLOR_NAME(iter->first);
-        PdfColor namedColor(PdfColor::FromString(COLOR_NAME.c_str()));
+        PdfColor namedColor(PdfColor::CreateFromString(COLOR_NAME.c_str()));
         PdfColor rgbColor(iter->second.first);
         PdfColor cmykColor(iter->second.second);
 
@@ -2199,3 +1593,21 @@ TEST_CASE("testRGBtoCMYKConversions")
         REQUIRE(rgbColor == cmykColor.ConvertToRGB());
     }
 }
+
+TEST_CASE("TestSeparationColor")
+{
+    // CHECK-ME: Check if this tests does something meaningful
+    PdfMemDocument doc;
+    auto& page = doc.GetPages().CreatePage();
+    PdfColor separationColor = PdfColor(0.6, 0, 1, 0);
+    PdfColorSpaceFilterPtr sportColorFilter(new PdfColorSpaceFilterSeparation("spot_1", separationColor));
+    auto spotColorSpace = doc.CreateColorSpace(sportColorFilter);
+
+    PdfPainter painter;
+    painter.SetCanvas(page);
+    painter.GraphicsState.SetNonStrokingColorSpace(*spotColorSpace);
+    painter.GraphicsState.SetNonStrokingColor(PdfColorRaw{ 0.5 });
+    painter.DrawRectangle(Rect(100, 600, 100, 50), PdfPathDrawMode::Fill);
+    painter.FinishDrawing();
+    doc.Save(TestUtils::GetTestOutputFilePath("TestSeparationColor.pdf"));
+}
